/*
 * $Id: MirrCmdProto.h,v 1.26 2004/03/19 21:45:47 ricardo Exp $
 * $Source: /proj/ncl/cvsroot/thinc/include/MirrCmdProto.h,v $
 *
 * The THINC protocol
 *
 * Authors:
 * Ricardo Baratto
 * Leo Kim
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

/*
 * This is the THINC protocol definition. Aims to be small, efficient and 
 * extensible.
 * There are Server->Client messages and Client->Server messages. Each
 * message has a corresponding type that identifies it. Types from
 * s->c may overlap types from c->s.
 * Each message is prepended by a header, thincHeader, which currently 
 * carries the message's type. 
 *
 * Rules for adding new messages
 * -----------------------------
 *
 * For message types:
 * Include the name of the structure that your particular type represents.
 * The limit is 255
 *
 * For the messages:
 * Every message should have a clear description of what it does, it's
 * parameters, the data (if any) that follows it, and the size of its 
 * header. 
 * If any of your fields take only a set of constant values use #define's 
 * and put them below your message's definition.
 * Follow the naming conventions of the existing message types.
 * If your particular message has even further "subdivisions", include
 * the main message header definition here, and the rest in an additional
 * header file, which gets #include'd at the end of this file.
 * On Size: 
 * You should create a #define showing your message's header size.
 * Some architectures align structures to 32bit boundaries, 
 * so by rule all message structures should include enough padding to
 * account for this (char pad[blah] should do it). However, when sending
 * and receiving them, the actual size (indicated by the #define mentioned 
 * above) should be used. In addition, all the fields should explicitly tell 
 * the compiler their size using bitfields.
 * Don't use unsigned fields, java (one of our "clients") doesn't know about 
 * those.
 */
/* 
 * Current connection setup process:
 * (msg type, client state)
 * [Note: client state is used internally by the server to keep track of
 * the stage at which a client is]
 *
 * 1. Client connects and sends its string version. (T_CSTART, TCLIENT_INIT)
 * 2. If server "likes" the client's version, it sends back its version
 * along with its endian-ness information and any other information
 * relevant to the client. (T_SSTART, TCLIENT_INIT). 
 * If it doesn't "like" it, the server sends back an negative T_SACK message 
 * and closes the connection.
 * 3. If the client "likes" the server version, its endian-ness, and has
 * no problems with any additional infomartion, it sends back an ack
 * telling the server to send the initial framebuffer (T_CACK, TCLIENT_START)
 * 4. Server sends initial framebuffer and cursor information. The cursor
 * information is optional, in case the server doesn't support hardware
 * cursors (T_FB_INIT, TCLIENT_FB)
 * 5. If client is capable of handling this framebuffer and cursor, it sends 
 * back another ack telling the server to send the fb (T_CACK, TCLIENT_CONN)
 * 6. Server sends initial cursor and framebuffer to client 
 * (message its data part of T_FB_INIT, TCLIENT_CONN).
 */

/* TODO:
 * - For the video, the server should inform the client at connection time
 *   if it'll be negotiating video formats or not.
 * - Add negotiation of features in connection setup.
 * - Add padding to message's data
 * - Add maximum size to a message (with data included)
 */


#ifndef _THINC_PROTO_H_
#define _THINC_PROTO_H_

/**************************** Message header ******************************/
typedef struct mirr_header_s {
	int type;	
	int	dstId;	/* The Id of the destination surface */
	int	srcId;    /* The Id of the source surface */
} MirrCmdHeader;

/* Note: At least in x86 this structure will be aligned to 32 bits. Since
 * it doesn't make sense to waste 24 bits on every message, write()/read()'s 
 * of the header should use this #define as the size parameter */
/* FIXME: Change name to T_HEADER_SIZE */


#define THINC_HEADERSIZE sizeof(MirrCmdHeader)

/*************************************************************************
 *                          SERVER => CLIENT                             *
 *************************************************************************/

/**************************** Message types ******************************/

#define T_SSTART	1	/* thinc_sStart */
/* 2 - 9 reserved for future use */
#define T_SACK		10	/* thinc_sAck */
/* Framebuffer painting messages */
#define T_FB_INIT	11	/* thinc_fbInit */
#define T_FB_RUPDATE	12	/* thinc_rawUpdate */
#define T_FB_COPY	13	/* thinc_fbCopy */
#define T_FB_SFILL	14	/* thinc_fbSFill */
#define T_FB_PFILL	15	/* thinc_fbPFill */
#define T_FB_GLYPH	16	/* thinc_fbGlyph */
#define T_FB_BILEVEL	17	/* thinc_fbBilevel */
/* - 39 reserved for framebuffer operations */
/* Cursor events */
#define T_CURSOR_CHANGE	40	/* thinc_cursorChange */
#define T_CURSOR_SHOWHIDE	41	/* thinc_cursorShowHide */
#define T_CURSOR_MOVE	42	/* thinc_cursorMove */
#define T_CURSOR_COLOR	43	/* thinc_cursorColor */

//#define T_CREATE_BITMAP 50 /* thinc_createBitmap */
//#define T_DESTROY_BITMAP 51 /* thinc_destroyBitmap */

/* 
 * current allocations:
 * resize: 50-60
 * video: 100-110
 */
/********************** helper structures ******************************/

/* description of a rectangular region */
typedef struct thinc_rectRegion_s {
	int x;
	int y;
	int width;
	int height;
} thinc_rectRegion;

#define T_RECTREGION_SIZE	sizeof(thinc_rectRegion)

/**************************** The messages *******************************/


/* T_FB_RUPDATE
 * Description: Message sent for a raw rectangular frame buffer update. 
 * Carries the description of the location and size of the update.
 * Size: 8 bytes
 * Data: Followed by height lines of size width*bpp, representing the
 * new contents of the described frame buffer region
 */
typedef struct thinc_rawUpdate_s {
	int x;	/* coords of top-left corner of the update */
	int y;
	int width;	/* dimensions of the update */
	int height;
} thinc_rawUpdate;

#define T_FB_RUPDATE_SIZE sizeof(thinc_rawUpdate)

/* flags */				/* order */
#define T_FB_RUPDATE_COMPRESSED	0x01	/* 3 */
#define T_FB_RUPDATE_RESIZED	0x02	/* 2 */
#define T_FB_RUPDATE_CACHED	0x04	/* 1 */
#define T_FB_RUPDATE_ADDCACHE	0x08	/* 1 */

/* T_FB_COPY
 * Description: Tells the client to copy the area of size width*height, 
 * with top left corner sx,sy to dx,dy.
 * Size: 12 bytes
 * Data: NONE
 */
typedef struct thinc_fbCopy_s {
	int sx;
	int sy;
	int dx;
	int dy;
	int width;
	int height;
} thinc_fbCopy;

#define T_FB_COPY_SIZE	sizeof(thinc_fbCopy)

/* T_FB_SFILL	
 * Description: Solid Fill command. Tells the client to fill the area 
 * described by x,y,w,h with the pixel value pixel.
 * Size: 12 bytes
 * Data: NONE
 */
typedef struct thinc_fbSFill_s {
	int pixel;
	int numrects;
	/* followed by numrects thinc_rectRegion structs describing each
	 * of the regions to be filled with this colr */
} thinc_fbSFill;

#define T_FB_SFILL_SIZE	sizeof(thinc_fbSFill)

#endif

/* T_FB_PFILL
 * Description: Pixmap Fill command: tile pixmap along rectangles. x,y
 * tells the client where's the origin of the tiling region. Note that
 * the rectangles may (and most of the time will) describe a subregion
 * of the tiling region, for example if only the lower left corner of
 * the region needs to be painted. This information is necessary for
 * proper alignment of the pixmap, ie. the client shouldn't just paint
 * the pixmap starting at the origin of each of the rectangles, clipping
 * and alignment needs to be done.
 * Some implementation notes:
 * -> The pixmap will probably have a maximum size, above which some
 * other type of update should be used.
 * -> If the region to be filled is smaller than the pixmap, another
 * type of message should be used.
 * Size: 10 bytes
 * Data: The pixmap (width*height*bpp bytes) followed by numrects
 * thinc_rectRegion structures describing each of the regions that should
 * be filled with this pixmap.
 */
typedef struct thinc_fbPFill_s {
	/* the pixmap's origin */
	int x;
	int y;
	/* the pixmap size */
	int width;
	int height;
	/* number of rectangles */
	int numrects;
	/* followed by the pixmap and numrects thinc_rectRegion's (see below) */
} thinc_fbPFill;

#define T_FB_PFILL_SIZE	sizeof(thinc_fbPFill)

/* T_FB_GLYPH
 * Description: Glyph Fill command. Fill the region described by (x,y)
 * (width, height) using the bitmap as a stipple to fill the region:
 * If there's a 1 on the bitmap the color specified in the message
 * should be applied. If there's a 0 no operation should be performed
 * on that pixel.
 * Size: 12 bytes
 * Data: The bitmap to use as stipple. Bitmap's size is: 
 * ceiling(width/8)*height
 * where 8 is the number of pixels that are packed on every byte of the
 * bitmap. There's a macro defined in the sample server and client 
 * implementation for this number and for the bitmap's size, given its 
 * width and height.
 */
typedef struct thinc_fbGlyph_s {
	/* the region to fill */
	int x;
	int y;
	int width;
	int height;
	/* the color to use */
	int color;
	/* number of rectangles. if 0 just blast the glyph onto the fb*/
	int numrects;
	/* followed by a bitmap (width * height)/8 bytes long */
	/* followed by numrects thinc_rectRegion */
} thinc_fbGlyph;

#define T_FB_GLYPH_SIZE sizeof(thinc_fbGlyph)

/* T_FB_BILEVEL
 * Description: Bilevel Fill command. Fill the region described by (x,y)
 * (width, height) using the bitmap as a stipple to fill the region:
 * If there's a 1 on the bitmap the fg color specified in the message
 * should be applied. If there's a 0, the bg color specified in the
 * message should be applied.
 * Size: 16 bytes
 * Data: The bitmap to use as stipple. Bitmap's size is: 
 * ceiling(width/8)*height
 * where 8 is the number of pixels that are packed on every byte of the
 * bitmap. There's a macro defined in the sample server and client 
 * implementation for this number and for the bitmap's size, given its 
 * width and height.
 */
typedef struct thinc_fbBilevel_s {
	/* the region to fill */
	int x;
	int y;
	int width;
	int height;
	/* the colors to use */
	int fg;
	int bg;
	/* number of rectangles. If 0, just blast the bilevel onto the fb */
	int numrects;
	/* followed by a bitmap (width * height)/8 bytes long */
	/* followed by numrects thinc_rectRegion */
} thinc_fbBilevel;

#define T_FB_BILEVEL_SIZE sizeof(thinc_fbBilevel)

/* flags for glyph and bilevel */
#define T_FB_BITMAP_CACHED	0x01	/* 1 */
#define T_FB_BITMAP_ADDCACHE	0x02	/* 1 */
#define T_FB_BITMAP_RESIZED	0x04

/**************** Cursor messages ******************/

/* T_CURSOR_CHANGE
 * Description: The cursor shape needs to be changed to the shape
 * being sent as data to this message.
 * Size: 4 bytes
 * Data: A bitmap of size 2*cursor_size corresponding
 * to the coursor's source and mask.
 */
typedef struct thinc_cursorChange_s {
	int xhot;
	int yhot;
} thinc_cursorChange;

#define T_CURSOR_CHANGE_SIZE	sizeof(thinc_cursorChange)

/* T_CURSOR_SHOWHIDE
 * Description: Client should show/hide its current cursor according to
 * the contents of showhide: T_SHOW, T_HIDE.
 * Cursor Hide is an implementation dependent operation (it is required).
 * The server may choose to support it or replace it with an equivalent
 * operation (like sending a transparent cursor), and the client is free
 * to implement it if its windowing system allows such an operation,
 * implement it some other way (such as having a transparent
 * cursor on hand and using it when a Hide command comes in), or
 * ignoring it. The reasons for this:
 * -> Some windowing systems (eg. X) don't support a cursor hiding operation.
 * -> The sample server is able to implement cursor hide as a normal operation
 * or as sending a transparent cursor. It may decide either way depending
 * on the knowledge it has of what its clients support.
 * Size: 1 byte
 * Data: None
 */
typedef struct thinc_cursorShowHide_s {
	int showhide;
} thinc_cursorShowHide;

#define T_SHOW	1
#define T_HIDE 	0

#define T_CURSOR_SHOWHIDE_SIZE	sizeof(thinc_cursorShowHide)

/* T_CURSOR_MOVE
 * Description: Client should move the cursor to the specified position.
 * Size: 4 bytes
 * Data: None
 */
typedef struct thinc_cursorMove_s {
	int x;
	int y;
} thinc_cursorMove;

#define T_CURSOR_MOVE_SIZE 	sizeof(thinc_cursorMove)

/* T_CURSOR_COLOR 
 * Description: Change the cursor's fg and bg colors.
 * Size: 8 bytes
 * Data: None
 */
typedef struct thinc_cursorColor_s {
	int fg;
	int bg;
} thinc_cursorColor;

#define T_CURSOR_COLOR_SIZE	sizeof(thinc_cursorColor)

