/* 
 *
 * Command structure definitions
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop = 8 */
#ifndef _THINC_COMMAND_H_
#define _THINC_COMMAND_H_

typedef struct thincCmd_S {
	int 		type:16;	/* the command type(thincProto.h)*/
	int		flags:16;	/* additional info about command */
	unsigned long	id;		/* An id which identifies the
					   command globally. Mostly used
					   for debugging. */
	BoxRec		area;		/* the area that's affected by this
					   command */
	int 		cmdSize;	/* size of command following us */
	int		refcnt:8;	/* XXX Deprecated XXX */
	/* A command may be in at most 2 lists at a time. Currently
	 * commands are most of the time in one list, and only in 2 when 
	 * they're buffered and the scheduler code puts them in a queue.
	 * So for now, we're ok ;) */
	thincListRec list;
	thincListRec sched;
	/* followed by the actual command (of size cmdSize bytes) */
} thincCmdRec, *thincCmdPtr;

#define T_CMDREC_SIZE	(sizeof(thincCmdRec))

/* return a pointer to the data part of this command. Not just (cmd+1)
 * since in the future we may want to align the command and its data
 * in some particular way and things would break */
#define cmd_data(cmd, type) ((type) ((char*)(cmd) + T_CMDREC_SIZE))

/* return a pointer to the lists in the command */
#define cmd_list(cmd) (&(cmd)->list)
#define cmd_sched(cmd) (&(cmd)->sched)

/** flags **/
#define T_NO_FLAGS	0x0000

#define T_CMD_NO_OVER	0x0001 /* don't overwrite this command */
#define T_CMD_TRANS 	0x0002	/* this command doesn't overwrite any
				   other commands */
/* sched flags */
#define T_CMD_RT	0x0080	/* command is high priority (real time) */

/* dependency flags */
#define T_CMD_DEPS_NEEDS 0x0010	/* the command requests that dependencies be found
				   whenever it's queued or buffered */
#define T_CMD_DEPS_MASTER 0x0020 /* another command(s) depend(s) on this one */
#define T_CMD_DEPS_SLAVE 0x0040	/* command depends on somebody else */ 

/* command-specific flags. careful they don't overlap with the generic
 * flags above */
#define T_BIT_FILL	0x8000	/* differentiate between a bitmap fill and just
				   a bitmap */

/** XXX Deprecated flags XXX **/
#define	T_CMD_INVALID	0x0004 /* command should be destroyed as soon as 
				all references to it are gone */

/*** per-command structures ***/
typedef struct rawData_S {
	unsigned int refcount;
	unsigned char *data;
} rawData;

struct rawCmd {
	unsigned  datasize:32;	/* we may need to store the actual raw
				   data inside the command instead of
				   getting it at flush time. If this is > 0,
				   then that's how much data we have in
				   store */
	BoxRec orig;		/* see thincCommandRaw.c for this */
	RegionRec reg;		/* Data points outside of the command block.
				   And that's ok */
#ifdef T_USERENDER
	int Bpp;		/* the color depth and bytes per pixel of
				   this update. We need this for RENDER
				   since for antialised text we send RAW
				   updates which are only 8bbp */
#endif
	rawData* data;
};

struct copyCmd {
	int deltax:16;
	int deltay:16;
	RegionRec reg;		/* The destination region. Be careful,
				   everywhere else, the region for a copy
				   is considered to be the source */
};

/* solid fill */
struct fillCmd {
	int color:32;		/* fill color */
	RegionRec reg;		/* the region for the fill */
};

/* pixmap fill */
struct pixmapFillCmd {
	int size:32;
	int x:16;
	int y:16;
	int w:16;
	int h:16;
	RegionRec reg;
	/* followed by pixmap */
};

/* bilevel fill */
struct bilevelCmd {
	/* bilevel info */
	int x:16;		
	int y:16;
	int w:16;
	int h:16;
	int fg:32;	/* bilevel colors */
	int bg:32;
	int size:32;	/* size of bitmap data */
	BoxRec orig_area;
	RegionRec reg;
	/* followed by bilevel data */
};

/* glyph fill */
struct glyphCmd {
	/* glyph info */
	int x:16;		
	int y:16;
	int w:16;
	int h:16;
	int color:32;		/* glyph color */
	int size:32;		/* size of bitmap data */
	int sched_sz:32;	/* size for the scheduler (for deps) */
	BoxRec orig_area;
	RegionRec reg;
	/* followed by glyph data */
};

/* cursor cmd defined in thincCursor.c */

/***** The functions to manipulate commands *****/

/* CREATE 
 * Return a pointer to a new command 
 * Return NULL on error
 */
typedef thincCmdPtr (*cmdCreateFunc) (
	thincScreenPtr 	/* pThinc */,
	RegionPtr	/* pReg */,
	va_list		/* command specific data goes here */
);

/* DESTROY 
 * Free memory used by passed command
 */
typedef void (*cmdDestroyFunc) (
	thincScreenPtr	/* pThinc */,
	thincCmdPtr	/* pCmd */
);

/* COPY 
 * Return a copy of the passed command. Return NULL on error
 */
typedef thincCmdPtr (*cmdCopyFunc) (
	thincScreenPtr	/* pThinc */,
	thincCmdPtr	/* pSrc */
);

/* MODIFY 
 * Modify the command. If clip is not NULL, the command should be clipped
 * to that region. If deltax and deltay are non-zero, the command should
 * be translated by that amount (could be positive or negative).
 * It should return the modified command, or NULL on error. 
 */
typedef thincCmdPtr (*cmdModifyFunc) (
	thincScreenPtr	/* pThinc */,
	thincCmdPtr	/* pCmd */,
	int		/* what modification? (see below) */,
	va_list		/* the args for the modification */
);

/* Modify commands */
#define T_CMD_CLIP	0x01 	/* clip because surface is smaller */
#define T_CMD_OVER	0x02 	/* clip because something's writing over */
#define T_CMD_MOVE	0x04
#define T_CMD_MERGE	0x08
#define T_CMD_SAVE	0x10
#define T_CMD_SCHEDSZ	0x20	/* size command should use for scheduling
				   (for correct dependency tracking) */

/* Returns the specified info from the command (see below). It can 
 * return multiple things at a time. Everything is returned through
 * additional parameters. Each command may support specific queries which
 * don't make sense for other commands. It's up to the command what to do
 * if an unknown query is made to it (normally it should just return an
 * error) */
typedef void (*cmdGetInfoFunc) (
	thincScreenPtr	/* pThinc */,
	thincCmdPtr	/* pCmd */,
	int 		/* info -> what do we want? */,
	...
);

#define T_CMD_REGION	0x01	/* the region to which this command draws */
#define T_CMD_SIZE	0x02	/* the size of the command when it's sent
				 * over the wire */
#define T_CMD_BITMAP	0x04	/* specific to bitmap commands. Gets bitmap
				   and bitmap size */

/* EXECUTE
 * Executes the passed command. Sort of misnamed, since it's purpose is to
 * pass commands from offscreen pixmap to client's queue. */
typedef void (*cmdExecuteFunc) (
	thincScreenPtr	/* pThinc */,
	WindowPtr	/* pWindow */,
	thincCmdPtr	/* pCmd */,
	RegionPtr	/* clip */,
	int		/* deltax */,
	int		/* deltay */
);

/* FLUSH
 * Flushes the command from the client's buffer. See below for return value */
typedef int (*cmdFlushFunc) (
	thincScreenPtr	/* pThinc */,
	thincClientPtr	/* cl */,
	thincCmdPtr	/* cmd */
);

/***/
#define T_MAXCMD_NAME	31

typedef struct thincCmdInfo_S {
	int	cmdId;		/* the command's id. see below for 
				   non-standard id's */
	cmdCreateFunc	cmdCreate;
	cmdDestroyFunc	cmdDestroy;
	cmdCopyFunc	cmdCopy;
	cmdModifyFunc 	cmdModify;
	cmdGetInfoFunc cmdGetInfo;
	cmdExecuteFunc	cmdExecute;
	cmdFlushFunc	cmdFlush;
	char 		cmdName[T_MAXCMD_NAME+1];  /* the command's name */
} thincCmdInfoRec, *thincCmdInfoPtr;

/* === ID stuff === */
/* This should be kept in sync with thincProto -> if fb commands are
 * moved to lower/higher numbers, change this. 
 * btw, with 10 we waste 2 slots in the command list... oh well, 64 bytes 
 * down the sink :*( */
#define T_FIRST_CMD		10
#define thincCmd_idx(id) 	((id)-T_FIRST_CMD)
/* cheating: all the cursor messages start at 40 so we can't use them
 * directly as indexes in table. So we create a generic cursor command
 * which wraps all the others. This is its id. Be careful if the protocol
 * messages change numbers */
#define T_FB_CURSOR		T_FB_LAST_MSG+1

unsigned long thinc_getCmdID();

/* allocate a new thincCmdInfoRec. Returns NULL on error
 * (in thincPixmap.c) */
thincCmdInfoPtr thincAllocCmdInfo(int cmdId, char* cmdName, int namelen, 
	cmdCreateFunc create, cmdDestroyFunc destroy, cmdCopyFunc copy,
	cmdModifyFunc modify, cmdExecuteFunc execute, 
	cmdGetInfoFunc getinfo, cmdFlushFunc flush);
void thincDestroyCmdInfo(thincCmdInfoPtr cmd);

/**** The functions for the current commands ****/

/* Create Functions */
thincCmdPtr thincRawCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data);
thincCmdPtr thincRawCreateV(thincScreenPtr pThinc, RegionPtr pReg, ...);
int thincRawInit(thincScreenPtr pThinc, thincCmdPtr cmd, RegionPtr pReg);
thincCmdPtr thincCopyCreate(thincScreenPtr pThinc, RegionPtr pReg,
	va_list data);
thincCmdPtr thincSFillCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data);
thincCmdPtr thincSFillCreateV(thincScreenPtr pThinc, RegionPtr pReg, ...);
thincCmdPtr thincPFillCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data);
thincCmdPtr thincGlyphCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data);
thincCmdPtr thincBilevelCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data);
thincCmdPtr thincCursorCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data);

/* Destroy Functions */
void thincRawDestroy(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincCopyDestroy(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincSFillDestroy(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincPFillDestroy(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincGlyphDestroy(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincBilevelDestroy(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincCursorDestroy(thincScreenPtr pThinc, thincCmdPtr cmd);

/* Copy Functions */
thincCmdPtr thincRawCopy(thincScreenPtr pThinc, thincCmdPtr cmd);
thincCmdPtr thincCopyCopy(thincScreenPtr pThinc, thincCmdPtr cmd);
thincCmdPtr thincSFillCopy(thincScreenPtr pThinc, thincCmdPtr cmd);
thincCmdPtr thincPFillCopy(thincScreenPtr pThinc, thincCmdPtr cmd);
thincCmdPtr thincGlyphCopy(thincScreenPtr pThinc, thincCmdPtr cmd);
thincCmdPtr thincBilevelCopy(thincScreenPtr pThinc, thincCmdPtr cmd);
thincCmdPtr thincCursorCopy(thincScreenPtr pThinc, thincCmdPtr cmd);

/* Modify Functions */
thincCmdPtr thincRawModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args);
thincCmdPtr thincCopyModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args); 
thincCmdPtr thincSFillModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args);
thincCmdPtr thincPFillModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args);
thincCmdPtr thincGlyphModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args);
thincCmdPtr thincBilevelModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args);
thincCmdPtr thincCursorModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args);

/* get affected region */
void thincRawGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...);
void thincCopyGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...);
void thincSFillGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...);
void thincPFillGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...);
void thincGlyphGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...);
void thincBilevelGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...);
void thincCursorGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...);

/* Execute Functions */
void thincRawExecute(thincScreenPtr pThinc, WindowPtr pWindow, thincCmdPtr cmd, 
	RegionPtr clip, int deltax, int deltay);
void thincCopyExecute(thincScreenPtr pThinc, WindowPtr pWindow, 
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay);
void thincSFillExecute(thincScreenPtr pThinc,  WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay);
void thincPFillExecute(thincScreenPtr pThinc, WindowPtr pWindow, 
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay);
void thincGlyphExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay);
void thincBilevelExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay);
void thincCursorExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay);

/* Flush Functions */
int thincRawFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd);
int thincCopyFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd); 
int thincSFillFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd);
int thincPFillFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd);
int thincGlyphFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd);
int thincBilevelFlush(thincScreenPtr pThinc, thincClientPtr cl, 
	thincCmdPtr cmd);
int thincCursorFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd);

/*** Functions which make calling command functions easier ***/

/* add a command to the list */
int thincAddCmdList(thincListPtr list, thincCmdPtr cmd);
#ifdef T_WIN32
int thincAddCmdListHead(thincListPtr list, thincCmdPtr cmd);
#endif
int thincAddCmdSched(thincListPtr list, thincCmdPtr cmd);
int thincCommandInvalid(thincCmdPtr cmd);
thincCmdPtr thincCreateCmd(thincScreenPtr pThinc, int type, RegionPtr reg, 
	va_list args);
thincCmdPtr thincModifyCommand(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, ...);
thincCmdPtr thincModifyCommandV(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args);
/* modify a command in a node. If the command is clipped out after
 * the modification, it is taken out of the list */
int thincModifyCmdInList(thincScreenPtr pThinc, thincCmdPtr cmd, int mod, ...);
void thincDestroyCmdListBuffer(thincScreenPtr pThinc, thincListPtr list);
void thincDestroyCmdListSched(thincScreenPtr pThinc, thincListPtr list);
void thincEmptyCmdListBuffer(thincScreenPtr pThinc, thincListPtr buffer);
void thincEmptyCmdListSched(thincScreenPtr pThinc, thincListPtr queue);
void thincDestroyCmd(thincScreenPtr pThinc, thincCmdPtr cmd);
/* takes the command out of the list and then destroys it */
void thincDeleteCmdAll(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincDeleteCmdList(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincDeleteCmdSched(thincScreenPtr pThinc, thincCmdPtr cmd);

void thincCmdGetRegion(thincScreenPtr pThinc, thincCmdPtr cmd, RegionRec **reg);
unsigned int thincCmdGetSize(thincScreenPtr pThinc, thincCmdPtr cmd);
void thincCmdGetBitmap(thincScreenPtr pThinc, thincCmdPtr cmd, 
	unsigned char **bitmap, int *bitmap_size);
/* Modifies all commands in list which are overwritten by a command
 * drawing in region pReg */
int thincOverwriteCmds(thincScreenPtr pThinc, thincListPtr list,
	RegionPtr pReg);
int thincTryMergeCmd(thincScreenPtr pThinc, thincListPtr list, 
	thincCmdPtr cmd);
int thincCmdSaveCopy(thincScreenPtr pThinc, thincListPtr list, 
	thincCmdPtr copy_cmd);
/* tries to merge to COPY commands */
int thincCopyMerge(thincScreenPtr pThinc, thincCmdPtr one, thincCmdPtr two);

/* returned by thincMergeBoxes() to tell us where merging happened */
#define T_MERGE_TOP 0	
#define T_MERGE_RIGHT  1
#define T_MERGE_BOTTOM 2
#define T_MERGE_LEFT 3
/* merge mrect into rect */
int thincMergeBoxes(BoxPtr rect, BoxPtr mrect, int *where);

/* make a copy of a command */
thincCmdPtr thincCopyCommand(thincScreenPtr pThinc, thincCmdPtr cmd);
/* flush a command */
int thincFlushCmd(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd);

/* additional useful things */
extern RegionRec thincNullReg;
extern BoxRec thincNullBox;

/* misc functions */
void thincRawWriteToFile(char *data, unsigned int w, 
	unsigned int h, unsigned int Bpp, unsigned int pitch);

/***** Command Scheduler (see thincSched.c) *****/
/* 10 normal + 1 high priority */
#define T_SCHED_NORMALQUEUES	10
#define T_SCHED_RTQUEUES	1
#define T_SCHED_NUMQUEUES	(T_SCHED_NORMALQUEUES+T_SCHED_RTQUEUES)

#define T_SCHED_FIRST	11	/* first queue starts at 2^T_SCHED_FIRST */
#define T_SCHED_MAXSIZE	( 1<<(T_SCHED_FIRST+T_SCHED_NORMALQUEUES-1) )

#define T_SCHED_RT_IDX	0

/* size that should be returned to get in the first queue */
#define T_SCHED_1STQ_SIZE	(1<<T_SCHED_FIRST)

typedef struct thincSched_S {
	/* TODO: Make this a dynamic array of num_queues entries */
	thincListRec	queues[T_SCHED_NUMQUEUES];
} thincSchedRec, *thincSchedPtr;

int thincSchedInit(thincClientPtr cl);
void thincSchedDestroy(thincClientPtr cl);
int thincAddCommand(thincClientPtr cl, thincCmdPtr cmd);
int thincFlushScheduler(thincScreenPtr pThinc, thincClientPtr cl);

#define T_RESPONSE	10	/* ms */

int thincShouldYield(thincClientPtr cl);
int thincClientsHaveEvents(thincScreenPtr pThinc);

/*** Dependency tracking ****/
int thincCmdFindDeps(thincScreenPtr pThinc, thincCmdPtr cmd, thincListPtr list);

#endif
