/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Functions for offscreen manipulation of FILL (solid/pixmap) commands
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

#include "thinc.h"

static int thincSFillMerge(thincScreenPtr pThinc, thincCmdPtr cmd,
	thincCmdPtr merge);
static int thincPFillMerge(thincScreenPtr pThinc, thincCmdPtr cmd,
	thincCmdPtr merge);
static int thincPFillSize(struct pixmapFillCmd *fill);
static int thincSFillSize(struct fillCmd *fill);

/* Fill Cmd is stored:
 * --------------------
 * | CmdRec | struct  |
 * |        | fillCmd |
 * --------------------
 *
 */
/* data contains the fill color */
thincCmdPtr thincSFillCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data)
{
	t_dbg_name(func)

	unsigned long color;
	struct fillCmd *fill;
	BoxPtr ext;
	int totalsize, cmdsize;
	char *thecmd;
	thincCmdPtr cmd;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: size of fillCmd: %d bytes\n", func,
		sizeof(struct fillCmd)));

	color = va_arg(data, unsigned long);

	cmdsize = sizeof(struct fillCmd);

	totalsize = cmdsize + T_CMDREC_SIZE;

	thincDbg(T_DBG_CMD, ("%s: fill color %#lx, "
		"command size %d, total size %d\n", func, color, cmdsize, 
		totalsize));
	print_reg(T_DBG_CMD, pThinc->pScreen, pReg, "fill region:");

	thecmd = malloc(totalsize);
	if(!thecmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
			"memory for FILL command\n", func));
		return NULL;
	}

	/* standard command stuff */
	cmd = (thincCmdPtr)thecmd;
	cmd->type = T_FB_SFILL;
	cmd->flags = 0;
	cmd->id = thinc_getCmdID();
	ext = REGION_EXTENTS(pThinc->pScreen, pReg);
	copy_boxp(&cmd->area, ext);
	cmd->cmdSize = cmdsize;

	/* fill fillCmd struct */
	fill = cmd_data(cmd, struct fillCmd*);

	fill->color = color;
	fill->reg = thincNullReg;
	if(!REGION_COPY(pThinc->pScreen, &fill->reg, pReg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy FILL "
			"command's region\n", func));
		free(thecmd);
		return NULL;
	}

	/* the end */
	thincDbg(T_DBG_CMD, ("%s: finished. command's id %#lx\n", func,
		cmd->id));
	stats_end(pThinc, t1, t2, sfillStats.create);

	return cmd;
}

thincCmdPtr thincSFillCreateV(thincScreenPtr pThinc, RegionPtr pReg, ...)
{
	va_list args;
	thincCmdPtr cmd;

	va_start(args, pReg);
	cmd = thincSFillCreate(pThinc, pReg, args);
	va_end(args);

	return cmd;
}

/* Pixmap Fill command is stored:
 * ------------------------------
 * | CmdRec | struct   | pixmap |
 * |        | pfillCmd |        |
 * ------------------------------
 *
 * data has:
 * the pixmap data (unsigned char *)
 * the pixmap coordinates (int x, y, w, h)
 */
thincCmdPtr thincPFillCreate(thincScreenPtr pThinc, RegionPtr pReg, 
	va_list data)
{
	t_dbg_name(func)

	struct pixmapFillCmd *pixmap;
	unsigned char *pixdata, *datacopy;
	int x, y, w, h;

	thincCmdPtr cmd;
	char *thecmd;
	BoxPtr ext;

	int totalsize, cmdsize, pixmapsize;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: size of pixmapFillCmd: %d bytes\n", 
		func, sizeof(struct pixmapFillCmd)));

	pixdata = va_arg(data, unsigned char*);
	x = va_arg(data, int);
	y = va_arg(data, int);
	w = va_arg(data, int);
	h = va_arg(data, int);

	pixmapsize = h*t_pix_to_bytes(w,pThinc);

	cmdsize = sizeof(struct pixmapFillCmd);
	cmdsize += pixmapsize;

	totalsize = cmdsize + T_CMDREC_SIZE;

	thincDbg(T_DBG_CMD, ("%s: pixmap size %d. "
		"Need to allocate %d bytes\n", func, pixmapsize,
		totalsize));

	thecmd = malloc(totalsize);
	if(!thecmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for PFILL command\n", func));
		return NULL;
	}
	
	/* pack the command */
	cmd = (thincCmdPtr)thecmd;
	cmd->type = T_FB_PFILL;
	cmd->flags = 0;
	cmd->id = thinc_getCmdID();
	ext = REGION_EXTENTS(pThinc->pScreen, pReg);
	copy_boxp(&cmd->area, ext);
	cmd->cmdSize = cmdsize;

	/* fill pixmap struct */
	pixmap = cmd_data(cmd, struct pixmapFillCmd*);
	pixmap->size = pixmapsize;
	pixmap->x = x;
	pixmap->y = y;
	pixmap->w = w;
	pixmap->h = h;

	pixmap->reg = thincNullReg;
	if(!REGION_COPY(pThinc->pScreen, &pixmap->reg, pReg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy PFILL "
			"command's region\n", func));
		free(thecmd);
		return NULL;
	}
	
	/* copy pixmap */
	datacopy = (unsigned char*) (pixmap+1);
	memcpy(datacopy, pixdata, pixmapsize);

	/* that's it */
	thincDbg(T_DBG_CMD, ("%s: finished. command's id %#lx\n", func,
		cmd->id));

	stats_end(pThinc, t1, t2, pfillStats.create);
	return cmd;
}

/*** DESTROY ***/

void thincSFillDestroy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	char *sfillcmd = (char*)cmd;
	struct fillCmd *fill;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: destroying command %#lx\n", func,
		cmd->id));

	fill = cmd_data(cmd, struct fillCmd*);
	REGION_UNINIT(pThinc->pScreen, &fill->reg);

	free(sfillcmd);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, sfillStats.destroy);
}

void thincPFillDestroy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	char *pfillcmd = (char*)cmd;
	struct pixmapFillCmd *pfill;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: destroying command %#lx\n", func,
		cmd->id));

	pfill = cmd_data(cmd, struct pixmapFillCmd*);
	REGION_UNINIT(pThinc->pScreen, &pfill->reg);

	free(pfillcmd);

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, pfillStats.destroy);
}

/*** COPY ***/

thincCmdPtr thincSFillCopy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)

	char *newcmd, *oldcmd;
	int totalsize;

	struct fillCmd *nfill, *ofill;
	thincCmdPtr ncmd;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: copying command of size %d\n", func,
		cmd->cmdSize));

	totalsize = cmd->cmdSize + T_CMDREC_SIZE;

	newcmd = malloc(totalsize);
	if(!newcmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for copy of command\n", func));
		return NULL;
	}

	oldcmd = (char*)cmd;
	memcpy(newcmd, oldcmd, totalsize);

	/* copy the region */
	ncmd = (thincCmdPtr) newcmd;
	nfill = cmd_data(ncmd, struct fillCmd*);
	ofill = cmd_data(cmd, struct fillCmd*);
	nfill->reg = thincNullReg;

	if(!REGION_COPY(pThinc->pScreen, &nfill->reg, &ofill->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy region "
			"of FILL command\n", func));
		free(newcmd);
		return NULL;
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, sfillStats.copy);

	return (thincCmdPtr)newcmd;
}

thincCmdPtr thincPFillCopy(thincScreenPtr pThinc, thincCmdPtr cmd)
{
	t_dbg_name(func)
	char *newcmd, *oldcmd;
	int totalsize;
	struct pixmapFillCmd *nfill, *ofill;
	thincCmdPtr ncmd;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: copying command of size %d\n", func,
		cmd->cmdSize));

	totalsize = cmd->cmdSize + T_CMDREC_SIZE;

	newcmd = malloc(totalsize);
	if(!newcmd) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for copy of command\n", func));
		return NULL;
	}

	oldcmd = (char*)cmd;
	memcpy(newcmd, oldcmd, totalsize);

	/* copy the region */
	ncmd = (thincCmdPtr) newcmd;
	nfill = cmd_data(ncmd, struct pixmapFillCmd*);
	ofill = cmd_data(cmd, struct pixmapFillCmd*);
	nfill->reg = thincNullReg;

	if(!REGION_COPY(pThinc->pScreen, &nfill->reg, &ofill->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't copy region "
			"of PFILL command\n", func));
		free(newcmd);
		return NULL;
	}

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, pfillStats.copy);
	return (thincCmdPtr)newcmd;
}

/*** MODIFY ***/

thincCmdPtr thincSFillModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	t_dbg_name(func)
	struct fillCmd *fill;
	RegionPtr reg;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);
	
	thincDbg(T_DBG_CMD, ("%s: modifying command %#lx, modifications "
                "%d\n", func, cmd->id, mod));

	/* retrieve the fill command */
	fill = cmd_data(cmd, struct fillCmd*);
	reg = &fill->reg;

	/* ignore OVER */
#ifdef T_WIN32
	if((mod & T_CMD_CLIP)|| (mod & T_CMD_OVER))  {
#else
	if(mod & T_CMD_CLIP) {
#endif 
		/* clip it */
		RegionPtr clip = va_arg(args, RegionPtr);
                REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
                if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
                        thincDbg(T_DBG_CMD, ("%s: command has been "
                                "completely obscured. finished\n", func));
                        return NULL;
                }

                thincDbg(T_DBG_CMD, ("%s: clipped the region, it now "
                        "has %ld rectangles\n", func, REGION_NUM_RECTS(reg)));
                print_reg(T_DBG_CMD,pThinc->pScreen,reg,"clipped region:");
        }

        if(mod & T_CMD_MOVE) {
                /* translate */
                int deltax = va_arg(args, int);
                int deltay = va_arg(args, int);
                REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
                
                print_reg(T_DBG_CMD, pThinc->pScreen, reg, 
                        "translated region:");
        }

	if(mod & T_CMD_MERGE) {
		thincCmdPtr merge = va_arg(args, thincCmdPtr);
		int *merged = va_arg(args, int*);
		int ret;

		if ( (ret = thincSFillMerge(pThinc, cmd, merge)) == T_ERROR ) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't merge "
				"commands\n", func));
			return (thincCmdPtr) T_ERROR;
		} else if(ret == T_TRUE) {
			thincDbg(T_DBG_CMD, ("%s: merged!\n", func));
			*merged = T_TRUE;
		} else {
			thincDbg(T_DBG_CMD, ("%s: Couldn't merge commands, "
				"sorry\n", func));
			*merged = T_FALSE;
		}
	}

	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, sfillStats.modify);
	return cmd;
}

#ifdef T_WIN32

//
// FIXME: It is just a hack. Merge them if their pixmaps have the same width and height
// 
//

static int thincPFillMerge(thincScreenPtr pThinc, thincCmdPtr cmd,
	thincCmdPtr merge)
{
	t_dbg_name(func)
	struct pixmapFillCmd *pfill, *mpfill;
	RegionPtr reg, mreg;
#if 0	/* part of #if 0 below */
	BoxPtr box, mbox;
#endif
	int ret;

	pfill = cmd_data(cmd, struct pixmapFillCmd*);
	reg = &pfill->reg;
	mpfill = cmd_data(merge, struct pixmapFillCmd*);
	mreg = &mpfill->reg;

	if (pfill->w != mpfill->w || pfill->h != mpfill->h){
		thincDbg(T_DBG_CMD, ("%s: commands have different color. "
			"finished\n", func));
		return T_FALSE;
	}

#if 0
	if ( (REGION_NUM_RECTS(reg) != 1 || REGION_NUM_RECTS(mreg) != 1) ) {
		thincDbg(T_DBG_CMD, ("%s: commands not simple enough, "
			"sorry. finished\n", func));
		return T_FALSE;
	}

	box = REGION_RECTS(reg);
	mbox = REGION_RECTS(mreg);

	/* we don't care where they merged */
	ret = thincMergeBoxes(box, mbox, NULL);
	if(ret) {
		/* can it happen that we have just one rectangle and that one
		 * is not the extents? well, just in case... */
		copy_boxp(&reg->extents, box);
	}
#else
	REGION_UNION(pThinc->pScreen, reg, reg, mreg);
	ret = T_TRUE;
#endif

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	return ret;
}

#endif

static int thincSFillMerge(thincScreenPtr pThinc, thincCmdPtr cmd,
	thincCmdPtr merge)
{
	t_dbg_name(func)
	struct fillCmd *fill, *mfill;
	RegionPtr reg, mreg;
#if 0	/* part of #if 0 below */
	BoxPtr box, mbox;
#endif
	int ret;

	fill = cmd_data(cmd, struct fillCmd*);
	reg = &fill->reg;
	mfill = cmd_data(merge, struct fillCmd*);
	mreg = &mfill->reg;

	if(fill->color != mfill->color) {
		thincDbg(T_DBG_CMD, ("%s: commands have different color. "
			"finished\n", func));
		return T_FALSE;
	}

#if 0
	if ( (REGION_NUM_RECTS(reg) != 1 || REGION_NUM_RECTS(mreg) != 1) ) {
		thincDbg(T_DBG_CMD, ("%s: commands not simple enough, "
			"sorry. finished\n", func));
		return T_FALSE;
	}

	box = REGION_RECTS(reg);
	mbox = REGION_RECTS(mreg);

	/* we don't care where they merged */
	ret = thincMergeBoxes(box, mbox, NULL);
	if(ret) {
		/* can it happen that we have just one rectangle and that one
		 * is not the extents? well, just in case... */
		copy_boxp(&reg->extents, box);
	}
#else
	REGION_UNION(pThinc->pScreen, reg, reg, mreg);
	ret = T_TRUE;
#endif

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	return ret;
}

thincCmdPtr thincPFillModify(thincScreenPtr pThinc, thincCmdPtr cmd, 
	int mod, va_list args)
{
	t_dbg_name(func)

	struct pixmapFillCmd *fill;
	RegionPtr reg;
	RegionPtr clip;
	int deltax, deltay;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);
	
	thincDbg(T_DBG_CMD, ("%s: modifying command %#lx, modifications "
                "%d\n", func, cmd->id, mod));

	/* retrieve the fill command */
	fill = cmd_data(cmd, struct pixmapFillCmd*);
	reg = &fill->reg;

	/* ignore OVER */
#ifdef T_WIN32
	if((mod & T_CMD_CLIP)|| (mod & T_CMD_OVER))  {
#else
	if(mod & T_CMD_CLIP) {
#endif
		/* clip it */
		clip = va_arg(args, RegionPtr);
		REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
		if(!(REGION_NOTEMPTY(pThinc->pScreen, reg))) {
			thincDbg(T_DBG_CMD, ("%s: command has been "
				"completely obscured. finished\n", func));
			return NULL;
		}

		thincDbg(T_DBG_CMD, ("%s: clipped the region, it now "
			"has %ld rectangles\n", func, REGION_NUM_RECTS(reg)));
		print_reg(T_DBG_CMD,pThinc->pScreen,reg,"clipped region:");
	}

	if(mod & T_CMD_MOVE) {
		/* translate */
		deltax = va_arg(args, int);
		deltay = va_arg(args, int);
		REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
		fill->x += deltax;
		fill->y += deltay;

		print_reg(T_DBG_CMD, pThinc->pScreen, reg, 
			"translated region:");
	}
#ifdef T_WIN32
	/* FIX_FOR_WIN32*/
	if(mod & T_CMD_MERGE) {
		thincCmdPtr merge = va_arg(args, thincCmdPtr);
		int *merged = va_arg(args, int*);
		int ret;
		
		if ( (ret = thincPFillMerge(pThinc, cmd, merge)) == T_ERROR ) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't merge "
				"commands\n", func));
			return (thincCmdPtr) T_ERROR;
		} else if(ret == T_TRUE) {
			thincDbg(T_DBG_CMD, ("%s: merged!\n", func));
			*merged = T_TRUE;
		} else {
			thincDbg(T_DBG_CMD, ("%s: Couldn't merge commands, "
				"sorry\n", func));
			*merged = T_FALSE;
		}
	}
#endif


	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));

	stats_end(pThinc, t1, t2, pfillStats.modify);
	return cmd;
}

void thincSFillGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...)
{
	va_list args;
	struct fillCmd *fill = cmd_data(cmd, struct fillCmd*);
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	va_start(args, info);

	if(info & T_CMD_REGION) {
		RegionRec **reg = va_arg(args, RegionRec**);
		*reg = &fill->reg;
	}

	if(info & T_CMD_SIZE) {
		unsigned int *size = va_arg(args, unsigned int*);
		*size = thincSFillSize(fill);
		/* cheat the scheduler. very hackish, but it's really the
		 * most efficient way to deal with pixmap and solid fills
		 * not being broken up as they're overwritten */
		if(pThinc->doSched) {
			if(*size >= T_SCHED_1STQ_SIZE)
				*size = T_SCHED_1STQ_SIZE - 1;
		}
	}

	va_end(args);
	stats_end(pThinc, t1, t2, sfillStats.info);
}

static int thincSFillSize(struct fillCmd *fill)
{
	unsigned int total = T_FB_SFILL_SIZE + THINC_HEADERSIZE;
	total += REGION_NUM_RECTS(&fill->reg) * T_RECTREGION_SIZE;
	return total;
}

void thincPFillGetInfo(thincScreenPtr pThinc, thincCmdPtr cmd, int info, ...)
{
	va_list args;
	struct pixmapFillCmd *fill = cmd_data(cmd, struct pixmapFillCmd*);
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	va_start(args, info);

	if(info & T_CMD_REGION) {
		RegionRec **reg = va_arg(args, RegionRec**);
		*reg = &fill->reg;
	}

	if(info & T_CMD_SIZE) {
		unsigned int *size = va_arg(args, unsigned int*);
		*size = thincPFillSize(fill);
		/* cheat the scheduler. very hackish, but it's really the
		 * most efficient way to deal with pixmap and solid fills
		 * not being broken up as they're overwritten */
		if(pThinc->doSched) {
			if(*size >= T_SCHED_1STQ_SIZE)
				*size = T_SCHED_1STQ_SIZE - 1;
		}
	}

	va_end(args);
	stats_end(pThinc, t1, t2, pfillStats.info);
}

static int thincPFillSize(struct pixmapFillCmd *fill)
{
	unsigned int total = T_FB_PFILL_SIZE + THINC_HEADERSIZE;
	total += fill->size;
	total += REGION_NUM_RECTS(&fill->reg) * T_RECTREGION_SIZE;

	return total;
}

/*** EXECUTE ***/

void thincSFillExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay)
{
	t_dbg_name(func)
	struct fillCmd *fill;
	RegionRec bak = thincNullReg;
	RegionPtr reg;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: deltax %d, deltay %d\n", 
		func, deltax, deltay));
	print_reg(T_DBG_CMD, pThinc->pScreen, clip, "clipping with:");

	fill = cmd_data(cmd, struct fillCmd*);

	/* TODO: Maybe we can check beforehand if the region will be actually
	 * changed before making the copy. Or change things so that
	 * thincBufferCmd() is the one doing the changes in its own
	 * copy */
	if(!REGION_COPY(pThinc->pScreen, &bak, &fill->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create copy of "
			"FILL region\n", func));
		return;
	}

	reg = &fill->reg;
	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "fill command region:");

	REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
	REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, reg)) {
		thincDbg(T_DBG_CMD, ("%s: finished. Queued command is "
			"completely obscured\n", func));
		REGION_UNINIT(pThinc->pScreen, reg);
		copy_reg(fill->reg, bak);
		return;
	}
	/* update this, or overwrite() will be useless! */
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));

	thincBufferCmd(pThinc, pWindow, cmd);

	/* undo time */
	REGION_UNINIT(pThinc->pScreen, reg);
	copy_reg(fill->reg, bak);
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, &fill->reg)));

	thincDbg(T_DBG_CMD, ("%s: finished\n", func));
	stats_end(pThinc, t1, t2, sfillStats.exec);
}

void thincPFillExecute(thincScreenPtr pThinc, WindowPtr pWindow,
	thincCmdPtr cmd, RegionPtr clip, int deltax, int deltay)
{
	t_dbg_name(func)
	struct pixmapFillCmd *pixmap;
	unsigned char *pixdata;

	RegionRec bak = thincNullReg;	/* backup of original reg */
	RegionPtr reg;
	stats_var2(t1, t2);

	stats_start(pThinc, t1);

	thincDbg(T_DBG_CMD, ("%s: deltax %d, deltay %d\n", func, deltax,
		deltay));
	print_reg(T_DBG_CMD, pThinc->pScreen, clip, "clipping with:");

	pixmap = cmd_data(cmd, struct pixmapFillCmd*);

	thincDbg(T_DBG_CMD, ("%s: got pixmap command, have pixmap of "
		"size %d\n", func, pixmap->size));

	/* TODO: Maybe we can check beforehand if the region will be actually
	 * changed before making the copy */
	if(!REGION_COPY(pThinc->pScreen, &bak, &pixmap->reg)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create copy of "
			"PFILL region\n", func));
		return;
	}

	reg = &pixmap->reg;
	thincDbg(T_DBG_CMD, ("%s: got region. have %ld rectangles\n",
		func, REGION_NUM_RECTS(reg)));
	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "pixmap region:");

	pixdata = (unsigned char*) (pixmap+1);

	REGION_TRANSLATE(pThinc->pScreen, reg, deltax, deltay);
	REGION_INTERSECT(pThinc->pScreen, reg, clip, reg);
	if(!REGION_NOTEMPTY(pThinc->pScreen, reg)) {
		thincDbg(T_DBG_CMD, ("%s: finished. Queued command is "
			"completely obscured\n", func));
		REGION_UNINIT(pThinc->pScreen, reg);
		copy_reg(pixmap->reg, bak);
		return;
	}
	pixmap->x += deltax;
	pixmap->y += deltay;

	thincDbg(T_DBG_CMD, ("%s: after clipping. have %ld rectangles\n",
		func, REGION_NUM_RECTS(reg)));
	print_reg(T_DBG_CMD, pThinc->pScreen, reg, "after clipping:");

	/* update this, or overwrite() will be useless! */
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, reg)));
	/* send the command */
	thincBufferCmd(pThinc, pWindow, cmd);

	/* cleanup */
	REGION_UNINIT(pThinc->pScreen, reg);
	copy_reg(pixmap->reg, bak);
	pixmap->x -= deltax;
	pixmap->y -= deltay;
	copy_boxp(&cmd->area, (REGION_EXTENTS(pThinc->pScreen, &pixmap->reg)));

	thincDbg(T_DBG_CMD, ("%s; finished\n", func));
	stats_end(pThinc, t1, t2, pfillStats.exec);
}
