/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Functions for data compression
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

/* TODO: use function pointers in pThinc */

#include "thinc.h"
#ifndef T_WIN32
#include <stdarg.h>
#include <netinet/in.h>
#endif /* T_WIN32 */
/* png is default */
#ifdef T_USEPNG
static int thinc_pngPrepare(thincEncodePtr encoder);
static int thinc_pngSetParams(thincEncodePtr encoder, int params, va_list args);
static int thinc_pngCompress(thincEncodePtr encoder);
static int thinc_pngFlush(thincEncodePtr encoder);
static int thinc_pngInit(thincEncodePtr encoder);
static void thinc_pngEnd(thincEncodePtr encoder);
static void thinc_pngWriteData(png_structp png_ptr, png_bytep data, 
	png_size_t length);
#if 0
static void thinc_pngFlushData(png_structp png_ptr);
static void thinc_pngWarning(png_structp png_ptr, png_const_charp error_msg);
static void thinc_pngError(png_structp png_ptr, png_const_charp error_msg);
static void thinc_pngWriteCallback(png_structp png_ptr, png_uint_32 row,
	int pass);
#endif
#else
/* zlib */
static int thinc_zlibPrepare(thincEncodePtr encoder);
static int thinc_zlibSetParams(thincEncodePtr encoder, int params, va_list args);
static int thinc_zlibCompress(thincEncodePtr encoder);
static int thinc_zlibFlush(thincEncodePtr encoder);
static int thinc_zlibInit(thincEncodePtr encoder);
static void thinc_zlibEnd(thincEncodePtr encoder);
#endif

int thinc_compressInit(thincScreenPtr pThinc, int compress_level)
{
	thincEncodePtr encoder;
	int err;

	encoder = malloc(sizeof(thincEncodeRec));
	if(!encoder)
		return T_ERROR;
	pThinc->encoder = encoder;
	encoder->level = compress_level;
	/* FIXME: What happens if update size is smaller than video frames?? */
	encoder->zbuff_size = t_pix_to_bytes(pThinc->updateSize, pThinc);
	encoder->zbuff_size += T_Z_OVERHEAD(encoder->zbuff_size);
	encoder->zbuff = malloc(encoder->zbuff_size);
	if(!encoder->zbuff) {
		thincError("ERROR: Couldn't allocate memory for compression "
			"buffer\n");
		return FALSE;
	}

#ifndef T_USEPNG
	err = thinc_zlibInit(encoder);
#else
	err = thinc_pngInit(encoder);
#endif
	if(err == T_ERROR) {
		free(encoder->zbuff);
		encoder->zbuff_size = -1;
		free(pThinc->encoder);
		return T_ERROR;
	}
	return T_OK;
}

void thinc_compressEnd(thincScreenPtr pThinc)
{
	thincEncodePtr encoder = pThinc->encoder;
#ifndef T_USEPNG
	thinc_zlibEnd(encoder);
#else
	thinc_pngEnd(encoder);
#endif
	free(encoder->zbuff);
	encoder->zbuff_size = -1;
	free(pThinc->encoder);
	pThinc->encoder = NULL;
}

int thinc_compressMaxSize(thincScreenPtr pThinc)
{
	return pThinc->encoder->zbuff_size;
}

unsigned char * thinc_compressGetBuffer(thincScreenPtr pThinc)
{
	return pThinc->encoder->zbuff;
}

/* similar to deflateReset() the idea is this function will be called when
 * one needs to make sure the compression engine is ready to be used */
int thinc_compressPrepare(thincScreenPtr pThinc)
{
#ifndef T_USEPNG
	int err = thinc_zlibPrepare(pThinc->encoder);
#else
	int err = thinc_pngPrepare(pThinc->encoder);
#endif
	return err;
}

/* 
 * if more than one type of params, assumed order is:
 * T_ZPARAM_IN
 * T_ZPARAM_OUT
 */
int thinc_compressSetParams(thincScreenPtr pThinc, int params, ...)
{
	va_list args;
	int err;
	va_start(args, params);
#ifndef T_USEPNG
	err = thinc_zlibSetParams(pThinc->encoder, params, args);
#else
	err = thinc_pngSetParams(pThinc->encoder, params, args);
#endif
	va_end(args);
	return err;
}

int thinc_compress(thincScreenPtr pThinc)
{
#ifndef T_USEPNG
	return thinc_zlibCompress(pThinc->encoder);
#else
	return thinc_pngCompress(pThinc->encoder);
#endif
}

/* Put all compressed data in return amount of data compressed or T_ERROR on error */
int thinc_compressFlush(thincScreenPtr pThinc)
{
#ifndef T_USEPNG
	return thinc_zlibFlush(pThinc->encoder);
#else
	return thinc_pngFlush(pThinc->encoder);
#endif
}

/**** ZLIB FUNCTIONS ****/
#ifndef T_USEPNG
static int thinc_zlibInit(thincEncodePtr encoder)
{
	t_dbg_name(func)
	z_stream *zstream;
	int err;

	encoder->zstream = malloc(sizeof(z_stream));
	if(!encoder->zstream) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for zlib stream object\n", func));
		return T_ERROR;
	}

	zstream = encoder->zstream;
	zstream->next_in = NULL;
	zstream->avail_in = 0;
	zstream->next_out = NULL;
	zstream->avail_out = 0;
	zstream->zalloc = (alloc_func)NULL;
	zstream->zfree = (free_func)NULL;
	zstream->opaque = (voidpf)NULL;

#if 0	/* zlevel set by config option */
	encoder->level = T_ZCOMPRESS_LEVEL;
	err = deflateInit(zstream, encoder->level);
#else
	/* possible values:
	 * Z_BEST_COMPRESSION
	 * Z_BEST_SPEED
	 * Z_NO_COMPRESSION
	 * Z_DEFAULT_COMPRESSION
	 */
#if 0	/* level set by config option */
	encoder->level = Z_BEST_SPEED;
#endif
	err = deflateInit2(zstream, encoder->level, Z_DEFLATED, 15, 9,
		Z_FILTERED);
#endif
	if(err != Z_OK) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initializate "
			"deflate: %d\n", func, err));
		return T_ERROR;
	}

	return T_OK;
}

static void thinc_zlibEnd(thincEncodePtr encoder)
{
	deflateEnd(encoder->zstream);
	free(encoder->zstream);
	encoder->zstream = NULL;
}

static int thinc_zlibPrepare(thincEncodePtr encoder)
{
	z_stream *zstream;
	int err;

	zstream = encoder->zstream;
	err = deflateReset(zstream);
	if(err != Z_OK) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't reset zlib: %d\n", 
			__func__, err));
		return T_ERROR;
	} else
		return T_OK;
}

static int thinc_zlibSetParams(thincEncodePtr encoder, int params, va_list args)
{
	t_dbg_name(func)
	void *ptr;
	unsigned long avail;
	z_stream *zstream = encoder->zstream;

	thincDbg(T_DBG_COMP, ("%s: setting compression parameters: %d\n",
		func, params));

	if(params & T_ZPARAM_IN) {
		ptr = va_arg(args, void*);
		zstream->next_in = ptr;
		avail = va_arg(args, unsigned long);
		zstream->avail_in = avail;
	}

	if(params & T_ZPARAM_OUT) {
		ptr = va_arg(args, void*);
		zstream->next_out = ptr;
		avail = va_arg(args, unsigned long);
		zstream->avail_out = avail;
	}

	thincDbg(T_DBG_COMP, ("%s: finished\n", func));
	return T_OK;
}

static int thinc_zlibCompress(thincEncodePtr encoder)
{
	t_dbg_name(func)
	int err;
	z_stream* zstream = encoder->zstream;

	if(!zstream->next_out || !zstream->next_in) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: buffers haven't been "
			"set. Use SetParams() before calling this function\n", 
			func));
		return T_ERROR;
	}

	err = deflate(zstream, 0);
	if(err != Z_OK) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't compress: %d\n",
			__func__, err));
		return T_ERROR;
	} else
		return T_OK;
}

static int thinc_zlibFlush(thincEncodePtr encoder)
{
	t_dbg_name(func)
	int err;
	z_stream* zstream = encoder->zstream;

	if(!zstream->next_out) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: output buffer hasn't been "
			"set. Use SetParams() before calling this function\n", 
			func));
		return T_ERROR;
	}
	err = deflate(zstream, Z_FINISH);
	if(err != Z_STREAM_END) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't compress: %d\n",
			__func__, err));
		return T_ERROR;
	} else 
		return zstream->total_out;
}

#else
/**** PNG FUNCTIONS ****/
/* FIXME: can we make the png structs persistent across compressions? */
static int thinc_pngInit(thincEncodePtr encoder)
{
	return T_OK;
}

static void thinc_pngEnd(thincEncodePtr encoder)
{
	return;
}

static int thinc_pngPrepare(thincEncodePtr encoder)
{
	png_structp png_ptr;

	instr_add("PNG_PREPARE", TI_START);

	/* FIXME: Add error handling. PNG's error handling really sucks.
	 * why can't they just return -1 on error, 0 on success?? :( */
#if 0
	encoder->png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
		(void*)encoder, thinc_pngError, thinc_pngWarning);
#else
	encoder->png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING,
		NULL, NULL, NULL);
#endif
	png_ptr = encoder->png_ptr;
	encoder->info_ptr = png_create_info_struct(png_ptr);

#if 0
	png_set_write_fn(png_ptr, (void*)encoder, thinc_pngWriteData,
		thinc_pngFlushData);
	png_set_write_status_fn(png_ptr, thinc_pngWriteCallback);
#else
	png_set_write_fn(png_ptr, (void*)encoder, thinc_pngWriteData,
		NULL);
#endif

	/* TODO: Set more zlib parameters? */
#if 1
	png_set_compression_level(png_ptr, encoder->level);
#else
	png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
	png_set_compression_strategy(png_ptr, Z_FILTERED);
	png_set_compression_mem_level(png_ptr, 9);
	png_set_compression_window_bits(png_ptr, 15);
#endif

	instr_add("PNG_PREPARE", TI_END);
	return T_OK;
}

/* IN has different meaning here. 
 * Call once to set w & h (and potentially first row).
 * Call repeatedly to set each row pointer */
static int thinc_pngSetParams(thincEncodePtr encoder, int params, va_list args)
{
	t_dbg_name(func)
	unsigned int w, h;
	png_structp png_ptr = encoder->png_ptr;
	png_infop info_ptr = encoder->info_ptr;

	instr_add("PNG_SETPARAMS", TI_START);

	thincDbg(T_DBG_COMP, ("%s: params: %d\n", func, params));

	if((params & T_ZPARAM_W) && (params & T_ZPARAM_H)) {
		w = va_arg(args, unsigned int);
		h = va_arg(args, unsigned int);

		/* set up filtering. possible filter values:
		 * PNG_FILTER_NONE 
		 * PNG_FILTER_SUB
		 * PNG_FILTER_UP 
		 * PNG_FILTER_AVG 
		 * PNG_FILTER_PAETH
		 * PNG_ALL_FILTERS
		 */
		/* sub is the simplest one (thus fastest) and the hit on
		 * data size is not that bad (~ 3%). none would be
		 * simpler but it has a much bigger impact on data size
		 * which hurts things, in particular on the wan. */
		png_set_filter(png_ptr, PNG_FILTER_TYPE_BASE, 
			PNG_FILTER_SUB);

		/* XXX: Do just RGB for now (instead of RGBA) */
#if 0
		png_set_IHDR(png_ptr, info_ptr, w, h, 8,
			PNG_COLOR_TYPE_RGBA, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
#else
		png_set_IHDR(png_ptr, info_ptr, w, h, 8,
			PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
			PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
#endif

		thincDbg(T_DBG_COMP, ("%s: compressing image of width %d and "
			"height %d\n", func, w, h));

		encoder->curr_size = 0;
		encoder->zcurr = encoder->zbuff;

		png_write_info(png_ptr, info_ptr);

		thincDbg(T_DBG_COMP, ("%s: wrote png info, setting up the "
			"rest of the parameters\n", func));

		/* PNG files store RGB pixels packed into 3 or 6 bytes.
		 * This code tells the library to strip input data that
		 * has 4 or 8 bytes per pixel down to 3 or 6 bytes (or
		 * strip 2 or 4-byte grayscale+filler data to 1 or 2
		 * bytes per pixel).
		 * 
		 * the 0 is unused, and the location is either
		 * PNG_FILLER_BEFORE or PNG_FILLER_AFTER, depending
		 * upon whether the filler byte in the pixel is
		 * stored XRGB or RGBX.
		 * FIXME: BEFORE or AFTER may depend on byte order of the
		 * machine??
		 */
		png_set_filler(png_ptr, 0, PNG_FILLER_AFTER);
	}

	if(params & T_ZPARAM_IN) {
		encoder->data = va_arg(args, unsigned char*);
	}

	thincDbg(T_DBG_COMP, ("%s: finished\n", func));

	instr_add("PNG_SETPARAMS", TI_END);
	return T_OK;
}

static int thinc_pngCompress(thincEncodePtr encoder)
{
	t_dbg_name(func)
	png_structp png_ptr = encoder->png_ptr;
	png_bytep row = encoder->data;

	instr_add("PNG_COMPRESS", TI_START);

	thincDbg(T_DBG_COMP, ("%s: compressing one more line\n", func));
	png_write_row(png_ptr, row);

	thincDbg(T_DBG_COMP, ("%s: done\n", func));
	instr_add("PNG_COMPRESS", TI_END);
	return T_OK;
}

/* FIXME: is it ok to use this as our end function for this session? */
static int thinc_pngFlush(thincEncodePtr encoder)
{
	png_structp png_ptr = encoder->png_ptr;
	png_infop info_ptr = encoder->info_ptr;

	instr_add("PNG_FLUSH", TI_START);

	png_write_end(png_ptr, NULL);
	png_destroy_write_struct(&png_ptr, &info_ptr);

	encoder->png_ptr = NULL;
	encoder->info_ptr = NULL;

	instr_add("PNG_FLUSH", TI_END);

	return encoder->curr_size;
}

static void thinc_pngWriteData(png_structp png_ptr, png_bytep data, 
	png_size_t length)
{
	t_dbg_name(func)
	thincEncodePtr encoder = (thincEncodePtr)png_get_io_ptr(png_ptr);

	thincDbg(T_DBG_COMP, ("%s: writing %d bytes\n", func, length));

	/* FIXME: Can this happen?? */
	if(encoder->curr_size + length > encoder->zbuff_size) {
		thincError("TOO MUCH PNG DATA... AGGGGGHHHHH!! Will have "
			"%d bytes, but only have space for %d\n",
			encoder->curr_size + length, encoder->zbuff_size);
		return;
	}

	memcpy(encoder->zcurr, data, length);
	encoder->curr_size += length;
	encoder->zcurr += length;
	thincDbg(T_DBG_COMP, ("%s: finished\n", func));
}

#if 0
static void thinc_pngFlushData(png_structp png_ptr)
{
	t_dbg_name(func)
	thincEncodePtr encoder = (thincEncodePtr)png_get_io_ptr(png_ptr);
	thincDbg(T_DBG_COMP, ("%s: curr_size %d\n", func, encoder->curr_size));
}

static void thinc_pngError(png_structp png_ptr, png_const_charp error_msg)
{
	thincError("%s: %s\n", __func__, error_msg);
}

static void thinc_pngWarning(png_structp png_ptr, png_const_charp warn_msg)
{
	thincError("%s: %s\n", __func__, warn_msg);
}

static void thinc_pngWriteCallback(png_structp png_ptr, png_uint_32 row,
	int pass)
{
	thincDbg(T_DBG_COMP, ("%s: row %lu, pass %d\n", __func__, row, pass));
}
#endif

#endif /* T_USEPNG */
