/* 
 *
 * Resize protocol for THINC
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

#ifndef _THINC_PROTO_RESIZE_H_
#define _THINC_PROTO_RESIZE_H_

/* how it works (see server-dir/doc/zoom for more details):
 *
 * 1. when client's window changes size (order of operation only used to
 * facilitate understanding of the process): 
 * => client sends a RESIZE EVENT to server with new dimensions (w1,h1).  
 * => server sends acknowledgement with new dimensions.  
 * => client resizes its local fb (should keep a copy of original size fb):
 * if new size is smaller than current "resize", just use resized copy to
 * display on the screen.  if new size is larger than current "resize", then go
 * back to original fb and resize. If _outdate-region_ is non-empty, then the
 * client sends a request to the server to update the contents of the region.
 * Until the server's ack is received, the client should discard all updates
 * which are flagged as RESIZED (since they may correspond to the old size).  
 * What to do after the ack but before the updates arrive is currently an
 * open question. Current implementation is as follows: 
 * ++ use original fb to fill in up-to-date regions.
 * ++ use current resized fb to fill in out-of-date regions, by enlarging them.
 * ++ as updates arrive, redraw the fb. Note that the updates can be sent by
 * the server resized, since they're no-different than normal draw commands.
 *
 * => special cases:
 * + resize to original size: client should send request for update of outdated
 * region. No future updates will have RESIZED flag set.
 * + resize to size larger than original: client is in charge of all resizing,
 * server can safely ignore this state.
 * 
 * 2. when draw request comes in to server: 
 * => if non-RAW update:
 * 	* send as always, unresized.  client draws to original fb, then resizes
 * 	affected area and copies to resized fb.  
 * => if RAW update:
 * 	* resize, and send to client. flag message with T_FB_RESIZED. Data is
 * 	wrapped with t_fb_resized message. Note that T_FB_RUPDATE message has 
 * 	unresized coordinates.
 * 	* client copies data to resized fb.  client adds unresized rectangle to
 * 	its outdated region.
 */

/* 
 * as it stands right now:
 * window resized => resz_event
 * if need updates => resz_update ... <= resz_ack from server
 *                                ... <= updates (as RAW updates)
 */

/*************************************************************************
 *                          CLIENT => SERVER                             *
 *************************************************************************/
#define T_RESZ_EVENT	50	/* thinc_resEvent */
#define T_RESZ_UPDATE	51	/* thinc_resUpdate */

/* T_RESZ_EVENT
 * Description: Tells the server that the client's destination window
 * has been resized to embedded dimensions
 * Size: 4 bytes
 * Data: None
 */
typedef struct thinc_reszEvent_s {
	int width	:16;
	int height	:16;
} thinc_reszEvent;

#define T_RESZ_EVENT_SIZE	4

/* T_RESZ_UPDATE
 * Description: Asks the server to send updated, unresized framebuffer 
 * contents for the passed rectangles. Server should reply with ack with
 * what it thinks are the client's current coordinates, and then send
 * the updates. The coordinates are used by the client in case the
 * server is replying to an old update request. The server only needs
 * to reply to the latest update request.
 * Size: 4 bytes
 * Data: Rectangles (thinc_rectRegion) to update
 */
typedef struct thinc_reszUpdate_s {
	int numrects	:32;	/* big in case we need to fit more things
				   in the future */
} thinc_reszUpdate;

#define T_RESZ_UPDATE_SIZE	4

/*************************************************************************
 *                          SERVER => CLIENT                             *
 *************************************************************************/
#define T_RESZ_ACK	50	/* thinc_resAck */

/* T_RESZ_ACK
 * Description: Tells the client that it received its update request
 * (T_RESZ_UPDATE) and it's preparing the updates for the regions it
 * requested
 * Size: 4 bytes
 * Data: None
 */
typedef struct thinc_reszAck_s {
	int width	:16;
	int height	:16;
} thinc_reszAck;

#define T_RESZ_ACK_SIZE		4

/* T_FB_RESIZED
 * Description: Wrapper for a resized RAW update. It contains what the server
 * thinks is the client's current size (which is the sizes it used to
 * resize the update) and how much data is following it. If the message
 * is also compressed, this message wraps the compressed header.
 * Size: 8 bytes
 * Data: The resized RAW data
 */
typedef struct thinc_reszData_s {
	int fb_width	:16;	/* these two work as id, so the client can */
	int fb_height	:16;	/* check that this update corresponds to its
				   actual size. */
	int size	:32;	/* how much data is being sent */
} thinc_reszData;

#define T_RESZ_DATA_SIZE	8

#endif
