/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Functions for RAW message
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

/*
 * Flush a RAW command:
 * Because RAW commands store multiple rectangles, but RAW updates only
 * support one rectangle, we need to break things up before the message
 * stack can be executed. The disadvantage here is that some processing
 * is done that's not accounted for in the statistics (I don't want to break
 * up statistics collecting to multiple places)
 * The way it works:
 * -> RawFlushReal: extracts information from RAW command. Then for each 
 *  rectangle, it checks if its size is appropriately "small" 
 *  (<= pThinc->updateSize). If yes, it calls RawFlushSimple(). If the 
 *  rectangle is too big it calls RawFlushMultiple().
 * -> RawFlushMultiple: breaks up the rectangle in rows and calls RawFlushRow.
 * -> RawFlushRow: breaks up a row into appropriately sized rectangles, each
 *  one being passed to RawFlushSimple(). After each rectangle is sent,
 *  checks are mode to verify if flush should stop or can keep going
 *  (see scheduler implementation).
 * -> RawFlushSimple: Creates a RAW object and passes it to the message stack.
 */

#include "thinc.h"
#ifdef T_WIN32
#else
#include <netinet/in.h>
#include <sys/poll.h>
#include "mi.h"

/* needed for WriteFile module */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif /* T_WIN32 */

/* if there's less than this space in the client's buffer, we flush it
 * before attempting to process our updates */
#define T_MINBUFSPACE		24

/* if update smaller than this, don't bother compressing it */
#define T_ZMINSIZE		64

/* updates larger than this are not cached (256x256x4). */
/* FIXME: This would be great if it wasn't for the fact that fucking
 * mozilla keeps drawing everything twice... */
#if 0
#define T_RCACHE_MAX_SZ		3145728
#endif

/* write each update to a file in /tmp/raw */
#if 0
#define T_RAW_TO_FILE
#endif

/* the methods */
#ifdef T_GETSTATS
static int thinc_rawPreStats(thincClientPtr, void *);
static int thinc_rawPostStats(thincClientPtr, void *);
#endif
static int thinc_rawSend(thincClientPtr cl, void *obj);
static int thinc_rawCompress(thincClientPtr cl, void *obj);
static int thinc_rawResize(thincClientPtr cl, void *msg);
#ifndef T_WIN32
static int thinc_rawWriteFile(thincClientPtr cl, void *msg);
#endif
static int thinc_rawCopyData(thincClientPtr cl, void *msg);
static int thinc_rawChkCache(thincClientPtr cl, void *msg);
static int thinc_rawAddCache(thincClientPtr cl, void *msg);
static int thinc_rawLogSize(thincClientPtr cl, void *msg);

/* the priorities. current order:
 * PRESTATS
 * RESIZE
 * COPYDATA (optional, but required by CHKCACHE and ADDCACHE) 
 * CHKCACHE (we're caching resized updates)
 * COMPRESS
 * SEND
 * ADDCACHE
 * POSTSTATS
 */
#ifdef T_GETSTATS
#define T_RAWPRIO_PRESTATS	T_PRIO_MIN
#define T_RAWPRIO_POSTSTATS	T_PRIO_MAX
#endif
#define T_RAWPRIO_SEND		T_PRIO_SEND
#define T_RAWPRIO_COMPRESS	T_RAWPRIO_SEND-10
#define T_RAWPRIO_RESIZE	T_RAWPRIO_SEND-100
#define T_RAWPRIO_COPYDATA	T_RAWPRIO_SEND-60 /* needed by chk cache */
#define T_RAWPRIO_CHKCACHE	T_RAWPRIO_SEND-50
#define T_RAWPRIO_ADDCACHE	T_RAWPRIO_SEND+10
/* change to modify where in the pipeline you want to save the data */
#if 0
#define T_RAWPRIO_WRITEFILE	T_RAWPRIO_CHKCACHE-1
#else
#define T_RAWPRIO_WRITEFILE	T_RAWPRIO_CHKCACHE+1
#endif
#define T_RAWPRIO_LOGSIZE	T_RAWPRIO_SEND-2

/* the static modules. From here, we make copies to each client's stack */
#ifdef T_GETSTATS
thincDefModule(rawPreStats, "RAW_PRESTATS", 13, T_RAWPRIO_PRESTATS);
thincDefModule(rawPostStats, "RAW_POSTSTATS", 14, T_RAWPRIO_POSTSTATS);
#endif
thincDefModule(rawSend, "RAW_SEND", 9, T_RAWPRIO_SEND);
thincDefModule(rawCompress, "RAW_COMPRESS", 13, T_RAWPRIO_COMPRESS);
thincDefModule(rawResize, "RAW_RESIZE", 11, T_RAWPRIO_RESIZE);
#if 0
thincDefModule(rawWriteFile, "RAW_WRITEFILE", 13, T_RAWPRIO_WRITEFILE);
#endif
thincDefModule(rawCopyData, "RAW_COPYDATA", 13, T_RAWPRIO_COPYDATA);
thincDefModule(rawChkCache, "RAW_CHKCACHE", 13, T_RAWPRIO_CHKCACHE);
thincDefModule(rawAddCache, "RAW_ADDCACHE", 13, T_RAWPRIO_ADDCACHE);
thincDefModule(rawLogSize, "RAW_LOGSIZE", 12, T_RAWPRIO_LOGSIZE);

/* local variables */
#ifdef T_GETSTATS
static unsigned long t_start;	/* needed for statistics */
#endif
static unsigned char *local_buf=NULL; /* used by copyData */

#define T_RAW_EXTRALEN	32

/* NOTES ON RAW OBJECT
 * Modules cannot set the data pointer to memory to which there are no other 
 * references, since modules will blindly replace that pointer as needed. DON'T
 * ASSUME THAT YOU CAN CREATE A POST_* MODULE AND FREE THE POINTER THERE.
 * You may end up freeing another pointer and cause a nuclear meltdown!
 */
typedef struct thincRawObj_S {
	unsigned int x:16;
	unsigned int y:16;
	unsigned int width:16;
	unsigned int height:16;
	/* XXX HACK ALERT XXX */
	/* for resize + png we need to know the real width and height of the
	 * update, but we cannot overwrite width and height because our
	 * stupid protocol sends unresized dimensions even when resizing
	 * the update... i think i was on really bad drugs when i wrote the
	 * resize protocol :( */
	unsigned int real_width:16;
	unsigned int real_height:16;

	/* information for write and stats. Basically how much total data
	 * was sent as a result of this raw object */
	unsigned int total_len:32;

	/* the raw data, may be NULL */
	unsigned char *data;
	unsigned int data_len:32;
	unsigned int data_width:32; /* width of one line of data (bytes) */
	unsigned int data_pitch:32; /* distance between lines */
	unsigned int data_height:16; /* how many lines of data are there */

	/* extra headers for the message */
	unsigned int flags:8;		/* message flags */
	unsigned int extra_len:8;	/* current length <= T_RAW_EXTRALEN */
	unsigned char *extra;		/* current end */
	unsigned char extra_data[T_RAW_EXTRALEN];

	/* cache information */
	unsigned int hash:32;
} thincRawObj;

/**** local functions ****/
/* FLUSH */
static int thincRawFlushReal(thincScreenPtr pThinc, thincClientPtr cl,
	thincCmdPtr cmd);
static int thincRawFlushMultiple(thincScreenPtr pThinc, thincClientPtr cl,
	BoxPtr rect, unsigned char *data, int pitch, RegionPtr* left);
static int thincRawFlushRow(thincScreenPtr pThinc, thincClientPtr cl,
	const BoxPtr row, BoxPtr notdone, unsigned char *data,
	int pitch);
static int thincRawFlushMultiple(thincScreenPtr pThinc, thincClientPtr cl,
	BoxPtr rect, unsigned char *data, int pitch, RegionPtr* left);
static int thincRawFlushSimple(thincScreenPtr pThinc, thincClientPtr cl,
	BoxPtr rect, unsigned char *data, int pitch);
/* helper functions to send module */
static int thinc_rawWriteUpdate(thincClientPtr cl, thincRawObj *obj);
static int thinc_rawWriteUpdateData(thincClientPtr cl, unsigned char *data,
	unsigned int width, unsigned int height, unsigned int pitch);

/* checks that we haven't overflowed the extra data buffer. Shouldn't happen
 * as long as extra messages and the buffer size are kept in sync. Should
 * be called every time something is added to the extra buffer */
#define checkExtraBuffer(obj) \
	if((obj)->extra_len > T_RAW_EXTRALEN) { \
		thincError("BUG: %s: %d: %s(): RAW update extra headers " \
			"overflowed static data buffer in RAW message " \
			"object! Current lenght is %u but buffer's size is " \
			"%u\n", __FILE__, __LINE__, modname, obj->extra_len, \
			T_RAW_EXTRALEN); \
		return T_ERROR; \
	}

#if 0
#define raw_addmodule(mod) { \
	err = thinc_msgStackAddModule(cl, cl->rawStack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add RAW module "\
			"%s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}
#else
#define raw_addmodule(cl, mod) { \
	err = thinc_msgAddUsingModule((cl), (cl)->rawStack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add RAW module "\
			"%s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}
#endif

/* initializes and adds appropriate modules to the RAW message stack */
int thinc_rawInitStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;
	thincScreenPtr pThinc = cl->parent;
	
	thincDbg(T_DBG_CLIENT, ("%s: creating RAW message stack\n", func));
	
	cl->rawStack = thinc_msgStackCreate(cl);
	if(!cl->rawStack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create RAW "
			"message stack\n", func));
		return T_ERROR;
	}

	/* add the modules */
#ifdef T_GETSTATS
	raw_addmodule(cl, &rawPreStats);
	raw_addmodule(cl, &rawPostStats);
#endif

	raw_addmodule(cl, &rawSend);
	if(pThinc->compress)
		raw_addmodule(cl, &rawCompress);

	if(pThinc->useRawCache) {
		/* cache needs the data contiguously */
		/* FIXME: This would be great if it wasn't for the fact 
		 * that fucking mozilla keeps drawing everything twice... 
		 * So for now, let's use the framebuffer size as the max
		 * size for our buffer. Should help things a little bit.
		 */
#if 0
		local_buf = malloc(T_RCACHE_MAX_SZ);
		if(!local_buf) {
			thincDbg(T_DBG_PROTO, ("ERROR: %s: Couldn't allocate "
				"memory for data copy\n", func));
			return T_ERROR;
		}
#else
		if(!local_buf) {
			local_buf = malloc(pThinc->widthBytes * 
				pThinc->height);
			if(!local_buf) {
				thincDbg(T_DBG_PROTO, ("ERROR: %s: Couldn't "
					"allocate memory for data "
					"copy\n", func)); 
				return T_ERROR; 
			}
		}
#endif 
		raw_addmodule(cl, &rawCopyData);
		raw_addmodule(cl, &rawChkCache);
		raw_addmodule(cl, &rawAddCache);
	}

	/* only for debugging */
#ifdef T_RAW_TO_FILE
	raw_addmodule(cl, &rawWriteFile);
#endif	
#if 0
	raw_addmodule(cl, &rawLogSize);
#endif

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

int thinc_rawStackClResized(thincClientPtr cl, int prevstate)
{
	t_dbg_name(func)
	int err;

	if(!cl->parent->allowResize)
		return T_OK;

	if(cl->resized == T_CL_RESZ_LESS) {
		if(prevstate != T_CL_RESZ_LESS)
			raw_addmodule(cl, &rawResize);
	} else {
		if(prevstate == T_CL_RESZ_LESS) {
			thinc_msgDelUsingModule(cl, cl->rawStack, &rawResize);
		}
	}
	return T_OK;
}

void thinc_rawDestroyStack(thincClientPtr cl)
{
	thinc_msgStackDestroy(cl, cl->rawStack);
	cl->rawStack = NULL;
}

/********** Flush a RAW command *************/
int thincRawFlush(thincScreenPtr pThinc, thincClientPtr cl, thincCmdPtr cmd)
{
#ifdef T_WIN32
	int res;
	BlockDriverDrawing();
	res = thincRawFlushReal(pThinc, cl, cmd);
	UnblockDriverDrawing();
	return res;
#else
	return thincRawFlushReal(pThinc, cl, cmd);
#endif
}

/* TODO:
 * Instead of passing data NULL and comparing it ten thousand times we should
 * just set the data pointer point to the framebuffer, and set pitch and friends
 * appropriately. For some reason, this doesn't currently work.
 */
static int thincRawFlushReal(thincScreenPtr pThinc, thincClientPtr cl,
	thincCmdPtr cmd)
{
	t_dbg_name(func)
	struct rawCmd *raw;
	BoxPtr rects;
	RegionPtr reg, left=NULL;
	int numrects, i, err = T_OK, rect_size, size;
	unsigned char *curr = NULL;
	/* T_PART handling vars */
	int remain, sent, hasdata = T_FALSE, pitch = 0;
	RegDataPtr data;
	BoxPtr first, orig = NULL;

	instr_add("RAW", TI_START);

	thincDbg(T_DBG_PROTO, ("%s: Flushing uncompressed RAW update for "
		"command %#lx\n", func, cmd->id));

#if 0
	cmd = thincRawModify(pThinc, cmd, T_CMD_SAVE, NULL);
#endif

	raw = cmd_data(cmd, struct rawCmd*);
	reg = &raw->reg;

	if(raw->datasize) {
		hasdata = T_TRUE;
		orig = &raw->orig;
	}
	
	numrects = REGION_NUM_RECTS(reg);
	rects = REGION_RECTS(reg);

#if 0
	thincError("orig box: %u,%u=>%u,%u\n", raw->orig.x1, 
		raw->orig.y1, raw->orig.x2, raw->orig.y2);
#endif

	thincDbg(T_DBG_PROTO, ("%s: have %d rectangles to send\n", func,
		numrects));

	for(i=0; i<numrects; ++i, ++rects) {
		thincDbg(T_DBG_PROTO, ("%s: rectangle %d (%d, %d) (%d, %d)\n",
			func, i, rects->x1, rects->y1, rects->x2, 
			rects->y2));

		/* FIXME: this won't work if updates are too big, and bypass 
		 * the buffer completely. Check if buffer size is 0 (then we
		 * know it was just flushed) */
		if(T_CL_MAXMSGDATA(cl, T_FB_RUPDATE) < T_MINBUFSPACE) {
			thincFlushBuffer(cl);
			/* can we write more? */
			err = thincDoPoll(cl->sock, POLLOUT, 0);
			if(err == T_ERROR) {
				thincDbg(T_DBG_PROTO, ("ERROR: %s: Couldn't "
					"check if would block\n", func));
				return T_ERROR;
			} else if(err == T_TIMEOUT) {
				/* can't write to this client now */
				thincDbg(T_DBG_PROTO, ("%s: Client is busy, "
					"must stop writing\n", func));
				err = T_OK;
				break;
			}
		}
		if(hasdata) {
			/* rect_size -> unused (it's actually rect_width)
			 * pitch we already have */
			pitch = box_width_bytes(pThinc, orig);
			t_get_datapointer(raw->data->data, orig,
					pitch, pThinc->Bpp, rects, curr);
		}
		size = box_width(rects)*box_height(rects);
		/* TODO: Need better heuristic. In good net, there's
		 * no reason to divide commands */
		if(size <= pThinc->updateSize)
			err = thincRawFlushSimple(pThinc, cl, rects, curr, 
				pitch);
		else
			err = thincRawFlushMultiple(pThinc, cl, rects, curr, 
				pitch, &left);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"RAW update\n", func));
			return T_ERROR;
		} else if(err == T_PART) {
			/* can only come from Multiple() */
			thincDbg(T_DBG_PROTO, ("%s: couldn't flush rectangle "
				"completely. Bailing out\n", func));
			break;
		}
		/* XXX: Not needed?? */
#if 1
		if(curr) {
			rect_size = box_size(pThinc, rects);
			curr += rect_size;
		}
#endif
	}

	if(i == numrects) {
		thincDbg(T_DBG_PROTO, ("%s: finished. Wrote command "
			"completely\n", func));
		instr_add("RAW", TI_END);
		return T_DONE;
	}

	/*
	 * TODO: What happens if command has data saved? We know where we
	 * are on the data stream regarding the rectangle. Need
	 * the underlying functions to tell us where they got to inside
	 * the rectangle they couldn't send.
	 */
	/* if err is T_PART, then we started sending the ith 
	 * rectangle and it needs to be taken out. Note that we
	 * may bail out before trying to send the ith rectangle
	 * (in which case err will be set to T_OK)
	 */
	remain = ( (err == T_PART) ? numrects - i -1: numrects -i);
	if(remain == numrects) {
		/* nice, no rectangles touched, get out of here */
		thincDbg(T_DBG_PROTO, ("%s: finished. Couldn't send any "
			"rectangles\n", func));
		instr_add("RAW", TI_END);
		return T_PART;
	}
	thincDbg(T_DBG_PROTO, ("%s: have %d rectangles remaining, err is "
		"%d\n", func, remain, err));
	if(remain == 0) {
		/* last rectangle in region was sent partially. just 
		 * overwrite with what we couldn't send and get out */
		REGION_COPY(pThinc->pScreen, reg, left);
		REGION_DESTROY(pThinc->pScreen, left);
		thincDbg(T_DBG_PROTO, ("%s: sent all rectangles except "
			"the last one, which was sent partially. finished\n",
			func));
		instr_add("RAW", TI_END);
		return T_PART;
	}

	/* the fun begins, don't hate me..
	 * scenario:
	 * region had more than 1 rectangles, say n. we couldn't sent one of 
	 * the first n-1 rectangles (nth rectangle is catched before we get 
	 * here)
	 * need to:
	 * => delete sent rectangles (0-remain)
	 * => move the rest to the front
	 */
	first = REGION_BOXPTR(reg);
	sent = numrects - remain;
	rects = first + sent;
	memmove(first, rects, remain*sizeof(BoxRec));
	reg->data = xrealloc(reg->data, REGION_SZOF(remain));
	if(!reg->data) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't fix "
			"partially sent RAW command\n", func));
		return T_ERROR;
	}
	data = reg->data;
	data->numRects = remain;
	data->size = remain;
	miSetExtents(reg);
	if(err == T_PART) {
		Bool blah;
		REGION_APPEND(pThinc->pScreen, reg, left);
		/* we don't even need this because we know that just
		 * by appending the rects will be YX-banded, but just in
		 * case. */
		REGION_VALIDATE(pThinc->pScreen, reg, &blah);
		REGION_DESTROY(pThinc->pScreen, left);
	}

	copy_boxp(&cmd->area, REGION_EXTENTS(pThinc->pScreen, reg));
	
	thincDbg(T_DBG_PROTO, ("%s: finished. Command was sent "
		"partially. There are %ld rectangles left to send\n", func,
		REGION_NUM_RECTS(&raw->reg)));
	instr_add("RAW", TI_END);
	return T_PART;
}

/*
 * how it works:
 * break up rect in rects of size <= max size
 * between each rect, check if can keep on writing. if not return. Region
 * left will have rectangles which couldn't be sent. Caller should take
 * out original rectangle and add the region in its place.
 */
static int thincRawFlushMultiple(thincScreenPtr pThinc, thincClientPtr cl,
	BoxPtr rect, unsigned char *data, int pitch, RegionPtr* left)
{
	t_dbg_name(func)
	int size, err;
	RegionPtr r;
	BoxRec row;
	xRectangle ret[2];
	BoxRec 	remain, /* complete rows which haven't been sent */ 
		curr; /* what couldn't be sent of current row */
	unsigned char *dptr;

	thincDbg(T_DBG_PROTO, ("%s\n", func));

	remain.x1 = row.x1 = rect->x1;
	remain.x2 = row.x2 = rect->x2;
	remain.y1 = row.y1 = rect->y1;
	row.y2 = rect->y1 + pThinc->updateHeight;
	remain.y2 = rect->y2;

	dptr = data;
	if(dptr)
		size = pThinc->updateHeight*pitch;
	else
		size = 0;

	while(row.y2 <= rect->y2) {
		remain.y1 += pThinc->updateHeight;	/* one less row */
		err = thincRawFlushRow(pThinc, cl, &row, &curr, dptr, pitch);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"rectangle's row %d->%d\n", func, row.y1, 
				row.y2));
			return T_ERROR;
		} else if(err == T_PART) {
			thincDbg(T_DBG_PROTO, ("%s: couldn't write "
				"complete update\n",func));
			print_box(T_DBG_PROTO, &curr, "curr [notdone]: ");
			print_box(T_DBG_PROTO, &remain, "remain: ");
			box_to_xrect_zerook(&curr, ret[0]);
			box_to_xrect_zerook(&remain, ret[1]);
			r = RECTS_TO_REGION(pThinc->pScreen, 2,
				ret, CT_YXBANDED);
			*left = r;
			return T_PART;
		}
		dptr += size;
		row.y1 = row.y2; /* next row */
		row.y2 += pThinc->updateHeight;
	}

	if(row.y1 != rect->y2) {
		remain.y1 = remain.y2; /* no complete rows left to send */
		row.y2 = rect->y2;
		err = thincRawFlushRow(pThinc, cl, &row, &curr, dptr, pitch);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"rectangle's row %d->%d\n", func, row.y1, 
				row.y2));
			return T_ERROR;
		} else if(err == T_PART) {
			thincDbg(T_DBG_PROTO, ("%s: couldn't write "
				"complete update\n",func));
			box_to_xrect_zerook(&curr, ret[0]);
			box_to_xrect_zerook(&remain, ret[1]);
			r = RECTS_TO_REGION(pThinc->pScreen, 2,
				ret, CT_YXBANDED);
			*left = r;
			return T_PART;
		}
	}

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;
}

static int thincRawFlushRow(thincScreenPtr pThinc, thincClientPtr cl,
	const BoxPtr row, BoxPtr notdone, unsigned char *data,
	int pitch)
{
	t_dbg_name(func)
	BoxRec curr;
	int err;
	unsigned char *dptr;

	thincDbg(T_DBG_PROTO, ("%s: sending row (%d,%d):(%d,%d)\n", func,
		row->x1, row->y1, row->x2, row->y2));

	notdone->y1 = curr.y1 = row->y1;
	notdone->y2 = curr.y2 = row->y2;
	notdone->x1 = curr.x1 = row->x1;
	curr.x2 = row->x1 + pThinc->updateWidth;
	notdone->x2 = row->x2;

	dptr = data;
	while(curr.x2 <= row->x2) {
		thincDbg(T_DBG_PROTO, ("%s: should i yield?\n", func));
		err = thincShouldYield(cl);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't verify "
				"if I should yield or not\n", func));
			return T_ERROR;
		} else if(err == T_YES) {
			thincDbg(T_DBG_PROTO, ("%s: yielding control to "
				"higher priority events\n", func));
			return T_PART;
		}

		if(data) {
			t_get_datapointer(data, row, pitch, pThinc->Bpp,
				&curr, dptr);
		}

		err = thincRawFlushSimple(pThinc, cl, &curr, dptr, pitch);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"sub-rectangle (%d,%d):(%d,%d) to client\n", 
				func, curr.x1, curr.y1, curr.x2, curr.y2));
			return T_ERROR;
		}
		notdone->x1 += pThinc->updateWidth; /* one rectangle less */
		curr.x1 = curr.x2;	/* next rectangle to send */
		curr.x2 += pThinc->updateWidth;
	}

	if(curr.x1 != row->x2) {
		/* last not perfect */
		curr.x2 = row->x2;
		err = thincRawFlushSimple(pThinc, cl, &curr, dptr, pitch);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"sub-rectangle (%d,%d):(%d,%d) to client\n", 
				func, curr.x1, curr.y1, curr.x2, curr.y2));
			return T_ERROR;
		}
		notdone->x1 = row->x2;	/* all sent */
	}
	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;
}

static int thincRawFlushSimple(thincScreenPtr pThinc, thincClientPtr cl,
	BoxPtr rect, unsigned char *data, int pitch)
{
	thincRawObj obj;

	obj.x = rect->x1;
	obj.y = rect->y1;
	obj.width = obj.real_width = box_width(rect);
	obj.height = obj.real_height = box_height(rect);

	/* FIXME: We should be getting the data from above, even if it
	 * just points to the framebuffer */
	if(data) {
		obj.data = data;
		obj.data_pitch = pitch;
		obj.data_width = t_pix_to_bytes(obj.width,pThinc);
	} else {
		/* get data from framebuffer. see FIXME above. */
		t_get_fbpointer(pThinc, rect, obj.data, obj.data_width,
			obj.data_pitch);
	}
	obj.data_height = obj.height;
	obj.data_len = obj.data_width*obj.data_height;
	obj.flags = 0;
	obj.extra_len = 0;
	obj.extra = obj.extra_data;
	obj.total_len = 0;
	
	return thinc_msgStackExec(cl, cl->rawStack, &obj);
}

/******************* RAW MODULES *********************/
#ifdef T_GETSTATS
static int thinc_rawPreStats(thincClientPtr cl, void *msg)
{
	stats_start_nostr(t_start);
	stats_init_simple(writeTime);
	stats_init_simple(blockTime);

	return T_OK;
}

static int thinc_rawPostStats(thincClientPtr cl, void *msg)
{
	unsigned long t_end;
	thincRawObj *obj = msg;

	/* TODO: Maybe stats should be per client?? */
	stats_end(cl->parent, t_start, t_end, rawStats.flush);
	stats_inc(cl->parent, rawStats.count);
	stats_add(cl->parent, rawStats.write, writeTime);
	stats_add(cl->parent, rawStats.block, blockTime);
	stats_add(cl->parent, rawStats.size, obj->total_len);
	return T_OK;
}
#endif /* T_GETSTATS */

static int thinc_rawCompress(thincClientPtr cl, void *msg)
{
	const char *modname = rawCompress.name;
	register thincRawObj *obj = msg;
	thinc_fbZData *z;
	int err;
	unsigned int z_max_size;
	register unsigned int i;
	register unsigned char *from;
	unsigned char *zbuff;
	thincScreenPtr pThinc = cl->parent;
	unsigned long zsize;

	instr_add("RAW_COMP", TI_START);
	
	thincDbg(T_DBG_PROTO, ("%s: compressing RAW update\n", modname));

	/* if data size is not big enough, just punt. This also caches
	 * cached updates */
	if(obj->data_len < T_ZMINSIZE) {
		thincDbg(T_DBG_PROTO, ("%s: update too small, finished\n",
			modname));
		instr_add("RAW_COMP", TI_END);
		return T_OK;
	}

	instr_add("RAW_COMP_PREP", TI_START);
	err = thinc_compressPrepare(pThinc);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't prepare "
			"compression engine\n", modname));
		return T_ERROR;
	}
	instr_add("RAW_COMP_PREP", TI_END);

	/* compress the thing. We need to compress one line at
	 * a time, since we don't have a contiguous buffer. The 
	 * procedure goes like:
	 * => set avail_in to line width
	 * => set next_in to start of line
	 * => call deflate
	 * => hopefully upon return, avail_in will be zero. In this
	 * case, we go to the next line and start again.
	 * When all the lines have been compressed, we call 
	 * compressFlush() to get all the output, 
	 * compressFlush returns how much compressed data was produced. 
	 *
	 * nice, eh?
	 */
	instr_add("RAW_COMP_PARAMS", TI_START);
	z_max_size = thinc_compressMaxSize(pThinc);
	zbuff = thinc_compressGetBuffer(pThinc);
#ifdef T_USEPNG
	thinc_compressSetParams(pThinc, T_ZPARAM_W|T_ZPARAM_H,
		obj->real_width, obj->real_height);
#else	
	thinc_compressSetParams(pThinc, T_ZPARAM_OUT,
			zbuff, z_max_size);
#endif
	thincDbg(T_DBG_PROTO, ("%s: compressing %u lines of data, each of "
		"width %u\n", modname, obj->data_height, obj->data_width));
	instr_add("RAW_COMP_PARAMS", TI_END);
	from = obj->data;
	for(i=0; i<obj->data_height; ++i) {
		instr_add("RAW_COMP_COMP", TI_START);
		thinc_compressSetParams(pThinc, T_ZPARAM_IN, from, 
			obj->data_width);
		err = thinc_compress(pThinc);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't compress "
				"line %u\n", modname, i));
			return T_ERROR;
		}
		from += obj->data_pitch;
		instr_add("RAW_COMP_COMP", TI_END);
	}
	thincDbg(T_DBG_PROTO, ("%s: all data compressed, flushing it to "
		"objects data\n", modname));
	/* XXX: For some reason, zlib doesn't flush all the data, unless
	 * it has something else as input (even tho is not using it).
	 * looks like a case of checking for parameters before figuring out
	 * what needs to be done. For flushing, it shouldn't need any
	 * additional input. oh well.
	 * FIXME: This is a bug in the client. */
#if 0
#ifndef T_USEPNG
	thinc_compressSetParams(pThinc, T_ZPARAM_IN, from, 1);
#endif
#endif
	instr_add("RAW_COMP_FLUSH", TI_START);
	err = thinc_compressFlush(pThinc);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush compressed "
			"data\n", modname));
		return T_ERROR;
	}
	instr_add("RAW_COMP_FLUSH", TI_END);
	/* sweet, now fill in appropriate parts of object and off we go */
	zsize = err;
	thincDbg(T_DBG_PROTO, ("%s: compressed RAW update from %lu bytes to "
		"%lu bytes\n", modname, (unsigned long)obj->data_len, zsize));

	obj->extra_len += T_FB_ZDATA_SIZE;
	/* sanity check */
	checkExtraBuffer(obj);
	z = (thinc_fbZData*)obj->extra;
	z->zsize = htonl(zsize);
	z->usize = htonl(obj->data_len);
	obj->extra += T_FB_ZDATA_SIZE;
	obj->flags |= T_FB_RUPDATE_COMPRESSED;

	zbuff = thinc_compressGetBuffer(pThinc);
	/* careful with this, eh? */
	obj->data = zbuff;
	obj->data_len = zsize;
	/* compressed data is one contiguous chunk of data */
	obj->data_height = 1;
	obj->data_width = obj->data_len;
	obj->data_pitch = 0;

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	instr_add("RAW_COMP", TI_END);
	return T_OK;
}

/* all the information it needs should be in the raw message object */
static int thinc_rawSend(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, rawSend.name)
	thincRawObj *obj = msg;
	unsigned int tobuffer, lines, write_len, notyet;
	int err, i;
	thinc_rawUpdate *r;
	register unsigned char *from, *to;

	instr_add("RAW_SEND", TI_START);
	thincDbg(T_DBG_PROTO, ("%s: sending RAW update to client %s\n",
		modname, cl->id));

	/* for stats */
	obj->total_len = THINC_HEADERSIZE + T_FB_RUPDATE_SIZE + 
		obj->extra_len + obj->data_len;
	/* can we buffer the message? If it's not small enough to fit in the
	 * write buffer for the client, flush the buffers and
	 * then send it directly. Ok, we actually try to fit at least
	 * some portion of the update in whatever space we have in the
	 * buffer */
	write_len = obj->data_len;
	if(!write_len) {
		/* a cached update */
		tobuffer = 0;
		lines = 0;
	} else {
		tobuffer = T_CL_MAXMSGDATA(cl, T_FB_RUPDATE);
		if(write_len < tobuffer)
			tobuffer = write_len;
		lines = tobuffer/write_len;
		/* just send complete lines */ 
		tobuffer = lines * obj->data_width;  
	}
	/* add extra stuff. We don't do this before to not interfere with the
	 * complete lines computation. */
	write_len += obj->extra_len;
	tobuffer += obj->extra_len;

	thincDbg(T_DBG_PROTO, ("%s: trying to buffer first %d bytes of %d\n",
		modname, tobuffer, write_len));

	/* TODO: Maybe thincBufferMsg() should make us think that we can always
	 * buffer data */
	instr_add("RAW_SEND_HEAD", TI_START);
	r = thincBufferMsg(cl, T_FB_RUPDATE, obj->flags, T_FB_RUPDATE_SIZE,
		tobuffer);
	instr_add("RAW_SEND_HEAD", TI_END);
	if(!r) {
		/* buffering's not being done */
		thincFlushBuffer(cl);
		instr_add("RAW_SEND_WRITE", TI_START);
		err = thinc_rawWriteUpdate(cl, obj);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not send RAW "
				"update\n", modname));
		} 
		instr_add("RAW_SEND_WRITE", TI_END);
		goto end;
	}
	/* build the message */
	r->x = htons(obj->x);
	r->y = htons(obj->y);
	r->width = htons(obj->width);
	r->height = htons(obj->height);

	/* copy extra headers */
	instr_add("RAW_SEND_EXTRA", TI_START);
	thincDbg(T_DBG_PROTO, ("%s: writing extra headers: size %u\n",
		modname, obj->extra_len));
	to = ((unsigned char*)r) + T_FB_RUPDATE_SIZE;
	from = obj->extra_data;
	memcpy(to, from, obj->extra_len);
	instr_add("RAW_SEND_EXTRA", TI_END);

	/* copy the data to the buffer */
	thincDbg(T_DBG_PROTO, ("%s: writing %d lines of %d bytes each\n",
		modname, lines, obj->data_width));
	to += obj->extra_len;
	from = obj->data;
	/* for cached updates, this will be skipped and nothing will be
	 * sent */
	instr_add("RAW_SEND_DATA_BUF", TI_START);
	for(i=0; i<lines; ++i) {
		memcpy(to, from, obj->data_width);
		to += obj->data_width;
		from += obj->data_pitch;
	}
	instr_add("RAW_SEND_DATA_BUF", TI_END);

	notyet = write_len - tobuffer;
	if(notyet == 0) {
		thincDbg(T_DBG_PROTO, ("%s: RAW update completely buffered, "
			"finished\n", modname));
		err = T_OK;
		goto end;
	}

	/* not so easy :( */
	instr_add("RAW_SEND_FLUSH", TI_START);
	thincFlushBuffer(cl);
	instr_add("RAW_SEND_FLUSH", TI_END);
	/* buffer is empty so we can figure out if the rest of the data can
	 * fit or not. If it can't, then just write it out normally. If
	 * it can, then buffer ir */
	tobuffer = T_CL_MAXDATA(cl);
	thincDbg(T_DBG_PROTO, ("%s: can put %d bytes in the buffer. I have "
		"%d\n", modname, tobuffer, notyet));
	if(tobuffer < notyet) {
		/* just punt */
		instr_add("RAW_SEND_DATA_WRITE", TI_START);
		err = thinc_rawWriteUpdateData(cl, from, obj->data_width,
			obj->data_height-lines, obj->data_pitch);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not send RAW "
				"update\n", modname));
		}
		instr_add("RAW_SEND_DATA_WRITE", TI_END);
		goto end;
	}

	/* sweet, can buffer it all */
	instr_add("RAW_SEND_DATA_BUFFER", TI_START);
	to = thincBufferMsg(cl, 0, 0, 0, notyet);
	if(!to) {
		err = T_ERROR;
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Could not buffer RAW "
			"update\n", modname));
		goto end;
	}
	for(i=lines; i<obj->data_height; ++i) {
		memcpy(to, from, obj->data_width);
		to += obj->data_width;
		from += obj->data_pitch;
	}
	instr_add("RAW_SEND_DATA_BUFFER", TI_END);

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	err = T_OK;
end:
	instr_add("RAW_SEND", TI_END);
	/* this is the end, my only friend, the end... */
	return err;
}

static int thinc_rawWriteUpdate(thincClientPtr cl, thincRawObj *obj)
{
	t_dbg_name(func)
	thinc_rawUpdate *r;
	int err;

	thincDbg(T_DBG_PROTO, ("%s\n", func));

	r = (thinc_rawUpdate*)t_msgBuffer;
	r->x = htons(obj->x);
	r->y = htons(obj->y);
	r->width = htons(obj->width);
	r->height = htons(obj->height);

	err = thincSendMsg(cl, T_FB_RUPDATE, T_FB_RUPDATE_SIZE, obj->flags);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send header for "
			"RAW update\n", func));
		return T_ERROR;
	}

	/* send extra headers */
	if(!obj->extra_len) 
		goto data;

	err = thincWriteData(cl, obj->extra_data, obj->extra_len);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send extra "
			"headers for RAW update\n", func));
		return T_ERROR;
	}

data:
	err = thinc_rawWriteUpdateData(cl, obj->data, obj->data_width,
		obj->data_height, obj->data_pitch);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send data for "
			"RAW update\n", func));
		return T_ERROR;
	}

	thincDbg(T_DBG_PROTO, ("%s: finished\n", func));
	return T_OK;
}

/* sends height lines of width width bytes separated by pitch bytes, starting
 * at data (changed semantics from old WriteFBUpdateData() */
static int thinc_rawWriteUpdateData(thincClientPtr cl, unsigned char *data,
	unsigned int width, unsigned int height, unsigned int pitch)
{
	register unsigned char *curr;
	register unsigned int i;
	int err;

	curr = data;
	for(i=0; i<height; ++i) {
		err = thincWriteData(cl, curr, width);
		if(err == T_ERROR) 
			return T_ERROR;
		curr += pitch;
	}

	return T_OK;
}

/* resize the update */
static int thinc_rawResize(thincClientPtr cl, void *msg)
{
	const char *modname = rawResize.name;
	thincRawObj *obj = msg;
	static unsigned char *rdata=NULL; /* where we put the resized data */
	thinc_reszData *r;
	unsigned int rw, rh, rdw, rlen;	/* rdw: resized data width */
	int err;

	thincDbg(T_DBG_PROTO, ("%s: resizing RAW command, current "
		"client dimensions: %ux%u\n", modname, cl->curr_width,
		cl->curr_height));

	rw = obj->width*cl->factor_x;
	rdw = t_pix_to_bytes(rw, cl->parent);
	rh = obj->height*cl->factor_y;
	rlen = rdw*rh;

	thincDbg(T_DBG_PROTO, ("%s: resized from %dx%d (%d) to %dx%d\n", 
		modname, obj->width, rw, rdw, obj->height, rh));

	if(!rw || !rh) {
		thincDbg(T_DBG_PROTO, ("%s: resized width or height are "
			"zero, discarding command\n", modname));
		return T_STOP;
	}

	/* i'm very lazy today... */
	if(rdata)
		free(rdata);
	rdata = malloc(rlen);
	if(!rdata) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate data "
			"for resized update\n", modname));
		return T_ERROR;
	}

	thincDbg(T_DBG_PROTO, ("%s: resizing update to %ux%u\n", modname,
		rw, rh));

#if 0
	err = thinc_resize32(obj->data, rdata, obj->width, obj->height,
		rw, rh, obj->data_pitch, rdw);
#else
	err = cl->resize(obj->data, rdata, obj->width, obj->height,
		rw, rh, obj->data_pitch, rdw);
#endif
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't resize update\n",
			modname));
		return T_ERROR;
	}

	obj->data = rdata;
#if 0
	obj->data_len = obj->data_width = rlen;
	obj->data_height = 1;
	obj->data_pitch = 0;
#else
	obj->data_len = rlen;
	obj->data_width = t_pix_to_bytes(rw, cl->parent);
	obj->data_pitch = obj->data_width;
	obj->data_height = rh;
	obj->real_width = rw;
	obj->real_height = rh;
#endif

	obj->extra_len += T_RESZ_DATA_SIZE;
	checkExtraBuffer(obj);
	r = (thinc_reszData*) obj->extra;
	r->fb_width = htons(cl->curr_width);
	r->fb_height = htons(cl->curr_height);
	r->size = htonl(obj->data_len);
	obj->flags |= T_FB_RUPDATE_RESIZED;

	obj->extra += T_RESZ_DATA_SIZE;

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));
	return T_OK;
}

/* Writes the update to a local file. useful for debugging. */
#define do_write(fd, ptr, len) \
	err = write(fd, ptr, len); \
	if(err == -1) { \
		thincDbg(T_DBG_ERROR, ("%s: Couldn't write RAW update to " \
			"file: %s\n", modname, strerror(errno))); \
		return T_ERROR; \
	}

#ifndef T_WIN32
/* TODO: Add config options for this?? */
static int thinc_rawWriteFile(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, rawWriteFile.name)
	static int count = 0;
	unsigned int a, i;
	char fn[255];
	int err, dump; /* fd */
	const char *base = "/tmp/raw/raw";
	thincRawObj *obj = msg;
	register unsigned char *curr;
	const int mode = S_IRUSR | S_IWUSR | S_IRGRP | S_IROTH;

	/* Write a header?? */
	const unsigned char write_header = T_TRUE;

	if(!obj->data) {
		return T_OK;
	}

	strncpy(fn, base, 12);
	sprintf(fn+12, "%d", count++);

	thincError("writing file %s\n", fn);

	dump = creat(fn, mode);
	if(dump == -1) {
		thincDbg(T_DBG_ERROR, ("%s: Couldn't open raw file %s. "
			"count %d: %s\n", modname,fn,count,strerror(errno)));
		return T_ERROR;
	}

	if(write_header) {
		a = obj->width;
		a = htonl(a);
		do_write(dump, &a, 4);
		a = obj->height;
		a = htonl(a);
		do_write(dump, &a, 4);
		a = cl->parent->Bpp;
		a = htonl(a);
		do_write(dump, &a, 4);
		a= 0;	/* pad */
		do_write(dump, &a, 4);
	}

	curr = obj->data;
	for(i=0; i<obj->data_height; ++i) {
		do_write(dump, curr, obj->data_width);
		curr += obj->data_pitch;
	}

	close(dump);
	return T_OK;
}

#endif /* T_WIN32 */ 

/* copies message data to a contiguous buffer */
static int thinc_rawCopyData(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, rawCopyData.name)
	register unsigned char *from, *to;
	register unsigned int j=0;
	thincRawObj *obj = msg;

	instr_add("RAW_COPY", TI_START);

	thincDbg(T_DBG_PROTO, ("%s: Copying %u bytes of data. There are %u "
		"lines, each of width %u and separated by %u bytes\n", 
		modname, obj->data_len, obj->data_height, obj->data_width,
		obj->data_pitch));

	/* FIXME: This would be great if it wasn't for the fact that
	 * fucking mozilla keeps drawing everything twice... */
#if 0
	if(obj->data_len > T_RCACHE_MAX_SZ) {
		thincDbg(T_DBG_PROTO, ("%s: update is too big for cache, "
			"finished\n", modname));
		instr_add("RAW_COPY", TI_END);
		return T_OK;
	}
#endif
	if((!obj->data_pitch) || (obj->data_width == obj->data_pitch)) {
		thincDbg(T_DBG_PROTO, ("%s: data is contiguous, finished\n",
			modname));
		instr_add("RAW_COPY", TI_END);
		return T_OK;
	}
	t_bug_on_2(
		!local_buf,
		T_BUG("RAW data buffer not initialized"),
		return T_ERROR
	);

	from = obj->data;
	to = local_buf;
	for(j=0; j<obj->data_height; ++j) {
		memcpy(to, from, obj->data_width);
		to += obj->data_width;
		from += obj->data_pitch;
	}

	obj->data = local_buf;
#if 0
	obj->data_width = obj->data_len;
	obj->data_height = 1;
	obj->data_pitch = 0;
#else
	obj->data_pitch = obj->data_width;
#endif

	thincDbg(T_DBG_PROTO, ("%s: finished. data_width is now %u (%u)\n", 
		modname, obj->data_width, obj->data_len));

	instr_add("RAW_COPY", TI_END);
	return T_OK;
}

/* checks if update is in cache and if yes it takes out the data and sets up
 * the cache message */
static int thinc_rawChkCache(thincClientPtr cl, void *msg)
{
	const char *modname = rawChkCache.name;
	thincRawObj *obj = msg;
	int err;
	unsigned long int hash;

	instr_add("RAW_CACHE_CHK", TI_START);

	thincDbg(T_DBG_PROTO, ("%s: checking if RAW update is in cache\n",
		modname));

	/* FIXME: This would be great if it wasn't for the fact that
	 * fucking mozilla keeps drawing everything twice... */
#if 0
	if(obj->data_len > T_RCACHE_MAX_SZ) {
		thincDbg(T_DBG_PROTO, ("%s: update is too big for cache, "
			"finished\n", modname));
		instr_add("RAW_CACHE_CHK", TI_END);
		return T_OK;
	}
#endif

#ifdef T_STRICT_CHECKS
	/* sanity check: is data contiguous? */
	if( (obj->data_pitch != 0) && (obj->data_pitch != obj->data_width) ) {
		T_BUG("Fixing uncontiguous RAW update. Is CopyData module "
			"not loaded?");
		err = thinc_rawCopyData(cl, msg);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't save "
				"cache data\n", modname));
			return T_ERROR;
		}
	}
#endif
	instr_add("RAW_CACHE_QUERY", TI_START);
	err = thinc_cache_query_data(cl->raw_cache, obj->data, obj->data_len,
		&hash);
	instr_add("RAW_CACHE_QUERY", TI_END);
	obj->hash = hash;
	if(err == T_NO) {
		thinc_cacheAdd *a;
		thincDbg(T_DBG_PROTO, ("%s: update is not in cache :*( "
			"Telling client to add it, hash is %u\n", 
			modname, obj->hash));
		obj->extra_len += T_FB_CACHE_ADD_SIZE;
		checkExtraBuffer(obj);
		a = (thinc_cacheAdd*) obj->extra;
		a->cache_id = htonl(obj->hash);
		obj->extra += T_FB_CACHE_ADD_SIZE;
		obj->flags |= T_FB_RUPDATE_ADDCACHE;
	} else {
		thinc_cacheData *c;
		thincDbg(T_DBG_PROTO, ("%s: sweeet, found cached data. hash "
			"is %u\n", modname, obj->hash));
		obj->extra_len += T_FB_CACHE_DATA_SIZE;
		checkExtraBuffer(obj);
		c = (thinc_cacheData*)obj->extra;
		c->cache_id = htonl(obj->hash);
		obj->extra += T_FB_CACHE_DATA_SIZE;
		obj->flags |= T_FB_RUPDATE_CACHED;

		/* fix up object so nobody sends data */
		obj->data = NULL;
		obj->data_len= obj->data_width= obj->data_height= \
			obj->data_pitch = 0;
	}

	thincDbg(T_DBG_PROTO, ("%s: finished\n", modname));

	instr_add("RAW_CACHE_CHK", TI_END);
	return T_OK;
}

static int thinc_rawAddCache(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, rawAddCache.name)
	thincRawObj *obj = msg;
	instr_add("RAW_CACHE_ADD", TI_START);

	thincDbg(T_DBG_PROTO, ("%s: possibly adding RAW update to cache\n",
		modname));

#if 0
	if(obj->flags & T_FB_RUPDATE_CACHED ) {
		thincDbg(T_DBG_PROTO, ("%s: update already cached, finished\n",
			modname));
		instr_add("RAW_CACHE_ADD", TI_END);
		return T_OK;
	} else {
		int err;
		instr_add("RAW_CACHE_ADD_ADD", TI_START);
		return T_OK;
		err = thinc_cache_add_hash(cl->raw_cache, obj->hash, NULL);
		instr_add("RAW_CACHE_ADD_ADD", TI_END);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add "
				"RAW update to cache\n", modname));
			return T_ERROR;
		} else {
			thincDbg(T_DBG_PROTO, ("%s: Added update to cache with "
				"hash %u, finished\n", modname, obj->hash));
			return T_OK;
		}
	}
#else
	if(obj->flags & T_FB_RUPDATE_ADDCACHE) {
		int err;
		instr_add("RAW_CACHE_ADD_ADD", TI_START);
		err = thinc_cache_add_hash(cl->raw_cache, obj->hash, NULL);
		instr_add("RAW_CACHE_ADD_ADD", TI_END);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add "
				"RAW update to cache\n", modname));
			return T_ERROR;
		} else {
			thincDbg(T_DBG_PROTO, ("%s: Added update to cache with "
				"hash %u, finished\n", modname, obj->hash));
			instr_add("RAW_CACHE_ADD", TI_END);
			return T_OK;
		}
	} else {
		thincDbg(T_DBG_PROTO, ("%s: update already cached (or not "
			"to be added to cache), finished\n", modname));
		instr_add("RAW_CACHE_ADD", TI_END);
		return T_OK;
	}

#endif
}

static int thinc_rawLogSize(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, rawLogSize.name)
	t_dbg_stmt(thincRawObj *obj = msg;)

	thincDbg(T_DBG_PROTO, ("%s: DIMENSIONS: %ux%u\n", modname,
		obj->width, obj->height));

	return T_OK;
}
