/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Text drawing functions
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 *
 */

/* set tabstop = 8 */

#include "thinc.h"

static int thincGlyphQueueCommand(thincScreenPtr pThinc, PixmapPtr pix,
	xRectangle *area, unsigned long color);

/**************************************************************************
 *                            Text Drawing Routines                       *
 **************************************************************************/

static void thincImageText( DrawablePtr pDraw, GCPtr pGC, int x, int y, 
	int count, unsigned short *chars, int which);
static int thincPolyText( DrawablePtr pDraw, GCPtr pGC, int x, int y, 
	int count, unsigned short *chars, int which);
/*
 * Aux. Function:
 * Returns rectangle which bounds the passed count characters
 * with the given font. It's not a minimal rectangle, since it uses 
 * characteristics of biggest character(s) in font.
 * Based on VNC's GetTextBoundingBox
 */
static void thincGetTextBounds(DrawablePtr pDrawable, FontPtr pFont, int x,
	int y, int count, xRectangle *pRec)
{
	int maxAscent, maxDescent, maxWidth, left;
	t_dbg_name(func)

	/* get info for biggest chars in the font. Need to chk in case
	 * some special char violates the max info */
	maxAscent = MAX(FONTMAXBOUNDS(pFont,ascent), FONTASCENT(pFont));
	maxDescent = MAX(FONTMAXBOUNDS(pFont, descent), FONTDESCENT(pFont));
	maxWidth = MAX(FONTMAXBOUNDS(pFont, characterWidth), 
		FONTMAXBOUNDS(pFont, rightSideBearing));

	thincDbg(T_DBG_DRAW, ("%s: ascent %d, descent %d, width %d\n", 
		func, maxAscent, maxDescent, maxWidth));

	/* in case a pFont extends outside (left) of its box */
	left = MIN(FONTMINBOUNDS(pFont, leftSideBearing), 0);

	pRec->x = pDrawable->x + x + left;
	pRec->y = pDrawable->y + y - maxAscent;
	pRec->width = maxWidth*count;
	pRec->height = maxAscent+maxDescent;

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

/********** High Level Text Routines *************/

void thincImageText8( DrawablePtr pDraw, GCPtr pGC, int x, int y, int count, 
	char *chars)
{
	thincImageText(pDraw, pGC, x, y, count, (unsigned short*)chars, 8);
}

void thincImageText16( DrawablePtr pDraw, GCPtr pGC, int x, int y, int count,
	unsigned short *chars)
{
	thincImageText(pDraw, pGC, x, y, count, chars, 16);
}

/* draws text. The text is drawn in the foreground color; the background 
 * color fills the remainder of the character rectangles. The coordinates 
 * specify the baseline and start of the text. */
static void thincImageText( DrawablePtr pDraw, GCPtr pGC, int x, int y, 
	int count, unsigned short *chars, int which)
{
	t_dbg_name(func)
	xRectangle update;
	RegionRec r;
	BoxPtr ext;
	int isOff;
	int res, size, width, height;
	unsigned char *bits;

	thincGC_pre(pGC);
	thincDbg(T_DBG_DRAW, ("%s count %d, (%d,%d). writing '%s'\n", 
		func, count, x, y, (char*)chars));
	thincDrawable_print(pDraw, "");

	/* call driver's func. */
	if(which == 8) {
		stats_inc(pThinc, countImageText8);
		(*pGC->ops->ImageText8)(pDraw, pGC, x, y, count, (char*)chars);
	} else if(which == 16) {
		stats_inc(pThinc, countImageText16);
		(*pGC->ops->ImageText16)(pDraw, pGC, x, y, count, chars);
	} else {
		T_BUG_ARG("Invalid argument [%d] to generic ImageText "
			"function", which);
		return;
	}

	if(count <= 0) {
		thincGC_post(pGC);
		thincDbg(T_DBG_DRAW, ("%s: not sending update\n", func));
		return;
	}

	thincGetTextBounds(pDraw, pGC->font, x, y, count, &update);

	/* clip update */
	call_clipUpdate(pGC, &r, update.x, update.y, update.width,
		update.height, noret);
	print_reg(T_DBG_DRAW, pScreen, &r, "sending update:");
	ext = REGION_EXTENTS(pScreen, &r);
	box_to_xrect(ext, update);

	width = update.width;
	height = update.height;
	size = bitmap_size(width, height);
	bits = (unsigned char*)ALLOCATE_LOCAL(size);
	if(!bits) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for bitmap\n", func));
		thincGC_post(pGC);
		return;
	}
	thincDbg(T_DBG_DRAW, ("%s: got bitmap of size %d bytes\n",func,size));
	
	isOff = is_offscreen(pDraw);

	if(pThincGC->flags & gc_weirdo) {
		thincDbg(T_DBG_DRAW, ("%s: Unsupported GC settings, bailing "
			"out\n", func));
		punt_to_raw(pThinc, isOff, res, pDraw, &r);
	} else if(pThinc->useBilevel) {
		if(isOff) {
			res = thinc_createBitmapPixmap(pThinc, 
				(PixmapPtr)pDraw, bits, &update, pGC->fgPixel);
			if(res == T_ERROR) {
				thincDbg(T_DBG_DRAW, ("ERROR: %s: Couldn't "
					"create bitmap\n", func));
			} else {
				res = thincQueueCommand(pThinc, 
					(PixmapPtr)pDraw, T_FB_BILEVEL, 
					&r, bits, 0, 0, width, height, 
					pGC->fgPixel, pGC->bgPixel, T_NO_FLAGS);
			}
		} else {
			res = thinc_createBitmapFB(pThinc, bits, &update, 
				pGC->fgPixel);
			if(res == T_ERROR) {
				thincDbg(T_DBG_DRAW, ("ERROR: %s: Couldn't "
					"create bitmap\n", func));
			} else {
				res = thincBufferNewCmdReg(pThinc, 
					(WindowPtr)pDraw, T_FB_BILEVEL, &r, 
					bits, 0, 0, width, height, 
					pGC->fgPixel, pGC->bgPixel, T_NO_FLAGS);
			}
		}
	} else {
		punt_to_raw(pThinc, isOff, res, pDraw, &r);
	}

	if(res != T_OK) {
		thincWarn("THINC: Couldn't process ImageText command\n");
	}

	REGION_UNINIT(pScreen, &r);

	thincGC_post(pGC);
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

int thincPolyText8( DrawablePtr pDraw, GCPtr pGC, int x, int y, int count, 
	char *chars)
{
	return thincPolyText(pDraw, pGC, x, y, count, 
		(unsigned short*) chars, 8);
}

int thincPolyText16( DrawablePtr pDraw, GCPtr pGC, int x, int y, int count, 
	unsigned short *chars)
{
	return thincPolyText(pDraw, pGC, x, y, count, chars, 16);
}

/* PolyText works like ImageText, except it draws with the current fill 
 * style for special effects such as shaded text.
 * XXX: Should we support all the other fill styles? Seems more work than
 * actually worth it. Right now only support FillSolid, so that it can
 * send a Bilevel */
static int thincPolyText( DrawablePtr pDraw, GCPtr pGC, int x, int y, 
	int count, unsigned short *chars, int which)
{
	t_dbg_name(func)
	int res = -1;
	xRectangle update;
	RegionRec r;
	int isOff;
	BoxPtr ext;
	int ret, size, width, height;
	unsigned char *bits;

	thincGC_pre(pGC);
	thincDbg(T_DBG_DRAW, ("%s. count %d, (%d, %d). bpp %d "
		"writing '%s'\n", func, count, x, y, pDraw->bitsPerPixel, 
		(char*)chars));
	thincDrawable_print(pDraw, "");

	if(count <= 0) {
		thincGC_post(pGC);
		thincDbg(T_DBG_DRAW, ("%s: finished. not sending update\n", 
			func));
		return res;
	}

	thincGetTextBounds(pDraw, pGC->font, x, y, count, &update);
	
	/* clip update */
	call_clipUpdate(pGC, &r, update.x, update.y, update.width,
			update.height, res);

	ext = REGION_EXTENTS(pScreen, &r);
	box_to_xrect_zerook(ext, update);

	/* call driver's func */
	if(which == 8) {
		stats_inc(pThinc, countPolyText8);
		res = (*pGC->ops->PolyText8)(pDraw, pGC, x, y, count, 
			(char*)chars);
	} else if(which == 16) {
		stats_inc(pThinc, countPolyText16);
		res = (*pGC->ops->PolyText16)(pDraw, pGC, x, y, count, chars);
	} else {
		T_BUG_ARG("Invalid argument [%d] to generic PolyText "
			"function", which);
		return T_ERROR;
	}
	isOff = is_offscreen(pDraw);

	/* create bitmap */
	if(pThincGC->flags & gc_weirdo) {
		thincDbg(T_DBG_DRAW, ("%s: Unsupported GC settings, bailing "
			"out\n", func));
		punt_to_raw(pThinc, isOff, ret, pDraw, &r);
	} else if(pThinc->useGlyph && pGC->fillStyle == FillSolid) {
		thincDbg(T_DBG_DRAW, ("%s: sending Glyph update, "
			"color %#lx\n", func, pGC->fgPixel));
		if(isOff) {
			ret = thincGlyphQueueCommand(pThinc, (PixmapPtr)pDraw,
				&update, pGC->fgPixel);
		} else {
			width = update.width;
			height = update.height;
			size = bitmap_size(width, height);
			bits = malloc(size);
			if(!bits) {
				thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
					"allocate memory "
					"for bitmap\n", func));
				thincGC_post(pGC);
				return res;
			}
			thincDbg(T_DBG_DRAW, ("%s: got bitmap of size %d "
				"bytes\n",func,size));
			ret = thinc_createBitmapFB(pThinc, bits, &update, 
				pGC->fgPixel);
			if(ret == T_ERROR) {
				thincDbg(T_DBG_DRAW, ("ERROR: %s: Couldn't "
					"create bitmap\n", func));
			} else {
				ret = thincBufferNewCmdReg(pThinc, 
					(WindowPtr)pDraw, T_FB_GLYPH, 
					&r, bits, 0, 0, width, height, 
					pGC->fgPixel, T_NO_FLAGS);
			}
			free(bits);
		}
	} else {
		punt_to_raw(pThinc, isOff, ret, pDraw, &r);
	}

	/* Should we return something bad? */
	if(ret != T_OK) {
		thincWarn("THINC: Couldn't process PolyText command\n");
	}

	REGION_UNINIT(pScreen, &r);

	thincGC_post(pGC);
	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	return res;
}

/********** Low Level Text Drawing Routines *************/

/* copy the glyphs defined by the bitmaps in pglyphBase and the font metrics 
 * in ppci to the DrawablePtr, pDrawable. It follows all fill, stipple
 * and tile rules.
 * Exact specification:
 * The glyph is painted with the current fill style. The character rectangle
 * is irrelevant for this operation.
 */
void thincPolyGlyphBlt( DrawablePtr pDrawable, GCPtr pGC, int x , int y, 
	unsigned int nglyph, CharInfoPtr *ppci, pointer pglyphBase)
{
	t_dbg_name(func)
	xRectangle update;
	RegionRec r;
	int isOff;
	int err;
	thincGC_pre(pGC);

	thincDbg(T_DBG_DRAW, ("%s (id: %ld), drawing %d glyphs at %d,%d\n", 
		func, pDrawable->id, nglyph, x, y));
	thincDrawable_print(pDrawable, "");

	stats_inc(pThinc, countPolyGlyphBlt);

	/* call driver's func */
	(*pGC->ops->PolyGlyphBlt)(pDrawable, pGC, x, y, nglyph,
		ppci, pglyphBase);

	if(nglyph <= 0) {
		thincGC_post(pGC);
		thincDbg(T_DBG_DRAW, ("%s: no glyphs to draw. finished\n",
			func));
		return;
	}

	thincGetTextBounds(pDrawable, pGC->font, x, y, nglyph, &update);

	/* clip the update */
	call_clipUpdate(pGC, &r, update.x, update.y, update.width,
		update.height, noret);

	thincDbg(T_DBG_DRAW, ("%s: sending update (%d,%d) -> (%d,%d)\n", func,
		update.x, update.y, update.width, update.height));

	isOff = is_offscreen(pDrawable);
	/* must be solid fill and no weird alu */
	if((pGC->fillStyle != FillSolid) || (pThincGC->flags & gc_weirdo) ) {
		thincDbg(T_DBG_DRAW, ("%s: FillStyle not solid or weird "
			"draw settings. sending RAW\n", func));
		punt_to_raw(pThinc, isOff, err, pDrawable, &r);
	} else if(pThinc->useGlyph) {
		thincDbg(T_DBG_DRAW, ("%s: sending Glyph update\n", func));
		/* FIXME: This doesn't work. need to create bitmap
		 * beforehand */
		if(isOff)
			err = thincQueueCommand(pThinc, (PixmapPtr)pDrawable,
				T_FB_GLYPH, &r, NULL, 0, 0, 0, 0, pGC->fgPixel,
				T_NO_FLAGS);
		else
			err = thincBufferNewCmdReg(pThinc,(WindowPtr)pDrawable,
				T_FB_GLYPH, &r, NULL, 0, 0, 0, 0, 
				pGC->fgPixel, T_NO_FLAGS);
	} else {
		thincDbg(T_DBG_DRAW, ("%s: Not using GLYPH updates, sending "
			"RAW then\n", func));
		punt_to_raw(pThinc, isOff, err, pDrawable, &r);
	}

	if(err != T_OK) {
		thincWarn("THINC: Couldn't process PolyGlyphBlt command\n");
	}

	REGION_UNINIT(pScreen, &r);

	/* tell high level function we sent an update */
	pThincGC->flags |= gc_lowlevelUpdate;

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

/* Copy the glyphs defined by the bitmaps in pglyphBase and the font metrics 
 * in ppci to the DrawablePtr, pDrawable.
 * Exact specification:
 * Fills the character rectangle with the background color, and then the 
 * glyph is applied in the foreground color. Note that the glyph can extend 
 * outside of the character rectangle.
 * Update:
 * Get the bounding box for the text being drawn and send a bilevel command
 * with the foreground and background colors.
 */
/* TODO:
 * Add offscreen stuff */
void thincImageGlyphBlt(DrawablePtr pDrawable, GCPtr pGC, int x , int y, 
	unsigned int nglyph, CharInfoPtr *ppci, pointer pglyphBase)
{
	t_dbg_name(func)
	xRectangle update;
	RegionRec r;
	int isOff;
	int err;
	thincGC_pre(pGC);

	thincDbg(T_DBG_DRAW, ("%s (id: %ld), drawing %d glyphs at %d,%d\n", 
		func, pDrawable->id, nglyph, x, y));
	thincDrawable_print(pDrawable, "");

	/* call driver's func */
	(*pGC->ops->ImageGlyphBlt)(pDrawable, pGC, x, y, nglyph,
		ppci, pglyphBase);

	stats_inc(pThinc, countImageGlyphBlt);

	if(nglyph <= 0) {
		thincGC_post(pGC);
		thincDbg(T_DBG_DRAW, ("%s: no glyphs to draw. finished\n",
			func));
		return;
	}	

	thincGetTextBounds(pDrawable, pGC->font, x, y, nglyph, &update);

	/* clip the update */
	call_clipUpdate(pGC, &r, update.x, update.y, update.width,
		update.height, noret);

	thincDbg(T_DBG_DRAW, ("%s: sending update (%d,%d) -> (%d,%d)\n", func,
		update.x, update.y, update.width, update.height));

	isOff = is_offscreen(pDrawable);
	if(pThincGC->flags & gc_weirdo ) {
		thincDbg(T_DBG_DRAW, ("%s: Weird draw settings. "
			"sending RAW\n", func));
		punt_to_raw(pThinc, isOff, err, pDrawable, &r);
	} else if(pThinc->useGlyph) {
		thincDbg(T_DBG_DRAW, ("%s: sending Glyph update\n", func));
		/* FIXME: This doesn't work, need to create bitmap here */
		if(isOff)
			err = thincQueueCommand(pThinc, (PixmapPtr)pDrawable,
				T_FB_BILEVEL, &r, NULL, 0, 0, 0, 0, 
				pGC->fgPixel, pGC->bgPixel, T_NO_FLAGS);
		else
			err = thincBufferNewCmdReg(pThinc,(WindowPtr)pDrawable,
				T_FB_BILEVEL, &r, NULL, 0, 0, 0, 0, 
				pGC->fgPixel, pGC->bgPixel, T_NO_FLAGS);
	} else {
		thincDbg(T_DBG_DRAW, ("%s: Not using GLYPH updates, sending "
			"raw then\n", func));
		punt_to_raw(pThinc, isOff, err, pDrawable, &r);
	}

	if(err != T_OK) {
		thincWarn("THINC: Couldn't process ImageGlyphBlt command\n");
	}

	REGION_UNINIT(pScreen, &r);

	/* tell high level function we sent an update */
	pThincGC->flags |= gc_lowlevelUpdate;

	thincGC_post(pGC);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
}

/* Use this to queue a glyph. It first finds transparent regions from the
 * bitmap. Then, if any clients are resized, it tries to convert the glyph
 * to a bilevel (resizing doesn't work with glyphs for reasons that are
 * too long to explain here :) */
static int thincGlyphQueueCommand(thincScreenPtr pThinc, PixmapPtr pix,
	xRectangle *area, unsigned long color)
{
	t_dbg_name(func)
	int err;
	xRectangle rect;
	BoxRec box;
	RegionRec reg;
	unsigned int bitmap_size;
	unsigned char *bitmap;
#if 0	/* part of the #if 0 below */
	unsigned long bg;
#endif

	thincDbg(T_DBG_DRAW, ("%s: extracting real glyph regions and "
		"queueing them\n", func));

	bitmap_size = thinc_getBitmapRect(pThinc, pix, area, color, &rect);
	if(bitmap_size == 0) {
		thincDbg(T_DBG_DRAW, ("%s: region transparent, finished\n",
			func));
		return T_OK;
	}

	thincDbg(T_DBG_DRAW, ("%s: got non-transparent rect %u,%u=>%ux%u\n",
		func, rect.x, rect.y, rect.width, rect.height));

	bitmap = malloc(bitmap_size);
	if(!bitmap) {
		thincDbg(T_DBG_DRAW, ("ERROR: %s: Couldn't allocate memory "
			"for bitmap\n", func));
		return T_ERROR;
	}

	xrect_to_box(rect, &box);
	REGION_INIT(pThinc->pScreen, &reg, &box, 1);

	/* FIXME: Not the best heuristic: If between now and the time this 
	 * command is flushed a client gets resized we'll have to punt the
	 * command as RAW, which may cost us */
#if 0
	if(pThinc->resizeCount) {
		thincDbg(T_DBG_DRAW, ("%s: have %d clients resized, chking "
			"if we get a bilevel\n", func, pThinc->resizeCount));
		err = thinc_createBitmapPixmapBilevel(pThinc, pix, bitmap,
			&rect, color, &bg);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s; Couldn't create "
				"bitmap\n", func));
			return T_ERROR;
		}
		if(bg != color) {
			thincDbg(T_DBG_DRAW, ("%s: got a bilevel! bg is %lu\n",
				func, bg));
			/* got a bilevel! */
			err = thincQueueCommand(pThinc, pix, T_FB_BILEVEL, 
				&reg, bitmap, 0, 0, rect.width, rect.height, 
				color, bg, T_NO_FLAGS);
		} else {
			thincDbg(T_DBG_DRAW, ("%s: errr.. not a bilevel\n",
				func));
			/* glyph, glyph, glyph.. will have to punt it at flush
			 * time :( */
			err = thincQueueCommand(pThinc, pix, T_FB_GLYPH, &reg, 
				bitmap, 0, 0, rect.width, rect.height, color,
				T_NO_FLAGS);
		}
	} else {
		err = thinc_createBitmapPixmap(pThinc, pix, bitmap, &rect, 
			color);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create "
				"bitmap\n", func));
			return T_ERROR;
		}
		err = thincQueueCommand(pThinc, pix, T_FB_GLYPH, &reg, bitmap, 
			0, 0, rect.width, rect.height, color, T_NO_FLAGS);
	}
#else
	err = thinc_createBitmapPixmap(pThinc, pix, bitmap, &rect, 
			color);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create "
					"bitmap\n", func));
		return T_ERROR;
	}
	err = thincQueueCommand(pThinc, pix, T_FB_GLYPH, &reg, bitmap, 
			0, 0, rect.width, rect.height, color, T_NO_FLAGS);
#endif
	REGION_UNINIT(pThinc->pScreen, &reg);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR:%s: Couldn't queue GLYPH "
			"update\n", func));
		return T_ERROR;
	}

	free(bitmap);

	thincDbg(T_DBG_DRAW, ("%s: finished\n", func));
	return T_OK;
}
