/*
 *
 * Header with utility functions for thinc
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */

/* set tabstop=8 */

#ifndef _THINC_UTIL_H_
#define _THINC_UTIL_H_
#ifdef T_WIN32
#include "winsock2.h"
#else
#include <sys/poll.h>
#include <sys/time.h>
#include <time.h>

/* for backtrace */
#include <execinfo.h>
#endif /* T_WIN32 */
/* Additional message buffer. Messages cannot be longer than T_MAXSIZE bytes.
 * Use in conjuction with thincSendMsg().
 * Only use to send messages that need to be sent to one specific client.
 * For messages directed to all clients, use thincBufferMsg() */
extern char *t_msgBuffer;

extern unsigned long writeTime;
extern unsigned long blockTime;
extern unsigned long tmpSize;

/* default timeout in ms when waiting for data to come in a client's socket.
 * We need this so that a client cannot DoS us by not sending enough data.
 */
#define T_CL_TIMEOUT	10000

/* default port. Changeable via command in XF86Config */
#define _THINC_PORT 	20000

/* the current time in milliseconds */
#define gettime_ms(t) do { \
	struct timeval __tv;		\
	gettimeofday(&__tv, NULL);	\
	t = (__tv.tv_sec*1000) + (__tv.tv_usec/1000); \
} while(0)

/* the current time in microseconds */
#define gettime_us(t) do { \
    struct timeval __tv;        \
    gettimeofday(&__tv, NULL);  \
    t = (__tv.tv_sec*1000000) + (__tv.tv_usec); \
} while(0)

/* useful macros */
#ifndef MIN
# define MIN(a,b) ((a) < (b) ? (a) : (b))
#endif
#ifndef MAX
# define MAX(a,b) ((a) > (b) ? (a) : (b))
#endif
/* these are only for positive numbers and may not be cast (or otherwise) safe
 * (undef first to make sure we get ours) */
#undef FLOOR
#define FLOOR(x) (int)(x)
#undef CEILING
#define CEILING(x) ((x) == FLOOR(x) ? (x) : FLOOR(x)+1)
#undef ABS
#define ABS(x) ( (x) < 0 ? -(x) : (x) )

/* thincNetwork.c */
#if 0
int tcpThincListen(const char *addr, int port);
#else
int tcpThincListen(const char *addr, int port, int *find_port);
#endif
int tcpThincClose(int sock);
int thincDoPollRevs(int sock, short events, short *revents, int timeout);
#define thincDoPoll(sock, events, timeout) \
	thincDoPollRevs(sock, events, NULL, timeout);
int thincDoSelect(const fd_set *fd_in, fd_set *fd_out, int maxfd);

#define T_POLLIN	POLLIN
#define T_POLLOUT	POLLOUT

/* thincUtil.c */
int thincEndian();

/* is the client completely connected? */
#ifdef T_PROTO2
#define t_cl_connected(cl)	((cl)->connected)
#else
#define t_cl_connected(cl)	((cl)->stage == TC_ST_CONN)
#endif	/* T_PROTO2 */

/* call function f for every client with p as arguments. p should be
 * the complete argument list enclosed in parentheses, and the client
 * argument should be called "cl" and should have been defined in
 * the function. For example:
 * call_for_all(thincSendFBCopy, (pThinc, cl, copyReg, dx, dy));
 * you can also deference cl, like:
 * call_for_all(thincWriteData, (cl->sock, cl->msgBuffer, len));
 *
 * Variables that need to be defined in the scope where we're called:
 * thincScreenPtr pThinc;
 * thincClientPtr cl;
 */
#define call_for_all(f, p) { 						\
	thincClientPtr __tmp; 						\
	t_list_for_each_entry(cl,__tmp,thincClientRec,&pThinc->clients,link)\
	{ \
		/* only send to fully connected clients */ 		\
		if(!t_cl_connected(cl)) { 				\
			continue;					\
		}							\
		/* function call */					\
		res = f p;						\
		if(res == T_ERROR) 					\
			cl->close_cl(cl);				\
	}								\
}

#define t_first_client(pThinc) \
	t_list_first_entry(&(pThinc)->clients, thincClientRec, link)

#define bits_per_byte 		8u
#define log2_bits_per_byte 	3u
#define bytes_per_pixel(bpp) ( bpp == 1 ? 1 : (bpp) >> log2_bits_per_byte)

#define t_pix_to_bytes(pix, pThinc) ((pix) << (pThinc)->Bpp_shift)

#define redo_region(reg, ext, data) \
	copy_box((reg)->extents, (ext)); \
	(reg)->data = (data);

#define reg_has_data(reg) \
	((reg)->data && ((reg)->data != &miEmptyData))

/* copy regions */

/* not a complete copy, since the region data is not actually copied.
 * For a real copy use REGION_COPY() */
#define copy_reg(dst, src) \
	copy_box((dst).extents, (src).extents); \
	(dst).data = (src).data;

#define copy_regp(dst, src) \
	copy_box((dst)->extents, (src)->extents); \
	(dst)->data = (src)->data;

/* copy boxes */
#define copy_box(dst, src) \
	(dst).x1 = (src).x1; \
	(dst).y1 = (src).y1; \
	(dst).x2 = (src).x2; \
	(dst).y2 = (src).y2;

#define copy_boxp(dst, src) \
	(dst)->x1 = (src)->x1; \
	(dst)->y1 = (src)->y1; \
	(dst)->x2 = (src)->x2; \
	(dst)->y2 = (src)->y2;

/* copy rects */
#define copy_rect(dst, src) \
	(dst).x = (src).x; \
	(dst).y = (src).y; \
	(dst).width = (src).width; \
	(dst).height = (src).height;

#define copy_rectp(dst, src) \
	(dst)->x = (src)->x; \
	(dst)->y = (src)->y; \
	(dst)->width = (src)->width; \
	(dst)->height = (src)->height;

#define copy_rects(to, from, num) { \
	int _copy_rects_idx; \
	for(_copy_rects_idx=0; _copy_rects_idx<num; ++_copy_rects_idx) { \
		thincDbg(T_DBG_HELP, ("%s: copying %u,%u=>%ux%u\n", __func__, \
			(from)[_copy_rects_idx].x, (from)[_copy_rects_idx].y, (from)[_copy_rects_idx].width, \
			(from)[_copy_rects_idx].height)); \
		copy_rect((to)[_copy_rects_idx], (from)[_copy_rects_idx]); \
	} \
}

/* copies a thinc_rectRegion. which are the same as xRects. but just in case
 * the definition of a rectangle changes */
#define copy_trect(dst, src) \
	copy_rect(dst, src)

#define copy_trectp(dst, src) \
	copy_rectp(dst, src)

/* convert a BoxPtr to a thinc_rectRegion */
#define box_to_rect(b, r) \
	(r).x = (b)->x1; \
	(r).y = (b)->y1; \
	(r).width = box_width(b); \
	(r).height = box_height(b);

/* convert a BoxPtr to a thinc_rectRegion whose members are in network
 * byte order */
#define box_to_rect_hton(b, r) \
	(r).x = htons((b)->x1); \
	(r).y = htons((b)->y1); \
	(r).width = htons(box_width(b)); \
	(r).height = htons(box_height(b));

/* r is a pointer to a thinc_rectRegion */
#define box_to_rectp(b, r) \
	(r)->x = (b)->x1; \
	(r)->y = (b)->y1; \
	(r)->width = box_width(b); \
	(r)->height = box_height(b);

/* r is a pointer to a thinc_rectRegion, whose elements are stored in
 * network byte order */
#define box_to_rectp_hton(b, r) \
	(r)->x = htons((b)->x1); \
	(r)->y = htons((b)->y1); \
	(r)->width = htons(box_width(b)); \
	(r)->height = htons(box_height(b));

/* convert box to xRectangle */
#define box_to_xrect(b, r) \
	(r).x = (b)->x1; \
	(r).y = (b)->y1; \
	(r).width = box_width(b); \
	(r).height = box_height(b);

#define box_to_xrect_zerook(b, r) \
	(r).x = (b)->x1; \
	(r).y = (b)->y1; \
	(r).width = box_width_zerook(b); \
	(r).height = box_height_zerook(b);

#define xrect_to_box(r, b) \
	(b)->x1 = (r).x; \
	(b)->y1 = (r).y; \
	(b)->x2 = (b)->x1 + (r).width; \
	(b)->y2 = (b)->y1 + (r).height;

#define trect_to_box(r, b) \
	xrect_to_box(r, b)

/* convert coords to BoxRec */
#define coords_to_box(b, x, y, w, h) \
	(b).x1 = (x); \
	(b).y1 = (y); \
	(b).x2 = (x)+(w); \
	(b).y2 = (y)+(h); 

/* convert from coordinates to a RegionPtr */
#define coords_to_reg(pScreen, reg, x, y, w, h) { \
	BoxRec b; \
	b.x1 = (x); \
	b.y1 = (y); \
	b.x2 = (x)+(w); \
	b.y2 = (y)+(h); \
	REGION_INIT(pScreen, (reg), &b, 1); \
}

#define box_width(boxp) \
	( ( (boxp)->x1 == (boxp)->x2) ? 1 : (boxp)->x2 - (boxp)->x1)

#define box_width_zerook(boxp) \
	( (boxp)->x2 - (boxp)->x1 )

#define box_width_careful(boxp) \
	( ( (boxp)->x1 >= (boxp)->x2) ? 0: (boxp)->x2 - (boxp)->x1 )
	
#define box_height(boxp) \
	( ( (boxp)->y1 == (boxp)->y2) ? 1 : (boxp)->y2-(boxp)->y1)

#define box_height_zerook(boxp) \
	( (boxp)->y2 - (boxp)->y1 )

#define box_height_careful(boxp) \
	( ( (boxp)->y1 >= (boxp)->y2) ? 0: (boxp)->y2 - (boxp)->y1 )

#define box_width_bytes(pThinc, boxp) \
	( t_pix_to_bytes(box_width(boxp),pThinc))

#define box_size(pThinc, boxp) \
	( box_width_bytes(pThinc, boxp) * box_height(boxp) )

#define box_size_pix(boxp) \
	( box_width(boxp) * box_height(boxp) )

#define rect_size(pThinc, rectp) \
	( rect_width_bytes(pThinc, rectp) * (rectp)->height )

#define rect_width_bytes(pThinc, rectp) \
	( t_pix_to_bytes((rectp)->width,pThinc)

#define trect_size(pThinc, rectp) \
	rect_size(pThinc, rectp)

#define trect_width_bytes(pThinc, rectp) \
	rect_width_bytes(pThinc, rectp)

/* FIXME: This should get a pThinc pointer and call REGION_EXTENTS */
#define region_is_empty(reg) \
	( (REGION_NUM_RECTS(reg) == 0) || (box_is_empty(&(reg)->extents)) )
	
#define box_is_empty(box) \
	( ((box)->x2 <= (box)->x1) || \
	  ((box)->y2 <= (box)->y1) )

/* true iff b1 is inside b2. both are BoxPtr */
#define box_in_box(b1, b2) \
	( ((b2)->x1 <= (b1)->x1) && \
	  ((b2)->x2 >= (b1)->x2) && \
	  ((b2)->y1 <= (b1)->y1) && \
	  ((b2)->y2 >= (b1)->y2) )

#ifdef T_WIN32
/* true if two box equal.*/
#define boxs_equal(b1, b2) \
	( ((b2)->x1 == (b1)->x1) && \
	  ((b2)->x2 == (b1)->x2) && \
	  ((b2)->y1 == (b1)->y1) && \
	  ((b2)->y2 == (b1)->y2) )
#endif

/* true iff boxes overlap. from EXTENTCHECK()  -> miregion.c */
#define t_intersectBox(b1, b2) \
	(!( ((b1)->x2 <= (b2)->x1)  || \
	    ((b1)->x1 >= (b2)->x2)  || \
	    ((b1)->y2 <= (b2)->y1)  || \
	    ((b1)->y1 >= (b2)->y2) ) )

#define t_get_fbpointer(pThinc, rect, p, rect_width, fb_width) { \
	int width = box_width(rect); \
	int x_off = t_pix_to_bytes((rect)->x1, pThinc); \
	rect_width = t_pix_to_bytes(width, pThinc); \
	fb_width = (pThinc)->widthBytes; \
	p = (pThinc)->theFB + x_off + ((rect)->y1 * fb_width); \
}

#define t_get_pixpointer(pThinc, pix, rect, p, Bpp, rect_width, pitch) { \
	int x_off; \
	int width = box_width(rect); \
	Bpp = (pix)->drawable.bitsPerPixel >> 3; \
	x_off = (rect)->x1 * bpp; \
	rect_width = width * bpp; \
	pitch = (pix)->devKind; \
	p = ((unsigned char*)thinc_getPixmap(pix)) + x_off + ((rect)->y1 * pitch); \
}

/* out: ptr
 * data -> pointer to start of data 
 * base -> base rectangle 
 * in: the rest */
#define t_get_datapointer(start_ptr, base_box, pitch, Bpp, rect, ptr) \
do { \
	unsigned int __x0 = (rect)->x1 - (base_box)->x1; \
	unsigned int __y0 = (rect)->y1 - (base_box)->y1; \
	unsigned int __x_off = __x0 * Bpp; \
	ptr = start_ptr + __x_off + (__y0 * pitch); \
} while(0);

#define box_translate(b, dx, dy) \
	(b)->x1 += (dx); \
	(b)->x2 += (dx); \
	(b)->y1 += (dy); \
	(b)->y2 += (dy);

#define box_equal(b1, b2) \
	( ( (b1)->x1 == (b2)->x1 ) && \
	  ( (b1)->x2 == (b2)->x2 ) && \
	  ( (b1)->y1 == (b2)->y1 ) && \
	  ( (b1)->y2 == (b2)->y2 ) )

#define reg_equal_simple(pScreen, reg1, reg2) \
	box_equal( REGION_EXTENTS(pScreen, (reg1)),  \
		REGION_EXTENTS(pScreen, (reg2)) )

#define reg_equal(pScreen, reg1, reg2) \
	( ( reg_equal_simple(pScreen, reg1, reg2) ) && \
	  ( (reg1)->data && (reg2)->data ) && \
	  ( REGION_NUM_RECTS(reg1) == REGION_NUM_RECTS(reg2) ) && \
	  ( !memcmp((reg1)->data, (reg2)->data,  \
		REGION_SZOF(REGION_NUM_RECTS(reg1))) ) )

/**** Bitmap manipulation macros ****/

/* width in bytes of a bitmap that covers w pixels */
#if 0
#define bitmap_width(w) \
	( CEILING((float)(w)/T_PIXELS_PER_BYTE) )
#else
/* standard case: (w+7) >> 3 */
#define bitmap_width(w) \
	( ((w) + (T_PIXELS_PER_BYTE-1)) >> T_LOG2_PIXELS_PER_BYTE)
#endif

/* pad size to 32 bits */
#define bitmap_width_pad(w) \
	( ( num_pad(w) ) >> T_LOG2_PIXELS_PER_BYTE)

#define bitmap_size(w, h) \
	( (bitmap_width(w)) * (h))

#define box_bitmap_width(box) \
	( bitmap_width(box_width(box)) )

#define box_bitmap_size(box) \
	 ( bitmap_size(box_width(box), box_height(box)) )

/* pad to next 32 bit boundary */
#define num_pad(num) \
	((num + T_PIXELS_PER_LONG-1) & ~(T_PIXELS_PER_LONG-1))

/******** DEBUGGING AND ERROR REPORTING *********/

#ifdef __GNUC__ /* make gcc check for correctness of messages */
void thincError(char *fmt, ...)
	__attribute__ ((format (__printf__, 1, 2)));

void thincWarn(char *fmt, ...)
	__attribute__ ((format (__printf__, 1, 2)));
void thincInfo(char *format, ...) 
	__attribute__ ((format (__printf__, 1, 2)));
void thincMsg(char *format, ...) 
	__attribute__ ((format (__printf__, 1, 2)));
/* print to log. __attribute__ makes gcc check that args are correct */
/* TODO: Should we add debug flag to print or not print the time? */
void thincLog(char *format, ...) 
	__attribute__ ((format (__printf__, 1, 2)));
#else
void thincError(char *fmt, ...);
void thincWarn(char *fmt, ...);
void thincLog(char *format, ...); 
void thincMsg(char *fmt, ...);
void thincInfo(char *fmt, ...);
#endif

/* where we store the current debug level */
extern unsigned short thincDebugLevel;

/* TODO: Define strings for each debug level which can be used
 * in the configuration file */
#ifdef T_DO_DEBUG

#define T_DBG_NONE 	0x0000 /* no debugging */

#define T_DBG_ERROR	0x0001 /* error messages from low level functions */

/* core messsages */
#define T_DBG_BASE	0x0002 /* initialization and basic thinc functions */
#define T_DBG_CLIENT	0x0004 /* client related messages */
#define T_DBG_IO	0x0008 /* i/o msgs (read/write/select) */
#define T_DBG_PROTO	0x0010 /* protocol msgs [16] */
#define T_DBG_STATS	0x0020 /* statistics messages [32] */
#define T_DBG_OTHER	0x0040 /* things that don't fit anywhere else [64] */
#define T_DBG_UNUSED	0x0080	/* == UNUSED == */

/* draw layer messages */
#define T_DBG_DRAW	0x0100 /* draw msgs (specially gc msgs) [256] */
#define T_DBG_WIN	0x0200 /* window msgs [512] */
#define T_DBG_CMD	0x0400 /* pixmap msgs [1024] */
#define T_DBG_INPUT	0x0800 /* keyboard, mouse, cursor (input devs) [2048]*/
#define T_DBG_VID	0x1000 /* Video [4096] */
#define T_DBG_COMP	0x2000 /* compression [8192] */

/* something else... */
#define T_DBG_TMP	0x4000 /* temp messages [16384] */
#define T_DBG_ANNOY	0x8000 /* messages that repeat all the time and annoy
				  us [32768] */

/* useful combinations for thincDebugLevel (add 1 to get Error messages) */
#define T_DBG_CORE	0x00FE	/* core messages (init, network, client, 
				   security, ... ) [254] */
#define T_DBG_X		0x3F00  /* X (drawing/displaying) messages [16128] */
#define T_DBG_ALL	0x7FFF 	/* everything minus annoying [32767] */
#define T_DBG_REALLYALL 0xFFFF	/* everything plus annoying messages (not
				   recommended) [65535] */

/* the default debugging level */
#define T_DBG_DEFAULT	T_DBG_NONE

#define T_DBG_MIN	T_DBG_NONE
#define T_DBG_MAX	T_DBG_REALLYALL

/* the macro used to print debugging info */
#define thincDbg(level, msg) \
	if(level & thincDebugLevel) \
		(thincLog msg);

#define t_dbg_name(f)	const char *f = __func__;
#define t_dbg_modname(m, name)	const char *m = name;
/* have a statement only when debugging is enabled */
#define t_dbg_stmt(s)	s

#else /* T_DO_DEBUG */
# define thincDbg(level, p)
#define t_dbg_name(func)
#define t_dbg_modname(m, name)
#define t_dbg_stmt(s)
#endif /* T_DO_DEBUG */

/* convert arg to a string */
#define t_str(arg)	#arg

/* why some many macros? we don't have variadic macros :(
 * (they're C99 and -pedantic as used in xfree86 doesn't like that) */
#define T_BUG(msg) \
	thincError("BUG: %s:%d:%s(): %s\n", __FILE__, __LINE__, __func__, msg);

#define T_BUG_ARG(msg, arg) \
	thincError("BUG: %s:%d:%s(): " msg "\n", __FILE__, __LINE__, \
		__func__, arg)

#define T_BUG_ARG_2(msg, arg1, arg2) \
	thincError("BUG: %s:%d:%s(): " msg "\n", __FILE__, __LINE__, \
		__func__, arg1, arg2)

#define T_BUG_ARG_3(msg, arg1, arg2, arg3) \
	thincError("BUG: %s:%d:%s(): " msg "\n", __FILE__, __LINE__, \
		__func__, arg1, arg2, arg3)

#define T_BUG_ARG_4(msg, arg1, arg2, arg3, arg4) \
	thincError("BUG: %s:%d:%s(): " msg "\n", __FILE__, __LINE__, \
		__func__, arg1, arg2, arg3, arg4)

/* execute something when a condition is true.  mutliple macros: the sad 
 * reality of not having variadic macros :( */
#define t_exec_on(cond, expr)	\
	if((cond)) {		\
		expr;		\
	}
#define t_exec_on_2(cond, expr1, expr2)	\
	if((cond)) {		\
		expr1;		\
		expr2;		\
	}

/* kind of assert but give more control to caller.  */
#ifdef T_STRICT_CHECKS
#define t_bug_on(cond, expr)	t_exec_on(cond, expr)
#define t_bug_on_2(cond, expr1, expr2)	t_exec_on_2(cond, expr1, expr2)
#else
#define t_bug_on(cond, expr)
#define t_bug_on_2(cond, expr1, expr2)
#endif

/* we can't use assert() in X, so we just define our own. use this one
 * with care, and preferably use t_bug_on() */
#define thinc_assert(cond) 	\
	t_exec_on_2(!(cond), T_BUG("Assertion failed"), exit(10))

/* useful for debugging */
#define print_box(level, box, s) \
	thincDbg(level, ("%s: %s x1 %d, y1 %d, x2 %d, y2 %d\n", func, (s), \
		(box)->x1, (box)->y1, (box)->x2, (box)->y2));

#define print_reg(level, pScreen, reg, s) \
	print_box(level, REGION_EXTENTS((pScreen), (reg)), (s)); 

#define really_print_reg(level, pScreen, reg, s) do { \
	BoxPtr _really_print_box; \
	int _really_print_idx; \
	print_reg(level, pScreen, reg, s); \
	_really_print_box = REGION_RECTS((reg)); \
	_really_print_idx = REGION_NUM_RECTS((reg)); \
	while(_really_print_idx) { \
		print_box(level, _really_print_box, s); \
		++_really_print_box; \
		--_really_print_idx; \
	} \
} while(0)

#define print_rect(level, rect, s) \
	thincDbg(level, ("%s: %s x %d, y %d, w %d, h %d\n", func, (s), \
		(rect)->x, (rect)->y, (rect)->width, (rect)->height));

#define print_rects(level, rects, num, s) { \
	int _print_rects_idx; \
	for(_print_rects_idx=0;_print_rects_idx<(num);++_print_rects_idx) {\
		print_rect(level, &(rects)[_print_rects_idx], s); \
	} \
}

/* call thincWriteData, check for return value and print a generic message
 * (with update type added in for flavor) */
#define call_writedata(err, cl, data, size, str) \
	err = thincWriteData(cl, data, size); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send %s update\n",\
			func, str)); \
		return T_ERROR; \
	}

#define call_sendmsg(err, cl, msgtype, msgsize, flags, str) \
	err = thincSendMsg(cl, msgtype, msgsize, flags); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send %s update\n",\
			func, str)); \
		return T_ERROR; \
	}

/* useful for debugging */
#define print_stack(level, s) { \
	void *trace[16]; \
	char **messages; \
	int i, trace_size; \
	trace_size = backtrace(trace, 16); \
	messages = backtrace_symbols(trace, trace_size); \
	thincDbg(level, ("%s backtrace: %s\n", func, s)); \
	for(i=0; i<trace_size; ++i) \
		thincDbg(level, ("=> [%d] %s\n", i, messages[i])); \
	free(messages); \
}
#ifdef T_WIN32
#define thinc_wrap(priv, real, mem, func)
#define thinc_unwrap(priv, real, mem)
#else /* T_WIN32 */
#define thinc_wrap(priv, real, mem, func) do {\
	priv->mem = real->mem; \
	real->mem = func; \
} while(0)
#define thinc_unwrap(priv, real, mem) \
	real->mem = priv->mem;
#endif /* T_WIN32 */

/* v2 protocol stuff */
#ifdef T_PROTO2

#define v2_newReply(rep, c_type, r_type) do {	\
	rep = (c_type*)t_msgBuffer;		\
	t_reply_set_type(rep, r_type);		\
} while(0)

#define v2_sendReply(err, cl, reply_size) do {		\
	err = thinc_v2SendReply(cl, reply_size);	\
	if(err == T_ERROR) {				\
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send reply "\
			"to client\n", func));	\
		return T_ERROR;			\
	}					\
} while(0)

#endif	/* T_PROTO2 */

#endif	/* THINC_UTIL_H */
