/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * XVideo support
 *
 * Author:
 * Ricardo Baratto
 * Network Computing Lab - Columbia University
 */
/* set tabstop = 8 */

/* 
 * Video is handled somewhat different than other messages to guarantee
 * maximum performance, while losing some flexibility and requiring some
 * more management. So a video stream is represented as a state machine,
 * where both client and server know the current state. The state
 * basically defines the characteristics of the video frame. State is
 * changed by the server by sending a message to the client describing
 * changes that need to be made to the video stream. This way, the
 * client doesn't need to check for flags on every message describing
 * the characteristics of the video frame being received (this is how
 * other messages are managed).
 */ 
/* TODO:
 * -> should we encapsulate video messages as command structures? they
 *  could be marked as RT.. the problem is that we'd need to save a copy
 *  of the video frame which could hurt performance.
 */

#include "thinc.h"
#include "xf86.h"
#include "xf86xv.h"
#include "Xv.h"
#include "fourcc.h"
#include <sys/poll.h>
#include <netinet/in.h>

#ifdef T_USEXVIDEO

#define t_vidPre(pScrn) \
	ScreenPtr pScreen = pScrn->pScreen; \
	thincScreenPtr pThinc = THINCPtr(pScreen);

/* this controls whether we wait for the client to acknowledge a new stream 
 * before we start sending data, or instead we assume that the stream can
 * be played and just start sending data right away.
 */
#define T_VID_NOWAITHANDSHAKE

#define T_NUM_ENCODINGS	1

/* encodings we support */
static XF86VideoEncodingRec thincVidEncodings[T_NUM_ENCODINGS] = 
{
	{ 0, "XV_IMAGE", 2046, 2046, {1, 1} }
};

/* we fill in the depth at initialization time according to our settings.
 * I don't know how we can do an overlay to support other depths... */
#define T_NUM_FORMATS 		1
static XF86VideoFormatRec thincVidFormats[T_NUM_FORMATS] = 
{
	{ 0, TrueColor}
};

/* number of ports is the number of separate data streams which the
 * adaptor can handle simultaneously. If you have more than one port,
 * the adaptor is expected to be able to render into more than one
 * window at a time */
#define T_MAX_PORTS		1

/* image formats */
#define T_NUM_IMAGES		1
static XF86ImageRec thincVidImages[T_NUM_IMAGES] = 
{
	/* defined for us in fourcc.h */
	XVIMAGE_YV12
};

/* states of a video stream */
#define T_STREAM_INACTIVE	0
#define T_STREAM_PENDING	1
#define T_STREAM_ACTIVE		2
#define T_STREAM_FAILED		3	/* client rejected this stream */

/* representation of a video stream */
typedef struct thincVideoStream_S {
	unsigned char state;	/* see above */
	unsigned char flags;	/* flags, like if the video stream is being
				   compressed */
	unsigned int id;
	unsigned int x;
	unsigned int y;
	unsigned int width;
	unsigned int height;
	unsigned int dst_width;	/* client scaling */
	unsigned int dst_height;
	unsigned int size;
	unsigned int timestamp;	/* synchronization */
	/* FIXME: Use our own structure so we can use function pointers
	 * to manage the stream */
	thinc_vidFormat format;

	/* FIXME: YUV specific stuff, move somewhere else. 
	 * Need to be int so we can use QueryImageAttributes */
	int yuv_offsets[3];
	int yuv_pitches[3];
} thincVideoStreamRec, *thincVideoStreamPtr;

/* this is a per-stream client video structure. It stores information of the
 * video stream that's just relevant to this client */
typedef struct thincClVideoStream_S {
	thincVideoStreamPtr stream;
	unsigned int width;	/* server scaling */
	unsigned int height;
	unsigned int size;	/* resized size */
	unsigned char flags;	/* misc flags (see below) */
	/* FIXME: YUV specific stuff, move somewhere else. 
	 * Need to be int so we can use QueryImageAttributes */
	int yuv_offsets[3];
	int yuv_pitches[3];
} thincClVideoStreamRec, *thincClVideoStreamPtr;

#define T_CLVIDEO_RESIZED	0x01	/* stream is resized */

/* this is our private video structure */
typedef struct thincVideo_S {
	thincScreenPtr	pThinc;
	int num_fmts;
	/* use same struct as protocol to make life easier */
	/* FIXME: Should have our own structure which wraps this one.
	 * The structure should have pointers to functions which know
	 * how to handle this format */
	thinc_vidFormat fmts[T_NUM_IMAGES];
	int num_streams;
	/* indexed by port number. 
	 * TODO: Remember to inactivate streams when closing clients */
	/* TODO: This is problematic with multiple clients */
	thincVideoStreamRec streams[T_MAX_PORTS];
} thincVideoRec, *thincVideoPtr;

/* video information for each client */
typedef struct thincClVideo_S {
	thincMsgStackPtr stack;
	int count;	/* number of video streams open with this client */
	/* fmts supported by this client */
	int num_fmts;
	/* we waste some space if the client doesn't support all formats...
	 * oh well */
	thinc_vidFormat fmts[T_NUM_IMAGES];
	/* streams associated with this client */
	thincClVideoStreamRec streams[T_MAX_PORTS];
} thincClVideoRec, *thincClVideoPtr;

/********* video delivery *********/
/* uses MsgStack */

/* methods */
static int thinc_videoSend(thincClientPtr cl, void *msg);
static int thinc_videoCompress(thincClientPtr cl, void *msg);
static int thinc_videoResize(thincClientPtr cl, void *msg);

/* priorities */
#define T_VIDPRIO_SEND		T_PRIO_SEND
#define T_VIDPRIO_COMPRESS	T_VIDPRIO_SEND-10
#define T_VIDPRIO_RESIZE	T_VIDPRIO_SEND-50

/* the modules */
thincDefModule(videoSend, "VIDEO_SEND", 11, T_VIDPRIO_SEND);
thincDefModule(videoCompress, "VIDEO_COMPRESS", 15, T_VIDPRIO_COMPRESS);
thincDefModule(videoResize, "VIDEO_RESIZE", 13, T_VIDPRIO_RESIZE);

/* video object for message stack */
typedef struct thincVideoObj_S {
	/* stream info */
	thincClVideoStreamPtr stream;
	/* the frame data */
	unsigned char *data;
} thincVideoObj;

/***** local functions *********/
/* Note: On error, functions called by X should return XvBadAlloc */

static int thincVideoInitStack(thincClientPtr cl);
static int thincVideoSendFmts(thincScreenPtr pThinc, thincClientPtr cl);
static void thincDestroyVideoStack(thincClientPtr cl);
static void thincVideoClientReady(thincClientPtr cl);

/* main video function */
static int thincVidPutImage(ScrnInfoPtr pScrn, 
	short src_x, short src_y, short drw_x, short drw_y, 
	short src_w, short src_h, short drw_w, short drw_h,
	int image, unsigned char *buf, short width, short height,
	Bool sync, RegionPtr clipBoxes, pointer data);
static int thincVidSetupStream(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, int format, long port_id, 
	unsigned char *image, int x, int y, int width, int height, 
	int dstw, int dsth);
static int thincVidEndStream(thincScreenPtr, thincVideoStreamPtr);
static int thincMoveStream(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, int x, int y);
static int thincScaleStream(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, int dstw, int dsth);

/* per client functions */
static int thincVidSetupStreamCl(thincClientPtr, thincVideoStreamPtr);
static int thincVidEndStreamCl(thincClientPtr cl, thincVideoStreamPtr stream);
static int thincMoveStreamCl(thincClientPtr cl, thincVideoStreamPtr stream);
static int thincScaleStreamCl(thincClientPtr cl, thincVideoStreamPtr stream);
static void thincVidParentStream(thincClientPtr cl, 
	thincClVideoStreamPtr cstream);
/* resize */
static int thincVidResizeStreamClient(thincClientPtr, thincClVideoStreamPtr);
static int thincVidResizeClStreams(thincClientPtr cl);
static void thincVidResizeStream(thincClientPtr cl, 
	thincClVideoStreamPtr cstream);

/* send */
static int thincVidSendImage(thincScreenPtr pThinc, thincVideoStreamPtr stream,
	unsigned char *image, int x, int y, int w, int h, int dstw, int dsth);
static int thincVidSendImageYUV12(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, unsigned char *image);
static int thincVidSendImageYUV12Cl(thincClientPtr cl, thincVideoStreamPtr,
	unsigned char*);

#if 0
static int thincVidSendImageYUV12Normal(thincScreenPtr pThinc, 
	thincClientPtr cl, thincVideoStreamPtr stream, unsigned char *image);
static int thincVidSendImageYUV12Compressed(thincScreenPtr pThinc, 
	thincClientPtr cl, thincVideoStreamPtr stream, unsigned char *image);
static int thincVidSendImageYUV12CompSimple(thincScreenPtr pThinc, 
	thincClientPtr cl, unsigned char *image, int size);
static int thincVidSendImageYUV12CompMultiple(thincScreenPtr pThinc, 
	thincClientPtr cl, thincVideoStreamPtr stream, unsigned char *image);
#endif

/* other (unused?) functions */
static int thincVidQueryImageAttributes(ScrnInfoPtr pScrn, int image, 
	unsigned short *width, unsigned short *height, int *pitches, 
	int *offsets);
void thincVidQueryBestSize(ScrnInfoPtr pScrn, Bool motion,
	short vid_w, short vid_h, short drw_w, short drw_h, 
	unsigned int *p_w, unsigned int *p_h, pointer data);
int thincVidGetPortAttribute(ScrnInfoPtr pScrn, Atom attribute,
	INT32 *value, pointer data);
int thincVidSetPortAttribute(ScrnInfoPtr pScrn, Atom attribute,
	INT32 value, pointer data);
void thincVidStopVideo(ScrnInfoPtr pScrn, pointer data, Bool Cleanup);

/************* INITIALIZATION *****************/
int thincInitVideo(thincScreenPtr pThinc)
{
	t_dbg_name(func)
	ScreenPtr pScreen = pThinc->pScreen;
	ScrnInfoPtr pScrn = xf86Screens[pScreen->myNum];
	XF86VideoAdaptorPtr adaptor;
	thincVideoPtr pVid;
	int i;

	thincDbg(T_DBG_VID, ("%s: Initializing Video Support\n", func));

	adaptor = xf86XVAllocateVideoAdaptorRec(pScrn);
	if(!adaptor) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for Video adaptor description\n", func));
		return T_FALSE;
	}

	/* can only put video into drawable, not get it */
	adaptor->type = XvInputMask;
	/* most drivers only support XvImages and video players 
	 * (mplayer/xine) don't seem to support anything else */
	adaptor->type |= XvImageMask;
	/* only draw to windows, kein pixmaps */
	adaptor->type |= XvWindowMask;

	/* TODO: Allow us to clip */
	adaptor->flags = VIDEO_NO_CLIPPING;
	adaptor->name = "THINC XVideo Adaptor";
	adaptor->nEncodings = T_NUM_ENCODINGS;
	adaptor->pEncodings = &thincVidEncodings[0];
	adaptor->nFormats = T_NUM_FORMATS;
	adaptor->pFormats = &thincVidFormats[0];
	adaptor->pFormats[0].depth = pThinc->depth;
	adaptor->nImages = T_NUM_IMAGES;
	adaptor->pImages = thincVidImages;
	/* TODO: Not yet used until we figure out all the other scheisse */
	adaptor->nAttributes = 0;
	adaptor->pAttributes = NULL;

	/* this is kind of sketchy. we store our private struct in
	 * here, but hidden from the xv layer. From mga and
	 * nv driver */
	adaptor->nPorts = T_MAX_PORTS;
	pVid = calloc(1,
		sizeof(thincVideoRec) + (sizeof(DevUnion)*T_MAX_PORTS));
	if(!pVid) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate "
			"memory for Video private data\n", func));
		xf86XVFreeVideoAdaptorRec(adaptor);
		return T_FALSE;
	}
	pThinc->thincVid = pVid;
	pThinc->thincVid->num_fmts = T_NUM_IMAGES;
	for(i=0; i<T_NUM_IMAGES; ++i) {
		pThinc->thincVid->fmts[i].id = thincVidImages[i].id;
	}
	adaptor->pPortPrivates = (DevUnion*)(&pVid[1]);

	/* we just use port numbers */
	for(i=0; i<T_MAX_PORTS; ++i) {
		adaptor->pPortPrivates[i].val = i;
		pThinc->thincVid->streams[i].state = T_STREAM_INACTIVE;
	}
	pThinc->thincVid->num_streams = 0;

	adaptor->PutVideo = NULL;
	adaptor->PutStill = NULL;
	adaptor->GetVideo = NULL;
	adaptor->GetStill = NULL;
	adaptor->StopVideo = thincVidStopVideo;
	adaptor->SetPortAttribute = thincVidSetPortAttribute;
	adaptor->GetPortAttribute = thincVidGetPortAttribute;
	adaptor->QueryBestSize = thincVidQueryBestSize;
	adaptor->PutImage = thincVidPutImage;
	adaptor->ReputImage = NULL;
	adaptor->QueryImageAttributes = thincVidQueryImageAttributes;

	thincDbg(T_DBG_VID, ("%s: ready to initialize XV\n", func));

	if(!xf86XVScreenInit(pScreen, &adaptor, 1)) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't initialize "
			"XVideo Extension\n", func));
		free(pThinc->thincVid);
		xf86XVFreeVideoAdaptorRec(adaptor);
		return T_FALSE;
	}

	/* clean up */
	xf86XVFreeVideoAdaptorRec(adaptor);
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_TRUE;
}

int thincVideoDisable(thincClientPtr cl)
{
	thincMsg("Disabling video on client %s\n", cl->id);
	/* do the whole shutdown process in case the disable comes in the 
	 * middle of the initialization */
	thincVidCloseClient(cl->parent, cl);
	cl->doVideo = T_FALSE;

	return T_OK;
}

/* Initialize video with this client */
int thincVInitNewClient(thincClientPtr cl)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = cl->parent;
	thincClVideoPtr clVid;
	int err;

	thincDbg(T_DBG_VID, ("%s: Initializing video for client %s\n", func,
		cl->id));

	if(!pThinc->useVideo || !pThinc->thincVid) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Video support not "
			"not present\n", func));
		return T_ERROR;
	}
#ifndef T_PROTO2
	if(cl->stage != TC_ST_VINIT || cl->state != TC_NONE) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Client %s is in "
			"unknown stage(%d)/state(%d)\n", func, cl->id,
			cl->stage, cl->state));
		return T_ERROR;
	}
#endif	/* NOT PROTO2 */

	/* allocate now so we can't fail later when it'd hurt more */
	clVid = malloc(sizeof(thincClVideoRec));
	if(!clVid) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't allocate memory "
			"for client's video information\n", func));
		return T_ERROR;
	}
	
	thincDbg(T_DBG_VID, ("%s: sending video formats to client. Have %d\n",
		func, pThinc->thincVid->num_fmts));

	/* send message to client with the video formats */
	err = thincVideoSendFmts(pThinc, cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send video "
			"formats to client\n", func));
		free(clVid);
		return T_ERROR;
	}

	clVid->stack = NULL;
	cl->video = clVid;

	/* create video message stack */
	err = thincVideoInitStack(cl);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create video "
			"message stack\n", func));
		free(cl->video);
		cl->video = NULL;
		return T_ERROR;
	}

#ifndef T_PROTO2
	/* update state */
	cl->state = TC_VINIT_INIT;
#endif	/* NOT T_PROTO2 */

	thincDbg(T_DBG_VID, ("%s: finished. Waiting for client to send "
		"its formats\n", func));
	return T_OK;
}

#ifdef T_PROTO2
static int thincVideoSendFmts(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_replyVideoServFmts* vf;
	thinc_vidSInit* sinit;
	thincVideoPtr pVid = pThinc->thincVid;
	int err, i;
	thinc_vidFormat f;

	thincDbg(T_DBG_VID, ("%s: sending video formats to client %s. "
		"Have %d\n", func, cl->id, pVid->num_fmts));

	v2_newReply(vf, thinc_replyVideoServFmts, T_REQ_VIDEO_SERV_FMTS);

	sinit = &vf->fmts;
	sinit->numformats = htonl(pVid->num_fmts);

	v2_sendReply(err, cl, t_reply_size(T_REPLY_VIDEO_SERV_FMTS));

	for(i=0; i<T_NUM_IMAGES; ++i) {
		f = pVid->fmts[i];
		f.id = htonl(f.id);
		err = thincWriteData(cl, &f, sizeof(thinc_vidFormat));
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO_SINIT message to client\n", func));
			return T_ERROR;
		}
	}
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

/* equivalent of thincHandleClientVInit() */
int thincVInitClientFmts(thincClientPtr cl, thinc_vidCInit *c)
{
	t_dbg_name(func)
	int err;

	thincDbg(T_DBG_CLIENT, ("%s: reading client video formats\n", func));

	c->numformats = ntohl(c->numformats);
	if(c->numformats == 0) {
		thincDbg(T_DBG_VID, ("%s: Client doesn't support "
			"any video formats\n", func));
		thincVideoDisable(cl);
		return T_OK;
	} else if(c->numformats > cl->parent->thincVid->num_fmts) {
		T_BUG_ARG("Broken client! It sent back more video formats (%d) "
			"than we support!", c->numformats);
		return T_ERROR;
	}

	/* read formats */
	err = thincReadData(cl, cl->video->fmts, 
		c->numformats * sizeof(thinc_vidFormat));
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client's "
			"video formats\n", func));
		return T_ERROR;
	} else if(err == T_CLOSED) {
		thincDbg(T_DBG_VID, ("%s: Client closed connection while "
			"waiting for client's video formats\n", func));
		return T_CLOSED;
	}
	cl->video->num_fmts = c->numformats;
	thincVideoClientReady(cl);

	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}
#else
static int thincVideoSendFmts(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_vidSInit* sinit;
	thincVideoPtr pVid = pThinc->thincVid;
	int err, i;
	thinc_vidFormat f;

	thincDbg(T_DBG_VID, ("%s: sending video formats to client %s. "
		"Have %d\n", func, cl->id, pVid->num_fmts));
	
	sinit = (thinc_vidSInit*) t_msgBuffer;
	sinit->numformats = htonl(pVid->num_fmts);

	err = thincSendMsg(cl, T_VIDEO_SINIT, T_VIDEO_SINIT_SIZE, 0);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send VIDEO_SINIT "
			"message to client\n", func));
		return T_ERROR;
	}
	for(i=0; i<T_NUM_IMAGES; ++i) {
		f = pVid->fmts[i];
		f.id = htonl(f.id);
		err = thincWriteData(cl, &f, sizeof(thinc_vidFormat));
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO_SINIT message to client\n", func));
			return T_ERROR;
		}
	}
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

int thincHandleClientVInit(thincClientPtr cl, int msg_type)
{
	t_dbg_name(func)
	thincScreenPtr pThinc = cl->parent;
	thinc_vidCInit c;
	int err;

	thincDbg(T_DBG_VID, ("%s: Handling client %s\n", func, cl->id));

	if(!pThinc->useVideo || !pThinc->thincVid) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Video support not "
			"not present\n", func));
		return T_ERROR;
	}
	if(cl->stage != TC_ST_VINIT || cl->state != TC_VINIT_INIT) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Client %s is in "
			"unknown stage(%d)/state(%d)\n", func, cl->id,
			cl->stage, cl->state));
		return T_ERROR;
	}
	if(msg_type != T_VIDEO_CINIT) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Invalid message %d "
			"received from client\n", func, msg_type));
		return T_ERROR;
	}

	err = thincReadAnyMsg(cl, &c, T_VIDEO_CINIT_SIZE);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client's "
			"message\n", func));
		return T_ERROR;
	} else if(err == T_CLOSED) {
		thincDbg(T_DBG_VID, ("%s: Client closed connection while "
			"waiting for VIDEO_CINIT message\n", func));
		return T_CLOSED;
	}

	c.numformats = ntohl(c.numformats);

	thincDbg(T_DBG_VID, ("%s: Got %d formats back from client\n",
		func, c.numformats));

	if(c.numformats == 0) {
		thincDbg(T_DBG_VID, ("%s: Client doesn't support "
			"any video formats\n", func));
		thincVideoDisable(cl);
		goto end;
	} else if(c.numformats > pThinc->thincVid->num_fmts) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Broken client!!! It sent "
			"back more formats (%d) than we support!!!\n", func,
			c.numformats));
		return T_ERROR;
	}

	/* read formats */
	err = thincReadData(cl, cl->video->fmts, 
		c.numformats * sizeof(thinc_vidFormat));
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client's "
			"video formats\n", func));
		return T_ERROR;
	} else if(err == T_CLOSED) {
		thincDbg(T_DBG_VID, ("%s: Client closed connection while "
			"waiting for client's video formats\n", func));
		return T_CLOSED;
	}
	cl->video->num_fmts = c.numformats;
	thincVideoClientReady(cl);

end:
	/* ready for next stage */
	cl->stage++;
	cl->state = TC_NONE;
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}
#endif	/* T_PROTO2 */

static void thincVideoClientReady(thincClientPtr cl)
{
	cl->video->count = 0;
	cl->doVideo = T_TRUE;
}

#define video_addmodule(mod) { \
	err = thinc_msgAddUsingModule(cl, video->stack, (mod)); \
	if(err == T_ERROR) { \
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't add VIDEO module "\
			"%s\n", func, (mod)->name)); \
		return T_ERROR; \
	} \
}

static int thincVideoInitStack(thincClientPtr cl)
{
	t_dbg_name(func)
	int err;
	thincScreenPtr pThinc = cl->parent;
	thincClVideoPtr video = cl->video;

	thincDbg(T_DBG_CLIENT, ("%s: creating VIDEO message stack\n", func));

	video->stack = thinc_msgStackCreate(cl);
	if(!video->stack) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't create VIDEO "
			"message stack\n", func));
		return T_ERROR;
	}

	/* add modules */
	video_addmodule(&videoSend);
	if(pThinc->compressVideo)
		video_addmodule(&videoCompress);
	thincDbg(T_DBG_CLIENT, ("%s: finished\n", func));
	return T_OK;
}

int thincHandleVidStartOK(thincClientPtr cl)
{
	t_dbg_name(func)
	thinc_vidStartOK ok;
	int err, id;
	thincScreenPtr pThinc = cl->parent;
	thincVideoStreamPtr stream;

	thincDbg(T_DBG_VID, ("%s: Reading video reply from client\n", func));

	err = thincReadAnyMsg(cl, &ok, T_VIDEO_STARTOK_SIZE);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't read client's "
			"message\n", func));
		return T_ERROR;
	} else if(err == T_CLOSED) {
		thincDbg(T_DBG_VID, ("%s: Client closed connection while "
			"waiting for VIDEO_CINIT message\n", func));
		return T_CLOSED;
	}

	id = ntohl(ok.id);

	thincDbg(T_DBG_VID, ("%s: Got video reply %d for stream %d\n", func,
		ok.ok, id));
	stream = &pThinc->thincVid->streams[id];

	if(ok.ok == T_VIDEO_NOTOK) {
		thincDbg(T_DBG_VID, ("%s: Client rejected video stream %d\n",
			func, id));
		stream->state = T_STREAM_FAILED;
	} else {
		thincDbg(T_DBG_VID, ("%s: woohoo!! client accepted video "
			"stream %d\n", func, id));
		stream->state = T_STREAM_ACTIVE;
	}
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

void thincVidCloseClient(thincScreenPtr pThinc, thincClientPtr cl)
{
	t_dbg_name(func)
	thincVideoPtr pVid;
	thincVideoStreamPtr stream;
	int i;

	thincDbg(T_DBG_VID, ("%s: client %s closed\n", func, cl->id));

	/* assumes clientCount hasn't been decreased */
	if(t_cl_connected(cl) && (pThinc->clientCount <= 1)) {
		/* no clients left after this one, deactivate all video
		 * streams */
		pVid = pThinc->thincVid;

		thincDbg(T_DBG_VID, ("%s: no active clients, deactivating %d "
			"streams\n", func, pVid->num_streams));

		for(i=0; i<pVid->num_streams; ++i) {
			stream = &pVid->streams[i];
			stream->state = T_STREAM_INACTIVE;
		}

		pVid->num_streams = 0;
	}

	if(cl->video) {
		thincDestroyVideoStack(cl);
		free(cl->video);
		cl->video = NULL;
	}

	thincDbg(T_DBG_VID, ("%s: finished\n", func));
}

static void thincDestroyVideoStack(thincClientPtr cl)
{
	if(cl->video->stack)
		thinc_msgStackDestroy(cl, cl->video->stack);
	cl->video->stack = NULL;
}

/************* RESIZE VIDEO STREAMS *************/
/* Client resized, should we do something?
 * chk if we should resize any active video streams. If yes, send messages
 * to client */
/* XXX: We just add the resize module, which we expect to check if the
 * video is bigger than the client's viewport and resize. If viewport is
 * smaller, the module shouldn't do anything */
int thincVidClientResized(thincClientPtr cl, int prevstate)
{
	t_dbg_name(func)
	thincClVideoPtr video; 
	int err;

	thincDbg(T_DBG_VID, ("%s: prevstate: %d\n", func, prevstate));

	if(!cl->parent->useVideo)
		return T_OK;
	if(!cl->doVideo)
		return T_OK;

	video = cl->video;

	if(!cl->parent->allowResize)
		return T_OK;

	/* resize all the streams for the client */
	if(cl->resized == T_CL_RESZ_LESS || prevstate == T_CL_RESZ_LESS) {
		err = thincVidResizeClStreams(cl);
		if(err == T_ERROR)
			return T_ERROR;
	}

	/* add resizing module */
	if(cl->resized == T_CL_RESZ_LESS) {
		thincDbg(T_DBG_VID, ("%s: client's window smaller than "
			"original %d\n", func, cl->resized));
		if(prevstate != T_CL_RESZ_LESS)
			video_addmodule(&videoResize);
	} else {
		thincDbg(T_DBG_VID, ("%s: client's window larger than "
			"original %d\n", func, cl->resized));
		if(prevstate == T_CL_RESZ_LESS)
			thinc_msgDelUsingModule(cl, video->stack,
				&videoResize);
	}

	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

/* resizes needed streams in the client */
static int thincVidResizeClStreams(thincClientPtr cl)
{
	thincClVideoPtr video = cl->video;
	thincClVideoStreamPtr cstream;
	unsigned int i;
	int err;

	for(i=0; i<video->count; ++i) {
		cstream = &video->streams[i];
		/* We can do this on PENDING assuming that the underlying
		 * network guarantees in-order delivery */
		if(	(cstream->stream->state == T_STREAM_INACTIVE) ||
			(cstream->stream->state == T_STREAM_FAILED)
		  )
			continue;
		err = thincVidResizeStreamClient(cl, cstream);
		if(err == T_ERROR)
			return T_ERROR;
	}

	return T_OK;
}

static int thincVidResizeStreamClient(thincClientPtr cl, 
	thincClVideoStreamPtr cstream)
{
	t_dbg_name(func)
	thincVideoStreamPtr stream = cstream->stream;
	thinc_vidResize *rsz;
	int err, nobuffer = T_FALSE;

	thincDbg(T_DBG_VID, ("%s: possibly resizing stream %d on client\n",
		func, stream->id));

	/* gotta resize */
	rsz = (thinc_vidResize*) thincBufferMsg(cl, T_VIDEO_RESIZE, 0,
		T_VIDEO_RESIZE_SIZE, 0);
	if(!rsz) {
		thincFlushBuffer(cl);
		nobuffer = T_TRUE;
		rsz = (thinc_vidResize*) t_msgBuffer;
	}

	if(cl->resized == T_CL_RESZ_LESS) {
		thincDbg(T_DBG_VID, ("%s: client's window smaller than "
			"original, resizing video for it\n", func));
		thincVidResizeStream(cl, cstream);
	} else {
		/* back to normal */
		thincDbg(T_DBG_VID, ("%s: client's window larger or equal "
			"than original, not doing anything to video stream\n",
			func));

		cstream->flags &= ~T_CLVIDEO_RESIZED;
		thincVidParentStream(cl, cstream);
	}

	thincDbg(T_DBG_VID, ("%s: resized stream to %ux%u. size is %u\n",
		func, cstream->width, cstream->height, cstream->size));
	rsz->id = htonl(stream->id);
	rsz->width = htons(cstream->width);
	rsz->height = htons(cstream->height);
	/* YUV_OFFSETS_PITCHES */
	rsz->yuv_yoffset = htonl(cstream->yuv_offsets[0]);
	rsz->yuv_uoffset = htonl(cstream->yuv_offsets[1]);
	rsz->yuv_voffset = htonl(cstream->yuv_offsets[2]);
	rsz->yuv_ypitch = htonl(cstream->yuv_pitches[0]);
	rsz->yuv_upitch = htonl(cstream->yuv_pitches[1]);
	rsz->yuv_vpitch = htonl(cstream->yuv_pitches[2]);

	if(nobuffer) {
		err = thincSendMsg(cl, T_VIDEO_RESIZE, T_VIDEO_RESIZE_SIZE,
			0);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO RESIZE message\n", func));
			return T_ERROR;
		}
	}

	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

static void thincVidResizeStream(thincClientPtr cl, 
	thincClVideoStreamPtr cstream)
{
	unsigned short w, h;
	int size;

	w = ((double)cstream->stream->width) * cl->factor_x;
	h = ((double)cstream->stream->height) * cl->factor_y;
	
	cstream->flags |= T_CLVIDEO_RESIZED;

	size = thincVidQueryImageAttributes(NULL, FOURCC_YV12, &w, &h,
		cstream->yuv_pitches, cstream->yuv_offsets);
	cstream->width = w;
	cstream->height = h;
	cstream->size = size;
}

/* i'm sorry, i can't think of a better name for this function. It just
 * basically inits the fields in the client stream struct to be the same
 * as the global stream struct (in contrast with ResizeStream that uses
 * the global stream to get an appropiately resized stream) */
static void thincVidParentStream(thincClientPtr cl, 
	thincClVideoStreamPtr cstream)
{
	thincVideoStreamPtr stream = cstream->stream;

	cstream->width = stream->width;
	cstream->height = stream->height;
	cstream->size = stream->size;

	/* YUV_OFFSETS_PITCHES */
	cstream->yuv_offsets[0] = stream->yuv_offsets[0];
	cstream->yuv_offsets[1] = stream->yuv_offsets[1];
	cstream->yuv_offsets[2] = stream->yuv_offsets[2];
	cstream->yuv_pitches[0] = stream->yuv_pitches[0];
	cstream->yuv_pitches[1] = stream->yuv_pitches[1];
	cstream->yuv_pitches[2] = stream->yuv_pitches[2];
}

/************** ENTRY POINT TO SENDING A VIDEO *****************/
/* from here we can:
 * -> initialize a new video stream (VidSetupStream)
 * -> destroy a stream (VidEndStream)
 * -> send the next video frame (VidSendImage)
 * -> change the characteristics of a stream (VidMoveStream, VidScaleStream)
 */
static int thincVidPutImage(ScrnInfoPtr pScrn, 
	short src_x, short src_y, short drw_x, short drw_y, 
	short src_w, short src_h, short drw_w, short drw_h,
	int format, unsigned char *buf, short width, short height,
	Bool sync, RegionPtr clipBoxes, pointer data)
{
	t_dbg_name(func)
	int err;
	long port_id = (long)data;
	thincVideoStreamPtr stream;
	thincVideoPtr pVid;
	t_vidPre(pScrn);

	thincDbg(T_DBG_VID, ("%s: putting video image with format %#x from "
		"location (%d,%d) => (%d, %d) into location "
		"(%d,%d) => (%d, %d). The image has dimensions %dx%d. "
		"sync is %d. Using port %ld\n",
		func, format,
		src_x, src_y, src_w, src_h,
		drw_x, drw_y, drw_w, drw_h, 
		width, height,
		sync, port_id));

	/* if no clients, nothing to do :P */
	/* FIXME: We should draw to the framebuffer!!! */
	if(pThinc->clientCount == 0) {
		thincDbg(T_DBG_VID, ("%s: No clients connected, ignoring "
			"request\n", func));
		return Success;
	}

	pVid = pThinc->thincVid;
	stream = &pVid->streams[port_id];

	/* for synchronization: get timestamp */
	gettime_us(stream->timestamp);
	thincDbg(T_DBG_VID, ("Video timestamp is %u\n", stream->timestamp));

	/* order:
	 * => inactive? set it up
	 * => ok, we have looked at it. did it change? stop and set up again.
	 * => ok, didn't change, do whatever's appropiate for its state
	 */
	if(stream->state == T_STREAM_INACTIVE) {
		thincDbg(T_DBG_VID, ("%s: stream is inactive. Setting "
			"it up\n", func));
		err = thincVidSetupStream(pThinc, stream, format,
			port_id, buf, drw_x, drw_y, width, height, 
			drw_w, drw_h);
#ifdef T_VID_NOWAITHANDSHAKE
		/* XXX: A little bit of a hack */
		if(err == T_ERROR) goto out;

		stream->state = T_STREAM_ACTIVE;

		/* now send first video frame. */
		err = thincVidSendImage(pThinc, stream, buf, 
			drw_x, drw_y, width, height, drw_w, drw_h);
#endif
	/* if stream characteristics changed, then we have a new video.
	 * shutdown and open again */
	} else if( 	(width != stream->width) || 
			(height != stream->height) ||
			(format != stream->format.id)
		) {
		thincDbg(T_DBG_VID, ("%s: stream changed. Ending "
			"and setting up again\n", func));
		err = thincVidEndStream(pThinc, stream);
		if(err == T_ERROR)
			goto out;
		err = thincVidSetupStream(pThinc, stream, format, port_id,
			buf, drw_x, drw_y, width, height, drw_w, drw_h);

#ifdef T_VID_NOWAITHANDSHAKE
		/* XXX: A little bit of a hack */
		if(err == T_ERROR) 
			goto out;

		/* assume client with positively ack the stream and start 
		 * sending video frames */
		stream->state = T_STREAM_ACTIVE;
		/* now send first video frame. */
		err = thincVidSendImage(pThinc, stream, buf, 
			drw_x, drw_y, width, height, drw_w, drw_h);
#endif
	} else if(stream->state == T_STREAM_PENDING) {
		/* TODO: What can we do?? */
		thincDbg(T_DBG_VID, ("%s: stream is pending, agghhh, "
			"where's the client???\n", func));
		err = T_OK;
	} else if(stream->state == T_STREAM_ACTIVE) {
		thincDbg(T_DBG_VID, ("%s: stream is active. sending "
			"data\n", func));
		err = thincVidSendImage(pThinc, stream, buf, 
			drw_x, drw_y, width, height, drw_w, drw_h);
	} else if(stream->state == T_STREAM_FAILED) {
		thincDbg(T_DBG_VID, ("%s: client rejected stream "
			"earlier. Ignoring request\n", func));
		err = T_OK;
	} else {
		T_BUG_ARG("Video stream is in an unknown state %d",
			stream->state);
		err = T_ERROR;
	}
out:

	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't put video image\n", 
			func));
		return BadAlloc;
	}

	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return Success;
}

/**** VIDEO STREAM SETUP ************/
static int thincVidSetupStream(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, int format, long port_id, 
	unsigned char *image, int x, int y, int width, int height, 
	int dstw, int dsth)
{
	t_dbg_name(func)
	thincClientPtr cl;
	thincVideoPtr pVid = pThinc->thincVid;
	int found = T_FALSE, res;
	unsigned int i;

	thincDbg(T_DBG_VID, ("%s: setting up video stream with format %#x. "
		"x %d, y %d, width %d, height %d. dstw %d, dsth %d\n", 
		func, format, x, y, width, height, dstw, dsth));

	/* find the appropiate format */
	for(i=0; i<pVid->num_fmts; ++i) {
		if(pVid->fmts[i].id == format) {
			stream->format = pVid->fmts[i];
			found = T_TRUE;
			break;
		}
	}
	if(!found) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Unsupported format %#x\n",
			func, format));
		return T_ERROR;
	}
	stream->id = port_id;
	/* TODO: Add compressed flag if necessary */
	stream->flags = 0;
	stream->x = x;
	stream->y = y;
	stream->width = width;
	stream->height = height;
	stream->dst_width = dstw;
	stream->dst_height = dsth;

	/* pre-compute size */
	/* YUV_OFFSETS_PITCHES  */
	stream->size = thincVidQueryImageAttributes(NULL, FOURCC_YV12,
		(unsigned short*)&stream->width,
		(unsigned short*)&stream->height,
		stream->yuv_pitches, stream->yuv_offsets);

	thincDbg(T_DBG_VID, ("%s: image has size %d bytes\n", func, 
		stream->size));

	call_for_all(thincVidSetupStreamCl, (cl, stream));
	/* nice */
	stream->state = T_STREAM_PENDING;
	pVid->num_streams++;
	thincDbg(T_DBG_VID, ("%s: finished. Have %d active streams\n", func,
		pVid->num_streams));
	return T_OK;
}

static int thincVidSetupStreamCl(thincClientPtr cl, thincVideoStreamPtr stream)
{
	t_dbg_name(func)
	thincClVideoStreamPtr clstream;
	thincClVideoPtr clvideo;
	thinc_vidStart *vs;
	int nobuffer = T_FALSE, err;

	thincDbg(T_DBG_VID, ("%s: setting up video stream for client %s\n",
		func, cl->id));

	if(!cl->doVideo) {
		thincDbg(T_DBG_VID, ("%s: client doesn't support video\n",
			func));
		return T_OK;
	}

	/* setup the stream for this client */
	clvideo = cl->video; 
	clstream = &clvideo->streams[stream->id];
	clstream->stream = stream;
	if(cl->resized == T_CL_RESZ_LESS) {
		thincVidResizeStream(cl, clstream);
	} else {
		clstream->flags = 0;
		thincVidParentStream(cl, clstream);
	}

	thincDbg(T_DBG_VID, ("%s: sending VIDEO_START message\n", func));

	/* TODO: Add compressed flag if necessary */
	vs = (thinc_vidStart*) thincBufferMsg(cl, T_VIDEO_START, 0,
		T_VIDEO_START_SIZE, 0);
	if(!vs) {
		thincDbg(T_DBG_VID, ("%s: Couldn't buffer VIDEO_START "
			"message\n", func));
		thincFlushBuffer(cl);
		vs = (thinc_vidStart*)t_msgBuffer;
		nobuffer = TRUE;
	}
	vs->id = htonl(stream->id);
	vs->fmt_id = htonl(stream->format.id);
	vs->x = htons(stream->x);
	vs->y = htons(stream->y);
	vs->width = htons(clstream->width);
	vs->height = htons(clstream->height);
	vs->dst_width = htons(stream->dst_width);
	vs->dst_height = htons(stream->dst_height);
	/* YUV_OFFSETS_PITCHES */
	vs->yuv_yoffset = htonl(clstream->yuv_offsets[0]);
	vs->yuv_uoffset = htonl(clstream->yuv_offsets[1]);
	vs->yuv_voffset = htonl(clstream->yuv_offsets[2]);
	vs->yuv_ypitch = htonl(clstream->yuv_pitches[0]);
	vs->yuv_upitch = htonl(clstream->yuv_pitches[1]);
	vs->yuv_vpitch = htonl(clstream->yuv_pitches[2]);

	if(nobuffer) {
		err = thincSendMsg(cl, T_VIDEO_START, T_VIDEO_START_SIZE, 0);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't "
				"send VIDEO_START message to client "
				"%s\n", func, cl->id));
			return T_ERROR;
		}
	}

	clvideo->count++;
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

static int thincVidEndStream(thincScreenPtr pThinc, thincVideoStreamPtr stream)
{
	t_dbg_name(func)
	thincClientPtr cl;
	int res;

	thincDbg(T_DBG_VID, ("%s: ending stream %d\n", func, stream->id));

	stream->state = T_STREAM_INACTIVE;

	pThinc->thincVid->num_streams--;
	call_for_all(thincVidEndStreamCl, (cl, stream));
	thincDbg(T_DBG_VID, ("%s: finished. Have %d streams left\n", func,
		pThinc->thincVid->num_streams));
	return T_OK;
}

static int thincVidEndStreamCl(thincClientPtr cl, thincVideoStreamPtr stream)
{
	t_dbg_name(func)
	thinc_vidEnd *end;
	int nobuffer = T_FALSE;
	int err;

	thincDbg(T_DBG_VID, ("%s: terminating stream with client %s\n",
		func, cl->id));

	if(!cl->doVideo) {
		thincDbg(T_DBG_VID, ("%s: client doesn't support video\n",
			func));
		return T_OK;
	}
	end = (thinc_vidEnd*) thincBufferMsg(cl, T_VIDEO_END, 0,
		T_VIDEO_END_SIZE, 0);
	if(!end) {
		thincDbg(T_DBG_VID, ("%s: Couldn't buffer VIDEO END "
			"message\n", func));
		thincFlushBuffer(cl);
		nobuffer = T_TRUE;
		end = (thinc_vidEnd*) t_msgBuffer;
	}
	end->id = stream->id;
	if(nobuffer) {
		err = thincSendMsg(cl, T_VIDEO_END, T_VIDEO_END_SIZE, 0);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO END message\n", func));
			return T_ERROR;
		}
	}

	cl->video->count--;
	thincDbg(T_DBG_VID, ("%s: finished, client has %u active streams\n", 
		func, cl->video->count));
	return T_OK;
}

/***** VIDEO STREAM MODIFICATION ********/
static int thincMoveStream(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, int x, int y)
{
	t_dbg_name(func)
	thincClientPtr cl;
	int res;

	thincDbg(T_DBG_VID, ("%s: moving stream %d to x %d, y %d\n", func,
		stream->id, x, y));

	thincError("moving video to %d %d\n", x, y);

	stream->x = x;
	stream->y = y;
	call_for_all(thincMoveStreamCl, (cl, stream));
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

static int thincMoveStreamCl(thincClientPtr cl, thincVideoStreamPtr stream)
{
	t_dbg_name(func)
	thinc_vidMove *move;
	int err, nobuffer = T_FALSE;
	
	if(!cl->doVideo) {
		thincDbg(T_DBG_VID, ("%s: client doesn't support video\n",
			func));
		return T_OK;
	}
	move = (thinc_vidMove*) thincBufferMsg(cl, T_VIDEO_MOVE, 0,
		T_VIDEO_MOVE_SIZE, 0);
	if(!move) {
		thincFlushBuffer(cl);
		nobuffer = T_TRUE;
		move = (thinc_vidMove*) t_msgBuffer;
	}
	move->id = htonl(stream->id);
	move->x = htons(stream->x);
	move->y = htons(stream->y);
	if(nobuffer) {
		err = thincSendMsg(cl, T_VIDEO_MOVE, T_VIDEO_MOVE_SIZE,
			0);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO MOVE message\n", func));
			return T_ERROR;
		}
	}
	return T_OK;
}

static int thincScaleStream(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, int dstw, int dsth)
{
	t_dbg_name(func)
	thincClientPtr cl;
	int res;

	thincDbg(T_DBG_VID, ("%s: scaling stream %d to w %d, h %d\n", func,
		stream->id, dstw, dsth));

	stream->dst_width = dstw;
	stream->dst_height = dsth;
	call_for_all(thincScaleStreamCl, (cl, stream));
	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

static int thincScaleStreamCl(thincClientPtr cl, thincVideoStreamPtr stream)
{
	t_dbg_name(func)
	thinc_vidScale *scale;
	int err, nobuffer = T_FALSE;

	if(!cl->doVideo) {
		thincDbg(T_DBG_VID, ("%s: client doesn't support video\n",
			func));
		return T_OK;
	}
	scale = (thinc_vidScale*) thincBufferMsg(cl, T_VIDEO_SCALE, 0,
		T_VIDEO_SCALE_SIZE, 0);
	if(!scale) {
		thincFlushBuffer(cl);
		nobuffer = TRUE;
		scale = (thinc_vidScale*) t_msgBuffer;
	}
	scale->id = htonl(stream->id);
	scale->dst_width = htons(stream->dst_width);
	scale->dst_height = htons(stream->dst_height);
	if(nobuffer) {
		err = thincSendMsg(cl, T_VIDEO_SCALE, T_VIDEO_SCALE_SIZE,
			0);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO SCALE message\n", func));
			return T_ERROR;
		}
	}

	return T_OK;
}
	
/***** VIDEO STREAM DELIVERY ***********/
/* we use a messaage stack similar to that used for RAW and BITMAP commands.
 * Depending on the situations we add and delete modules from the 
 * message stack. */
static int thincVidSendImage(thincScreenPtr pThinc, thincVideoStreamPtr stream, 
	unsigned char *image, int x, int y, int w, int h, int dstw, int dsth)
{
	t_dbg_name(func)
	int err = T_OK;

	thincDbg(T_DBG_VID, ("%s: sending image for stream %d\n", func,
		stream->id));

	/* check if stream has changed, and take appropiate actions */
	if(x != stream->x || y != stream->y ) {
		thincDbg(T_DBG_VID, ("%s: video stream moved from (%d,%d) to "
			"(%d, %d)\n", func, stream->x, stream->y, x, y));
		/* Move video */
		err = thincMoveStream(pThinc, stream, x, y);
	}
	if( (dstw != stream->dst_width) || (dsth != stream->dst_height) ) {
		thincDbg(T_DBG_VID, ("%s: video stream scaled from (%d,%d) to "
			"(%d, %d)\n", func, stream->dst_width, 
			stream->dst_height, dstw, dsth));
		err = thincScaleStream(pThinc, stream, dstw, dsth);
	}
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't modify stream "
			"characteristics\n", func));
		return T_ERROR;
	}

	/* FIXME: flush the command buffer before sending the video. There
	 * should be a centralized place where functions that don't buffer
	 * stuff can ask the server to clean things up for them before
	 * they can go ahead and break the order */
	err = thincFlushCmdAll(pThinc);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush command "
			"buffer for clients before sending video frame\n",
			func));
		return T_ERROR;
	}

	/* FIXME: Should use function pointers! */
	switch(stream->format.id) {
		case FOURCC_YV12:
			err = thincVidSendImageYUV12(pThinc, stream,
				image);
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Unsupported image "
				"format %#x\n", func, stream->format.id));
			return T_ERROR;
	}

	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send next image "
			"for stream %d\n", func, stream->id));
		return T_ERROR;
	}

	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

static int thincVidSendImageYUV12(thincScreenPtr pThinc, 
	thincVideoStreamPtr stream, unsigned char *image)
{
	t_dbg_name(func)
	unsigned int sendstamp;
	thincClientPtr cl;
	int res;

	thincDbg(T_DBG_VID, ("%s: sending YUV12 image for stream %d\n", func,
		stream->id));

	gettime_us(sendstamp);
	thincDbg(T_DBG_VID, ("%s: video sendstamp is %u\n", func, sendstamp));
	call_for_all(thincVidSendImageYUV12Cl, (cl, stream, image));

	thincDbg(T_DBG_VID, ("%s: finished\n", func));
	return T_OK;
}

static int thincVidSendImageYUV12Cl(thincClientPtr cl, 
	thincVideoStreamPtr stream, unsigned char *image)
{
	thincVideoObj obj;

	if(!cl->doVideo)
		return T_OK;
#if 0	/* If can't write right now, drop the frame. Not used for now because
	   it makes video playback too sensitive to small fluctuations in the
	   network. need something more resilient */
	const char *func = __func__;
	int err;

	err = thincDoPoll(cl->sock, POLLOUT, 0);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't flush video "
			"frame\n", func));
		return T_ERROR;
	} else if(err == T_TIMEOUT) {
		++cl->parent->droppedFrames;
		thincDbg(T_DBG_ANNOY, ("%s: Client is busy, dropping frame "
			"(%d)\n", func, cl->parent->droppedFrames));
		return T_OK;
	} else {
		obj.data = image;
		obj.stream = &cl->video->streams[stream->id];
		return thinc_msgStackExec(cl, cl->video->stack, &obj);
	}
#else
	obj.data = image;
	obj.stream = &cl->video->streams[stream->id];
	return thinc_msgStackExec(cl, cl->video->stack, &obj);
#endif
}

/* The modules */
static int thinc_videoSend(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, videoSend.name)
	thinc_vidNext *next;
	int err, nobuffer = T_FALSE;
	thincVideoObj *obj = msg;
	unsigned char *data;

	next = (thinc_vidNext*) thincBufferMsg(cl, T_VIDEO_NEXT, 0,
		T_VIDEO_NEXT_SIZE, obj->stream->size);
	if(!next) {
		thincDbg(T_DBG_VID, ("%s: Couldn't buffer VIDEO_NEXT "
			"message\n", modname));
		thincFlushBuffer(cl);
		next = (thinc_vidNext*)t_msgBuffer;
		nobuffer = T_TRUE;
	}

	/* get timestamp */
	next->id = htonl(obj->stream->stream->id);
	/* potentially resized size */
	next->size = htonl(obj->stream->size);
	next->timestamp = htonl(obj->stream->stream->timestamp);

	thincDbg(T_DBG_VID, ("%s: sending video frame of size %u\n", modname,
		obj->stream->size));
	if(nobuffer) {
		err = thincSendMsg(cl, T_VIDEO_NEXT, T_VIDEO_NEXT_SIZE,
			0);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO NEXT message\n", modname));
			return T_ERROR;
		}
		/* send the video data */
		err = thincWriteData(cl, obj->data, obj->stream->size);
		if(err == T_ERROR) {
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't send "
				"VIDEO NEXT data\n", modname));
			return T_ERROR;
		}
	} else {
		data = ((unsigned char*)next) + T_VIDEO_NEXT_SIZE;
		memcpy(data, obj->data, obj->stream->size);
	}

	thincDbg(T_DBG_VID, ("%s: finished\n", modname));
	return T_OK;
}

static int thinc_videoCompress(thincClientPtr cl, void *msg)
{
	return T_OK;
}

/* check if the
 * video is bigger than the client's viewport and resize. If viewport is
 * smaller, shouldn't do anything
 * Remember to change stream->size (client's stream) */
static int thinc_videoResize(thincClientPtr cl, void *msg)
{
	t_dbg_modname(modname, videoResize.name)
	thincVideoObj *obj = msg;
	static unsigned char *rdata = NULL;
	static unsigned int rdata_size = 0;
	thincClVideoStreamPtr cstream = obj->stream;
	int err;

	thincDbg(T_DBG_VID, ("%s: resizing video frame\n", modname));

	if(! (cstream->flags & T_CLVIDEO_RESIZED) ) {
		thincDbg(T_DBG_VID, ("%s: frame not resized, finished\n",
			modname));
		return T_OK;
	}

	if(rdata) {
		/* worth doing this since all frames have the same size */
		if(cstream->size != rdata_size) {
			rdata = realloc(rdata, cstream->size);
		}
	} else {
		rdata = malloc(cstream->size); 
	}
	if(!rdata) {
		thincDbg(T_DBG_ERROR, ("%s: Couldn't allocate memory for "
			"resized video frame\n", modname));
		return T_ERROR;
	}
	rdata_size = cstream->size;

	thincDbg(T_DBG_VID, ("%s: resizing frame to %ux%u\n", modname,
		cstream->width, cstream->height));

	err = thinc_resizeyuv(
		obj->data + cstream->stream->yuv_offsets[0],
		obj->data + cstream->stream->yuv_offsets[1],
		obj->data + cstream->stream->yuv_offsets[2],
		rdata + cstream->yuv_offsets[0],
		rdata + cstream->yuv_offsets[1],
		rdata + cstream->yuv_offsets[2],
		cstream->stream->width,	cstream->stream->height, 
		cstream->width, cstream->height);
	if(err == T_ERROR) {
		thincDbg(T_DBG_ERROR, ("ERROR: %s: Couldn't resize video "
			"frame\n", modname));
		return T_ERROR;
	}

	obj->data = rdata;

	thincDbg(T_DBG_VID, ("%s: finished\n", modname));
	return T_OK;
}

/******* end modules ********/

/******************* FUNCTIONS REQUIRED BY XVIDEO *************/
/* let the driver specify how data for an image of size widthxheight should
 * be stored. Should return size of the memory required for the image.
 * And should correct width and height and if not NULL, pitches and offsets 
 */
/* TODO: Seems like this function is always called before PutImage.
 * Maybe we can send the message about the video to the client here to
 * buy some time.
 */
/* pitches -> width of one line
 * offsets -> distance to each plane
 */
static int thincVidQueryImageAttributes(ScrnInfoPtr pScrn, int id, 
	unsigned short *w, unsigned short *h, int *pitches, 
	int *offsets)
{
	t_dbg_name(func)
	int size, tmp;

	thincDbg(T_DBG_VID, ("%s: id %#x, width %d, height %d\n", 
		func, id, *w, *h));

	/* this is same code as all drivers currently have */
	if(*w > 2046) *w = 2046;
	if(*h > 2046) *h = 2046;

	/* make w multiple of 4 so that resizing works properly */
	*w = (*w + 3) & ~3;
	if(offsets) offsets[0] = 0;

	switch(id) {
		case FOURCC_YV12:
			/* make h be even */
			*h = (*h + 1) & ~1;
			/* make w be multiple of 4 (ie. pad it) */
			size = (*w + 3) & ~3;
			/* width of a Y row => width of image */
			if(pitches) pitches[0] = size;
			/* offset of U plane => w*h */
			size *= *h;
			if(offsets) offsets[1] = size;
			/* width of U, V row => width/2 */
			tmp = ((*w >> 1) +3) & ~3;
			if(pitches) pitches[1] = pitches[2] = tmp;
			/* offset of V => Y plane + U plane (w*h + w/2*h/2) */
			tmp *= (*h >> 1);
			size += tmp;
			if(offsets) offsets[2] = size;
			size += tmp;
			break;
		default:
			thincDbg(T_DBG_ERROR, ("ERROR: %s: Unsupported image "
				"format %#x\n", func, id));
			return 0;
	}

	thincDbg(T_DBG_VID, ("%s: finished. size %d\n", func, size));
	return size;
}

/* FUNCTIONS PART OF THE XVIDEO INTERFACE BUT UNUSED FOR NOW *********/
/* If Cleanup is false the video is being stopped temporarily due to clipping
 * or moving of the window, etc...  and video will likely be restarted soon so
 * the driver should not deallocate any offscreen areas associated with that
 * port.
 */
void thincVidStopVideo(ScrnInfoPtr pScrn, pointer data, Bool Cleanup)
{
	t_dbg_name(func)
	long port_id = (long)data;

	thincVideoStreamPtr stream;
	thincVideoPtr pVid;
	t_vidPre(pScrn);

	if(!Cleanup)
		return;

	thincDbg(T_DBG_VID, ("%s: stopping video for port %ld\n", func, 
		port_id));

	pVid = pThinc->thincVid;
	/* FIXME: Check port_id is valid index! */
	stream = &pVid->streams[port_id];

	thincVidEndStream(pThinc, stream);

	thincDbg(T_DBG_VID, ("%s: finished\n", func));
}

int thincVidSetPortAttribute(ScrnInfoPtr pScrn, Atom attribute,
	INT32 value, pointer data)
{
	const char *func = __func__;

	thincError("%s: I was called!\n", func);

	return Success;
}

int thincVidGetPortAttribute(ScrnInfoPtr pScrn, Atom attribute,
	INT32 *value, pointer data)
{
	const char *func = __func__;

	thincError("%s: I was called!\n", func);

	return Success;
}

void thincVidQueryBestSize(ScrnInfoPtr pScrn, Bool motion,
	short vid_w, short vid_h, short drw_w, short drw_h, 
	unsigned int *p_w, unsigned int *p_h, pointer data)
{
	const char *func = __func__;

	thincError("%s: I was called!\n", func);
}

#else
/*********** NO VIDEO SUPPORT **************/
/*** just null entry point functions *******/
int thincInitVideo(thincScreenPtr pThinc)
{
	return T_TRUE;
}

/* Shouldn't be called if we don't have video support */
int thincHandleVidStartOK(thincClientPtr cl)
{
	return T_FALSE;
}
#endif /* XvExtension */
