/* Copyright 2003, 2004, 2005 Ricardo Baratto */
/* This file is part of THINC.

    THINC is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    THINC is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with THINC.  If not, see <http://www.gnu.org/licenses/>.
*/

/*
 *
 * Authentication library for THINC
 *
 * Author:
 * Sheikh M. Omer-Farooq <smo2001@columbia.edu>
 * Columbia University
 */

/*
 * authenticates a user based on
 *  1. session owner == user
 *  2. user has an account on the machone
 *  3. user provides correct info for account on machine
 */

#include "thinc_auth.h"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#ifndef T_WIN32
#include <termios.h>
#include <unistd.h>
#include <sys/types.h>
#include <pwd.h>
#endif /* T_WIN32 */

#define T_AUTH_DEFAULT_LEN	1

/* local funcs */
static int thinc_conv(int num_msg, const struct pam_message **msgm,
	struct pam_response **response, void *appdata_ptr);
static int thinc_check_login(struct thinc_user *info);

int thinc_authenticate(struct thinc_user* info, const char *owner,
	unsigned int owner_len)
{
#ifdef T_WIN32
/*
 TODO: We don't want to do anything right now
*/
	return T_AUTH_OK;
#else
	unsigned int len;
	/* check to see if all the params passed in are not null */
	if (!info || !info->user || !info->passwd || !owner)
		return T_AUTH_ERROR;

	len = (info->user_len <= owner_len ? owner_len : info->user_len);

	/* chk user and owner match */
	if (!strncmp(info->user, owner, len)) {
		/* kool. then try to authenticate the user */
		return thinc_check_login(info);
	} else {
		return T_AUTH_NOMATCH;
	}
#endif /* T_WIN32 */
}

/* returns username and password in info (it will allocate memory for them) */
int thinc_get_user_info(struct thinc_user* info)
{
#ifdef T_WIN32
/*
 TODO: We don't want to do anything right now
*/
#else
	char c = ' ';
	unsigned int x = 0, size = T_AUTH_DEFAULT_LEN;
	struct termios old_set, new_set;

	info->user = malloc(size * sizeof(char));
	if(!info->user)
		return T_AUTH_ERROR;

	fprintf(stdout, "Please enter your username: ");

	while ( (c = getchar()) != '\n') {
		/* we read chars till we get to newline */
		if (x == size-1) {
			size *= 2;
			/* if we are abt to exceed size resize buffer */
			info->user = realloc(info->user, size*sizeof(char));
			if(!info->user) {
				return T_AUTH_ERROR;
			}
		}
		/* save char */
		info->user[x++]= c;
	}

	info->user[x] = '\0';

	/* shrink down to actual size */
	info->user = realloc(info->user, (x+1)*sizeof(char));
	if(!info->user) {
		/* impossible :( */
		return T_AUTH_ERROR;
	}
	/* doesn't include the '\0' */
	info->user_len = x;

	/* now the password */
	x=0, c =' ', size = T_AUTH_DEFAULT_LEN;

	info->passwd = malloc(size * sizeof(char));
	if(!info->passwd) {
		free(info->user);
		return T_AUTH_ERROR;
	}

	fprintf(stdout, "Please enter your password: ");

	/* turn terminal echo off */
	tcgetattr(0,&old_set); /* get the settings of the current terminal */
	new_set = old_set;
	new_set.c_lflag &= (~ECHO);  /* set the echo flag to off */
	/* set current attrib to newset, change is immediate */
	tcsetattr(0,TCSANOW,&new_set);

	/* terminal echo is off, can now get the password */
	while ( (c = getchar()) != '\n') {
		/* we read chars till we get to newline */
		if (x > size - 3) {
			size *= 2;
			/* if we are abt to exceed size resize buffer */
			info->passwd = realloc(info->passwd, size*sizeof(char));
			if(!info->passwd) {
				free(info->user);
				return T_AUTH_ERROR;
			}
		}
		info->passwd[x++]= c; /* save char */
	}

	info->passwd[x] = '\0';

	/* revert back to old settings change immediate */
	tcsetattr(0,TCSANOW,&old_set);
	fprintf(stdout, "\n");

	/* shrink down to actual size */
	info->passwd = realloc(info->passwd, (x+1)*sizeof(char));
	if(!info->passwd) {
		/* no way :( */
		free(info->user);
		return T_AUTH_ERROR;
	}
	/* doesn't include the '\0' */
	info->pass_len = x;
#endif /* T_WIN32 */
	return T_AUTH_OK;
}

/* it also clears the password before freeing it */
void thinc_free_user_info(struct thinc_user *info)
{
#ifdef T_WIN32
/*
 TODO: We don't want to do anything right now
*/
#else
	memset(info->passwd, 0, info->pass_len+1);
	free(info->user);
	free(info->passwd);
	/* we're cleanup freaks! */
	info->user = info->passwd = NULL;
	info->user_len = info->pass_len = 0;
#endif /* T_WIN32 */
}

/* return the name associated with uid */
int thinc_get_owner(int uid, char **owner, unsigned int *len)
{
#ifdef T_WIN32
/*
 TODO: We don't want to do anything right now
*/
#else
       struct passwd *p = getpwuid(uid);
	if(!p)
		return T_AUTH_ERROR;

	*len = strlen(p->pw_name);
	*owner = malloc(*len+1);	/* +1 for '\0' */
	if(!*owner)
		return T_AUTH_ERROR;

	strncpy(*owner, p->pw_name, *len);
	(*owner)[*len] = '\0';
#endif /* T_WIN32 */
	return T_AUTH_OK;
}

static int thinc_conv(int num_msg, const struct pam_message **msgm,
	struct pam_response **response, void *appdata_ptr)
{
#ifdef T_WIN32
/*
 TODO: We don't want to do anything right now
*/
	return T_AUTH_OK;
#else
	struct pam_response *thinc_resp;
	unsigned int i;
	struct thinc_user* info;

	if (appdata_ptr == NULL)
		return PAM_CONV_ERR;

	info = (struct thinc_user*)appdata_ptr; 

	/* if num pam messages is not in range */
	if (num_msg <= 0 || num_msg > PAM_MAX_NUM_MSG)
		return PAM_CONV_ERR;

	thinc_resp = calloc(num_msg, sizeof(*thinc_resp));
	if(!thinc_resp)
		return PAM_BUF_ERR;

	/* this loop goes through all the messages that are passed to the
	 * conversation struct and for each message based on what the message
	 * is, it gives the response.  */
	for (i = 0; i < num_msg; ++i) {
		thinc_resp[i].resp_retcode = 0;
		thinc_resp[i].resp = NULL;

		switch (msgm[i]->msg_style) {
			case PAM_PROMPT_ECHO_OFF:
				/* pam wants password with no echo
				 * we send pam the password which we place
				 * in the appdata_ptr param */
				thinc_resp[i].resp =
					malloc(info->pass_len+1);
				if(!thinc_resp[i].resp) {
					goto fail;
				}

				strncpy(thinc_resp[i].resp, info->passwd,
					info->pass_len+1);
				break;
			case PAM_PROMPT_ECHO_ON:
				/* pam wants password with echo
				 * we send pam the password which we place in
				 * the appdata_ptr param */
				thinc_resp[i].resp =
					malloc(info->pass_len+1);
				if(!thinc_resp[i].resp) {
					goto fail;
				}

				strncpy(thinc_resp[i].resp, info->passwd,
					info->pass_len+1);
				break;
			case PAM_ERROR_MSG:
				/* pam wants to print error message */
				/* print to stderr */
				fprintf(stderr, "%s\n", msgm[i]->msg);
				break;
			case PAM_TEXT_INFO:
				/* pam wants to print some info */
				/* prinf to stdout */
				fprintf(stdout, "%s\n", msgm[i]->msg);
				break;
			default:
				goto fail;
		}
	}

	*response = thinc_resp;  /* this is our response */
	return PAM_SUCCESS;

fail:
	/* here we free everything , if something wasn't right */
	for (i = 0; i < num_msg; ++i) {
		if (thinc_resp[i].resp) {
			free(thinc_resp[i].resp);
		}
	}

	*response = NULL;
	return PAM_CONV_ERR;
#endif	/*T_WIN32*/
}

static int thinc_check_login(struct thinc_user *info)
{
#ifdef T_WIN32
/*
 TODO: We don't want to do anything right now
*/
#else

	pam_handle_t *pamhandle = NULL;
	int val;

	/* construct a conversation struct */
	struct pam_conv conv = {
		thinc_conv,
		NULL
	};

	conv.appdata_ptr = (void *)info;

	/* start pam using the thinc module and our conversation struct */
	val = pam_start("thinc", info->user, &conv, &pamhandle);
	if (val != PAM_SUCCESS)
		goto end;

	/* check if user is valid */
	val = pam_authenticate(pamhandle, PAM_SILENT);
	if (val != PAM_SUCCESS)
		goto end;

	/* check if this user has permissiong */
	val = pam_acct_mgmt(pamhandle, 0);
	if(val != PAM_SUCCESS)
		goto end;

end:
	if (pam_end(pamhandle,val) != PAM_SUCCESS || val != PAM_SUCCESS)
		return T_AUTH_ERROR;
#endif /* T_WIN32 */
	return T_AUTH_OK;
}
